#========================================================================================
# package      DE_EPAGES::Content::API::Object::ContentObject
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for ContentObject.
#========================================================================================
package DE_EPAGES::Content::API::Object::ContentObject;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use DE_EPAGES::Core::API::Image qw ( GetImageInfo );
use DE_EPAGES::Core::API::PerlTools qw( ElementOf trim );
use DE_EPAGES::Content::API::Table::ShortURL qw ( DeleteShortURL GetInfoShortURLsByContentObjectID );
use DE_EPAGES::Content::API::Table::ShortUrlRedirection qw( DeleteShortUrlRedirection
                                                            GetPKeyShortUrlRedirectionsBySiteID
                                                            InsertShortUrlRedirection );
use DE_EPAGES::Core::API::Error qw (Error);
use DE_EPAGES::Object::API::Factory qw (LoadClassByAlias);
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Presentation::API::PageType qw (LoadPageTypeByViewAction);
use DE_EPAGES::Permission::API::Permission qw (ActionByName);

#========================================================================================
# function     featureName
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FeatureName = $Forum->featureName;
#----------------------------------------------------------------------------------------
# description  Returns the feature name.
#----------------------------------------------------------------------------------------
# return       $FeatureName | feature name (undef means no feature count) | String
#========================================================================================
sub featureName () { return undef; }

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $ContentObject->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'ContentObject' }

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $ContentObject->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     requiredChildClass
# state        public
#----------------------------------------------------------------------------------------
# syntax       $RequiredChildClass = $ContentObject->requiredChildClass;
#----------------------------------------------------------------------------------------
# description  Returns the name of the class a new child is required to be an instance
#               of or undef if no class is allowed (no children).
#----------------------------------------------------------------------------------------
# return       $RequiredChildClass | name of a class, i.e. "Object" | String
#========================================================================================
sub requiredChildClass { return 'ContentObject'; }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ContentObject = DE_EPAGES::Content::API::Object::ContentObject->insert($hInfo);
# example      my $Class = LoadClassByAlias('ContentObject');
#               my $ContentObject = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertContentObject' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object | ref.hash
# return       $ContentObject | new object | object
# hook         OBJ_InsertContentObject | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ContentObjectID - object id            - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    my $Parent = $hInfo->{'Parent'};
    Error('OBJECT_EXISTS', { 'Alias' => $hInfo->{'Alias'} }) if $Parent->existsChild($hInfo->{'Alias'});

    my $Shop = $Parent->getSite();

    # increment the usage counter of the Pages feature
    my $FeatureName = $class->featureName;
    $Shop->featureIncrementValue($FeatureName) if defined $FeatureName;

    my $self = $class->SUPER::insert($hInfo);
    $self->correctTemplateType();

    TriggerHook('OBJ_InsertContentObject', {'Object' => $self, 'ContentObjectID' => $self->id });
    delete $hInfo->{$_} foreach qw ( IsVisible Date VisibleInNavigation ContentSite ImageSmall ImageMedium ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ContentObject->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteContentObject' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteContentObject | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ContentObjectID - object id            - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteContentObject', {'Object' => $self, 'ContentObjectID' => $self->id });
    LoadClassByAlias('ContentTag')->deleteObjects('ByContentObjectID', [$self->id]);

    my $hOldShortURL = {
        'Object' => $self,
    };

    $hOldShortURL->{'ContentObjectID'} = $self->get('Parent')->id;
    # Get correct parent of the object.
    TriggerHook('API_BeforeInsertShortURLRedirection', $hOldShortURL);

    foreach my $hShortURL (@{GetInfoShortURLsByContentObjectID($self->id)}) {
        DeleteShortURL($hShortURL->{'ShortUrlID'});
        # Do not save redirects to the shop object.
        if(defined $hOldShortURL->{'ContentObjectID'} and $hOldShortURL->{'ContentObjectID'} != $hShortURL->{'SiteID'}) {
            # Insert deleted shorturl in shorturlredirection table.
            delete $hShortURL->{'ShortUrlID'};
            $hShortURL->{'ContentObjectID'} = $hOldShortURL->{'ContentObjectID'};

            # delete any previous redirections for the same uri
            my ($SiteID, $Uri) = @{$hShortURL}{qw(SiteID Uri)};
            DeleteShortUrlRedirection($_) foreach @{GetPKeyShortUrlRedirectionsBySiteID($SiteID, $Uri)};

            InsertShortUrlRedirection($hShortURL);
        }
    }

    # remember the shop and feature
    my $Shop   = $self->getSite();
    my $FeatureName = $self->featureName;
    my $SelfClass = $self->class();

    $self->SUPER::delete;

    # decrement the usage counter of the feature
    $Shop->featureIncrementValue( $FeatureName, -1) if defined $FeatureName;
    TriggerHook('API_AfterDeleteContentObject', {'ObjectClass' => $SelfClass, 'ShopObject' => $Shop, 'ShopContentObjectID' => $Shop->id });
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ContentObject->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateContentObject' and 'OBJ_AfterUpdateContentObject'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateContentObject | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ContentObjectID - object id            - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateContentObject | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ContentObjectID - object id            - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID, $hAdditionalOptions) = @_;

    TriggerHook('OBJ_BeforeUpdateContentObject', {'Object'=>$self, 'Values'=> $hValues, 'ContentObjectID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID, $hAdditionalOptions);
    TriggerHook('OBJ_AfterUpdateContentObject',  {'Object'=>$self, 'Values'=> $hValues, 'ContentObjectID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     canInvalidatePageCache
# state        public
#----------------------------------------------------------------------------------------
# syntax       $CanInvalidate = $Object->canInvalidatePageCache;
#----------------------------------------------------------------------------------------
# description  Overwrites method DE_EPAGES::Object::API::Object::Object->canInvalidatePageCache
#               to invalidate page cache.
#----------------------------------------------------------------------------------------
# return       $CanInvalidate | true if the object affects the page cache | boolean
#========================================================================================
sub canInvalidatePageCache {
    return 1;
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Content::API::Object::ContentObject->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'IsVisible' => 0,
        'VisibleInNavigation' => 1,
    };
}

#========================================================================================
# function     template
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Name = $Object->template($Name, $ObjectPageType);
#----------------------------------------------------------------------------------------
# description  This function is used to overwrite the given template for the purpose that
#               each content object can defined there own template (visualisation).
#----------------------------------------------------------------------------------------
# input        $Name | include name | string
# input        $ObjectPageType | object related pagetype | DE_EPAGES::Presentation::API::Object::ObjectPageType
# return       $Name | include name | string
#========================================================================================
sub template { ## no critic qw(RequireCommentForExportedFunctions RequireArgUnpacking)
    my $self = shift;
    return $self->classTemplate(@_);
}

#========================================================================================
# function     correctTemplateType
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Object->correctTemplateType();
#----------------------------------------------------------------------------------------
# description  Correct a TemplateType of a IncludeTemplates folder to a featured
#               TemplateType if needed. This is useful, when the default template type
#               of a template group has a unfulfilled feature.
#========================================================================================
sub correctTemplateType {
    my $self = shift;

    my $Shop = $self->get('Site');

    # most likely the shop is in creation right now, so features are not set correctly
    return if $Shop->get('ClosedOrClosedTempByProvider');

    my $Action = ActionByName( $self->class, 'View' );
    my $PageType = LoadPageTypeByViewAction( $Action );

    foreach my $TemplateTypeFolder (@{$PageType->templateTypeFolders}) {
        my $Template = $PageType->getTemplate( $TemplateTypeFolder->alias );
        my $TemplateType = $Template->getTemplateMap($self)->templateType;
        my $Feature = $TemplateType->get('Feature');
        if( defined $Feature && 0 == $Shop->featureMaxValue($Feature) ) {
            my $FirstAllowedTemplateType = _FindFirstAllowedTemplateType( $TemplateTypeFolder, $Shop );
            if (defined $FirstAllowedTemplateType) {
                $Template->setTemplateMap($self, $FirstAllowedTemplateType);
            }
        }
    }
    return;
}

# find first template type in list, which has a valid feature
sub _FindFirstAllowedTemplateType {
    my ($TemplateTypeFolder, $Shop) = @_;

    foreach my $TemplateType ( @{ $TemplateTypeFolder->children } ) {
        my $Feature = $TemplateType->get('Feature');
        return $TemplateType if( !defined $Feature || $Shop->featureMaxValue($Feature) );
    }
    return undef;
}

sub effectedVisibilityAttributes {
    return {
        'IsVisibleContentView'=>1,
        'VisibleInNavigation'=>1,
        'ReadPermissionFor'=>1
    };
}

#========================================================================================
# function     isAllowedChild
# state        public
#----------------------------------------------------------------------------------------
# syntax       $isAllowed = $Object->isAllowedChild($PossibleChild);
#----------------------------------------------------------------------------------------
# description  Returns if an Object could be the parent of a PossibleChild.
#----------------------------------------------------------------------------------------
# input        $PossibleChild | proposed child object | DE_EPAGES::Content::API::Object::ContentObject
# return       $isAllowed | is a allowed child type | boolean
#========================================================================================
sub isAllowedChild {
    my $self = shift;
    my ($PossibleChild) = @_;

    return 0 unless $PossibleChild->get('Parent');
    foreach my $ChildClass (@{$self->class->get('ContentChildClasses')}) {
        return 1 if $PossibleChild->instanceOf($ChildClass->alias);
    }
    return 0;
}

#========================================================================================
# function     clone
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Clone = $Object->clone( $hInfo );
# syntax       $Clone = $Object->clone( $hInfo, $Recursive );
#----------------------------------------------------------------------------------------
# description  Inserts a new object into the database as copy from old object.
#----------------------------------------------------------------------------------------
# input        $hInfo | hash with attribute values that are different in the clone object.
#                       Note that the clone cannot have the same Parent and the same Alias.
#                    | hash ref
# input        $Recursive | clone child object too (optional, default is false) | boolean
# return       $Clone | the new object | object
#========================================================================================
sub clone {
    my $self = shift;
    my ($hInfo, $Recursive) = @_;

    my $hAttributes = $self->exportableAttributes;
    # set URI undef for each language
    if (defined $hAttributes->{'URI'} && $hAttributes->{'URI'} ne '') {
        foreach my $LanguageID (keys %{$hAttributes->{'URI'}}) {
            $hInfo->{'URI'}{$LanguageID} = undef;
        }
    }
    return $self->SUPER::clone($hInfo, $Recursive);
}

#========================================================================================
# function     createURI
# state        public
#----------------------------------------------------------------------------------------
# syntax       $URI = $Object->createURI($Name, $Parent, $LanguageID);
#----------------------------------------------------------------------------------------
# description  Returns a suggestion for a short url.
#----------------------------------------------------------------------------------------
# input        $Name | suggested URI | string
# input        $Parent | parent object | object
# input        $LanguageID | language id | integer
# return       $URI | URI path | string
#========================================================================================
sub createURI {
    my $class = shift;
    my ($Name, $Parent, $LanguageID) = @_;
    return $Name;
}

#========================================================================================
# function     defaultObject
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ContentObject = $Object->defaultObject();
#----------------------------------------------------------------------------------------
# description  Returns the default object (has to be overwritten for products with variation).
#----------------------------------------------------------------------------------------
# return       $ContentObject | content object | object
#========================================================================================
sub defaultObject {
    my $self = shift;
    return $self;
}

#========================================================================================
# function     exportableChildren
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aChildren = $Object->exportableChildren
#----------------------------------------------------------------------------------------
# example      $Description = $Object->exportableChildren;
#----------------------------------------------------------------------------------------
# description  Returns all children of the object which should be exported.
#----------------------------------------------------------------------------------------
# return       $aChildren | children of the object | array ref
#========================================================================================
sub exportableChildren {
    my $self = shift;
    my $Shop = $self->getSite();
    my @Children = grep {
        !$_->can('featureName') ||
        !defined $_->featureName ||
        $Shop->featureMaxValue($_->featureName) >= 1
    } @{$self->children};

    return \@Children;
}

#========================================================================================
# function     useFilePathPart
# state        public
#----------------------------------------------------------------------------------------
# syntax       $useFilePathPart = $Object->useFilePathPart;
#----------------------------------------------------------------------------------------
# description  Use FilePathPart for content objects.
#----------------------------------------------------------------------------------------
# return       $useFilePathPart | false if not use filepathpart | boolean
#========================================================================================
sub useFilePathPart {
    return 1;
}

#========================================================================================
# function     addImageInformations
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object->addImageInformations($aAllImageAttributes, $hNewValues);
#----------------------------------------------------------------------------------------
# description  rewrite the JSON attribute ImageInformations for content objects.
#               Should be called by "set" and "clone" functions on the changed content object like products
#----------------------------------------------------------------------------------------
# input        $aAllImageAttributes | all image attribute names for this object | ref.array
# input        $hNewValues | changed values | ref.hash
#========================================================================================
sub addImageInformations {
    my $self = shift;
    my ($aAllImageAttributes, $hNewValues) = @_;


    # test if we have changed image attributes and prepare...
    my $hChangedImageAttributes = {};
    if(defined $hNewValues){
        foreach my $Attribute (@$aAllImageAttributes){
            $hChangedImageAttributes->{$Attribute} = $hNewValues->{$Attribute} if exists $hNewValues->{$Attribute};
        }
    }
    # no actions needed if we have $hNewValues but no image attributes has changed
    return if (defined $hNewValues and (not scalar keys(%$hChangedImageAttributes)));

    # get changed file from $hChangedImageAttributes and 
    # use GetImageInfo for fetching image dimensions
    # write this informations to the object as JSON attribute "ImageInformations"
    # if no entries exists in ImageInformations then they are added
    my $hExistingImageInformations = $self->get('ImageInformations') // {};
    my $ObjectFilePath = $self->get('PublicPath');

    # test if exists and need an update
    foreach my $ImageAttributeName (@$aAllImageAttributes) {
        # rewrite or not exists
        if( (not defined $hNewValues) or (not exists $hExistingImageInformations->{$ImageAttributeName}) ){
           $hExistingImageInformations->{$ImageAttributeName} = $self->getImageInfo($ObjectFilePath, $self->get($ImageAttributeName));
        # changed attribute 
        } elsif( ElementOf($ImageAttributeName, keys(%$hChangedImageAttributes)) ){
           $hExistingImageInformations->{$ImageAttributeName} = $self->getImageInfo($ObjectFilePath, $hChangedImageAttributes->{$ImageAttributeName});
        }
        # cleanup if we have no entries (e.g. if deleted)
        delete $hExistingImageInformations->{$ImageAttributeName} unless defined $hExistingImageInformations->{$ImageAttributeName};
    }

    # set attribute
    $hExistingImageInformations = undef unless keys(%$hExistingImageInformations);
    $self->set({'ImageInformations' => $hExistingImageInformations });

    return ;
}

#========================================================================================
# function     getImageInfo
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hImageInfo = $Object->getImageInfo;
#----------------------------------------------------------------------------------------
# description  get width, height and size for a filename or filenames.
#----------------------------------------------------------------------------------------
# input        $ObjectFilePath | public file path for this object | string
# input        $FileNames | filename or filenames separated by ; | string
# return       $hInfo | hash with file informations | ref.hash
#========================================================================================
sub getImageInfo {
    my $self = shift;
    my ($ObjectFilePath, $FileNames) = @_;
    
    return undef unless $FileNames;

    # Separate file names in case we have more than one image in the attribute (like ImageSlideShowString)
    my @Files = trim(split /;/,$FileNames);
    my $hInfo = {};
    foreach my $FileName (@Files) {
        my $FullFileName = $ObjectFilePath.'/'.$FileName;
        next unless -f $FullFileName;
        my $hImageInfo = GetImageInfo($FullFileName);   
        $hInfo->{$FileName} = { 
                                'Width'  => $hImageInfo->{'Width'},
                                'Height' => $hImageInfo->{'Height'},
                                'Size'   => $hImageInfo->{'Size'},
                               };
    } 
    return $hInfo;
}

1;
