use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Error qw( GetError ExistsError );
use DE_EPAGES::Object::API::Factory qw( LoadObject );
use DE_EPAGES::Core::API::Log qw( LogError );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore GetCurrentDBHandle );
use DE_EPAGES::Shop::API::Feature qw( LoadFeatureByAlias );
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );
use DE_EPAGES::Product::API::Constants qw ( UNITY_IMAGE_ATTRIBUTES);
#========================================================================================
# function         UpdateImageInformations
# state            public
#----------------------------------------------------------------------------------------
# syntax           UpdateImageInformations();
#----------------------------------------------------------------------------------------
# description      Fills the ImageInformations attribute on products for a Unity shop
#========================================================================================
sub UpdateImageInformations {
    my ($NoOutput) = @_;

    print "Reading all Unity shops\n" unless $NoOutput;
    my $Feature = LoadFeatureByAlias('Unity');
    my $aShopIDs = $Feature->findShops;

    my $aAllImageAttributes = [UNITY_IMAGE_ATTRIBUTES];
    my $dbi = GetCurrentDBHandle();

    foreach my $ShopID (@$aShopIDs) {
        # update ContentObject Product (all products and sub products) on shop
        printf "update image informations on shop %s\n", $ShopID unless $NoOutput;       

        eval {

            my $aaResultSet = $dbi->execute(<<'SQL'  ,[$ShopID] );
SELECT p.productid
FROM product p
INNER JOIN contentobject co ON (p.productid = co.contentobjectid)
WHERE p.shopid = ?
ORDER BY p.productid ASC
SQL
            foreach my $aProductID (@$aaResultSet) {
                my $Product = LoadObject($aProductID->[0]);
                printf "create image informations for product %s\n", $aProductID->[0] unless $NoOutput;
                $Product->addImageInformations($aAllImageAttributes, undef);
                $Product->extendImageInformationsSlideShowString(undef);
            };
        };
        warn('Update image informations failed for shop ID ' . $ShopID . " with error \"" . GetError() . "\".") if ExistsError();
    }

    eval { SynchronizeCache(); };
    LogError(GetError()) if ExistsError();

    return;
}

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $Help;
    my $NoOutput;

    GetOptions(
        'help'        => \$Help,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'nooutput'    => \$NoOutput,
    );

    usage() if $Help;

    unless ( defined $StoreName ) {
        print STDERR "Missing parameter -storename.\n\n";
        usage();
    }
    my $time1 = time;

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub { UpdateImageInformations($NoOutput); },
    );
    my $time2 = time;
    printf "need %.1f seconds\n", $time2 - $time1;
}

sub usage {
    print <<END_USAGE;
Description:
    Fills the ImageInformations attribute on products for a Unity shop.

Usage:
    perl $0 [options] [flags] datafiles

options:
    -passwd     database user password                          ()
    -storename  name of store                                   ()
flags:
    -help       show the command line options
    -nooutput   do not print any output

Example:
    perl $0 -storename Store
END_USAGE
    exit 2;
}

RunScript( 'Sub' => \&Main );