#==============================================================================
# §package      DE_EPAGES::Customer::UI::Customer
# §state        private
#==============================================================================
package DE_EPAGES::Customer::UI::Customer;
use base DE_EPAGES::Presentation::UI::Object;

use strict;
use DE_EPAGES::Address::API::Address qw ( IsValidBankName );
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError Error );
use DE_EPAGES::Core::API::Log qw ( LogError );
use DE_EPAGES::Core::API::Password qw ( CryptPassword );
use DE_EPAGES::Customer::API::Shop qw ( InsertUserForCustomer );
use DE_EPAGES::Customer::API::Table::CustomerUser qw (ExistsCustomerUser);
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias );
use DE_EPAGES::Object::API::Table::Country qw ( GetInfoCountry );
use DE_EPAGES::Presentation::API::User qw ( GetUsersByEMail );
use DE_EPAGES::Shop::API::Shop qw ( GetNextNo );
use DE_EPAGES::ShopMailType::API::ShopMailType qw( GetShopMailType );
use DE_EPAGES::MailType::API::NewPassword qw( GetUserNewPasswordUrl );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Design::UI::Address;

#==============================================================================
# Event handlers
#==============================================================================
sub Save  {
    my $self = shift;
    my ($Servlet) = @_;

    my $Form = $Servlet->form;

    my $Customer = $Servlet->object;

    # update customer
    $self->SUPER::Save($Servlet);

    eval {
        # update/insert billing address
        my $BillingAddress = $Customer->get('BillingAddress');
        if (defined $BillingAddress) {
            # EPG-24791: Verify bank names having non-ascii characters or being too long
            my $BankName = $Servlet->params('BankName');
            if ( !IsValidBankName($BankName) ) {
                $Servlet->form->addFormError({
                    Reason => 'BankNameTooLong',
                    Form   => 'CustomerForm',
                    Name   => 'BankName',
                });
                $Servlet->form->executeFormError;
            }
            # update address
            $Servlet->params('ObjectID', $BillingAddress->id);
            $self->SUPER::Save($Servlet);
            # correct servlet.object
            $Servlet->params('ObjectID', $Customer->id);
        } else {
            # insert address
            $Servlet->params('ClassID', LoadClassByAlias('Address')->id);
            $BillingAddress = $self->NewObject($Servlet, $Customer->folder('Addresses'));
            $Customer->set({'BillingAddress' => $BillingAddress});
        }

        #START: synchronize bank account data with user EPG-28128
        my $aCustomerUsers = $Customer->get('Users');
        if (scalar(@$aCustomerUsers)) {
            my $User = $aCustomerUsers->[0];
            my $UserBillingAddress = $User->get('BillingAddress');
            $UserBillingAddress->set({
                'BankCode' => $Form->value('BankCode'),
                'BankAccountHolder' => $Form->value('BankAccountHolder'),
                'BankName' => $Form->value('BankName'),
                'BankAccountNo' => $Form->value('BankAccountNo')
            });
        }
        #END: synchronize bank account data with user EPG-28128
    };
    if (ExistsError()) {
        my $Error = GetError();
        $Servlet->params('ObjectID', $Customer->id);
        Error($Error);
    }

    return;
}

sub UnitySaveCustomer {
    my $self = shift;
    my ($Servlet) = @_;

    my $Customer = $Servlet->object;
    my $Shop = $Customer->getSite();

    # save address to existing customers user and update also customers billing address
    if(scalar(@{$Customer->get('Users')})) {
        my $User = $Customer->get('Users')->[0];

        my $UseShippingAddress = $Servlet->form->value('UseShippingAddress');

        foreach my $AddressType (qw(BillingAddress ShippingAddress)){

            my $FormValues = $Servlet->form->form($Customer, 'UnitySave'.$AddressType);
            my $Address = $User->get($AddressType);

            # test if we have a ShippingAddress in form
            if(not $UseShippingAddress and $AddressType eq 'ShippingAddress'){
                $User->set({'ShippingAddress' => undef});
                next;
            }

            eval {
                # update user addresses (Shipping and Billing)
                my %AddressValues = $self->_prepareAddressValues($AddressType,$FormValues);

                if (defined $Address) {
                    $Address->set(\%AddressValues);
                } else {
                    # add new address for user
                    my $Folder = $User->folder('Addresses');
                    my $NewAddress = $Folder->insertChild('Address',{Alias => $Folder->nextAlias('Address%d')});
                    $NewAddress->set(\%AddressValues);
                    $User->set({$AddressType => $NewAddress});
                }
            };
            if (ExistsError()) {
                my $Error = GetError();
                $Servlet->params('ObjectID', $Customer->id);
                Error($Error);
            }
        }

        #cleanup unused addresses (we allow only one shipping address and one billing address)
        my $aAddresses = $User->folder('Addresses')->children;
        # if we have more than 2 addresses we have to cleanup w.o. destroying needed addresses
        if(scalar(@$aAddresses) > 2) {
            my $hAddresses = $User->get(['ShippingAddress', 'BillingAddress']);
            foreach my $Address (@$aAddresses) {
                next if (defined $hAddresses->{'BillingAddress'}) and ($Address->id eq $hAddresses->{'BillingAddress'}->id);
                next if (defined $hAddresses->{'ShippingAddress'}) and ($Address->id eq $hAddresses->{'ShippingAddress'}->id);
                $Address->delete();
            }
        }

    } else {

        # used for new unregistered customer
        # save address to customer only if we have no user and no order
        unless (scalar(@{$Customer->get('Orders')})){
            my $AddressType = 'BillingAddress';
            my $FormValues = $Servlet->form->form($Customer, 'UnitySave'. $AddressType);
            my $Address = $Customer->get($AddressType);

            eval {
                    # update user billing address
                    my %AddressValues = $self->_prepareAddressValues($AddressType,$FormValues);
                    $Address->set(\%AddressValues);
                };
            if (ExistsError()) {
                my $Error = GetError();
                $Servlet->params('ObjectID', $Customer->id);
                Error($Error);
            }
        }

    }

    # save new Alias....
    my $NewAlias = $Servlet->form->value('Alias');

    unless ($NewAlias =~ /\S/){
        $Servlet->form->addFormError({
            Reason => 'AliasEmpty',
            Form   => 'CustomerForm',
            Name   => 'Alias',
        });
    }

    # test if we have already the Alias
    if (($NewAlias ne $Customer->alias) and $Shop->folder('Customers')->existsChild($NewAlias) ) {
        $Servlet->form->addFormError({
            Reason => 'ObjectExists',
            Form   => 'CustomerForm',
            Name   => 'Alias',
        });
    }

    $Servlet->form->executeFormError;

    # set the new next customer number if we try to change the alias to the next customer number
    if($NewAlias eq GetNextNo($Shop, 'Customer', undef, 0) ){
        $NewAlias = GetNextNo($Shop, 'Customer');
    }

    $Customer->set({'Alias' => $NewAlias});

}

sub _prepareAddressValues {
    my $self = shift;
    my ($AddressType, $FormValues) = @_;

    my %AddressValues;
    foreach my $Key ( grep { /^$AddressType/ } keys %$FormValues) {
        my ($AttributeName) = $Key =~ /^$AddressType\.(.*)/;
        $AddressValues{$AttributeName} = $FormValues->{$Key};
    }

    return %AddressValues;
}

sub SaveUserBillingAddress {
    my $self = shift;
    my ($Servlet) = @_;

    my $Customer = $Servlet->object;
    my $SelectedUserID = $Servlet->form->value('SelectedUserID');
    my $User = $Customer->getSite()->loadObject($SelectedUserID, 'User');

    Error('NOT_A_VALID_USER', {
        'CustomerPath'  => $Customer->get('Path'),
        'UserPath'      => $User->get('Path')}) unless ExistsCustomerUser($User->id, $Customer->id);

    my $FormValues = $Servlet->form->form($Customer, 'SaveBillingAddress');

    eval {
        # update user billing address
        my $BillingAddress = $User->get('BillingAddress');
        if (defined $BillingAddress) {
            my %Address = $self->_prepareAddressValues('BillingAddress',$FormValues);
            $BillingAddress->set(\%Address);
        }
    };
    if (ExistsError()) {
        my $Error = GetError();
        $Servlet->params('ObjectID', $Customer->id);
        Error($Error);
    }

    return;
}

sub SaveUserShippingAddress {
    my $self = shift;
    my ($Servlet) = @_;

    my $Customer = $Servlet->object;
    my $SelectedUserID = $Servlet->form->value('SelectedUserID');
    my $User = $Customer->getSite()->loadObject($SelectedUserID, 'User');

    Error('NOT_A_VALID_USER', {
        'CustomerPath'  => $Customer->get('Path'),
        'UserPath'      => $User->get('Path')}) unless ExistsCustomerUser($User->id, $Customer->id);

    my $FormValues = $Servlet->form->form($Customer, 'SaveShippingAddress');
    eval {
        # update user shipping address
        my $ShippingAddress = $User->get('ShippingAddress');
        if (defined $ShippingAddress) {
            my %Address = $self->_prepareAddressValues('ShippingAddress',$FormValues);
            $ShippingAddress->set(\%Address);
        }
    };
    if (ExistsError()) {
        my $Error = GetError();
        $Servlet->params('ObjectID', $Customer->id);
        Error($Error);
    }

    return;
}

# update existing user
sub SaveUserLogin {
    my $self = shift;
    my ($Servlet) = @_;

    my $Customer = $Servlet->object;
    my $Form = $Servlet->form->form($Customer->getSite(), 'SaveUser');

    # update user data
    my $SelectedUserID = $Servlet->form->value('SelectedUserID');
    my $User = $Customer->getSite()->loadObject($SelectedUserID, 'User');

    Error('NOT_A_VALID_USER', {
        'CustomerPath'  => $Customer->get('Path'),
        'UserPath'      => $User->get('Path')}) unless ExistsCustomerUser($User->id, $Customer->id);

    my $UserIsMerchant = $User->get('IsMerchant');

    # for merchants we do not allow changes on the email, password, and activation
    if( $UserIsMerchant ) {
        #dont allow to deactivate the merchant
        if( not $Form->{'UserIsActivated'} ) {
            $Servlet->form->executeFormError({
                'Reason'=> 'CantDeactivateMerchant',
                'Name'  => 'UserIsActivated',
                'Form'  => 'SaveUser',
            });
        }
        my $OldPassword = $User->get('Password');
        # test if user is merchant and attempt to change password
        $Servlet->form->executeFormError({
            'Reason'=> 'CantChangeMerchantPassword',
            'Name'  => 'UserPassword',
            'Form'  => 'SaveUser',
        }) if $OldPassword ne $Form->{'UserPassword'};

        #test if Merchant try to change the email address
        if ($User->get('EMail') ne $Form->{'EMail'}){
            $Servlet->form->executeFormError({
                'Reason'=> 'CantRenameMerchant',
                'Name'  => 'UserAlias',
                'Form'  => 'SaveUser',
            });
        }
    }

    my $ExistingUserEMail = $User->get('EMail');
    if (defined $ExistingUserEMail and $ExistingUserEMail ne $Form->{'UserEMail'}){
        # user will change email, search if the mail address is already used by a user alias or email
        my $aUsers = GetUsersByEMail($Form->{'UserEMail'},$User->getSite);
        if( scalar(@$aUsers) )  {
            $Servlet->vars('LOGIN_EXISTS',1);
            $Servlet->form->executeFormError({
                'Reason'=> 'LOGIN_EXISTS',
                'Name'  => 'UserEMail',
                'Form'  => 'SaveUser',
            });
        }
    }

    # test password and confirmation
    if ((not defined $Form->{'UserPasswordConf'} and
        not defined $Form->{'UserPassword'}) or
        (defined $Form->{'UserPasswordConf'} and
        defined $Form->{'UserPassword'} and
                $Form->{'UserPasswordConf'} eq $Form->{'UserPassword'})){
        # no password has entered or password and password confirmation are same
        delete $Form->{'UserPasswordConf'};
    } else {
        $Servlet->form->executeFormError({
            'Reason'=> 'PASSWORD_NOT_SAME',
            'Name'  => 'UserPasswordConf',
            'Form'  => 'SaveUser',
        });

    }

    # map form attributes
    my %Attributes;
    $Attributes{$_} = $Form->{'User' . $_} foreach map { s/^User//; $_; } grep { /^User/ } keys %$Form;

    my $NewPassword;

    # crypt password
    if ($Attributes{'Password'} and $Attributes{'Password'} =~ /\S/) {
        $NewPassword = $Attributes{'Password'} ;

        $Attributes{'Password'} = eval { CryptPassword($NewPassword); };
        if (ExistsError()) {
            my $Error = GetError();
            Error($Error) unless $Error->code eq 'NativeDie' and $Error->shortMessage eq 'Wide character in crypt';
            $Servlet->form->executeFormError({
                'Reason'=> 'PASSWORD_WRONG',
                'Form'  => 'SaveUser',
                'Name'  => 'UserPassword',
            }) ;
        }
        TriggerHook(
            'UI_BeforeSaveCustomerPassword',
            {
                'Servlet'           => $Servlet,
                'User'              => $User,
                'Password'          => $NewPassword,
                'FormName'          => 'SaveUser',
                'FormField'         => 'UserPassword',
            }
        );
        $Servlet->form->executeFormError;

    } else {
        delete $Attributes{'Password'};
    }

    # set all attributes
    $User->set(\%Attributes);

    if ( defined $NewPassword ) {
        TriggerHook(
            'UI_AfterSaveCustomerPassword',
            {
                'Servlet'           => $Servlet,
                'User'              => $User,
                'Password'          => $NewPassword
            }
        );
    }

    return;
}

sub SendUserByMerchantWelcomeMail ($$$) {
    my $self = shift;
    my ($Servlet, $Customer, $User) = @_;

    #send order confirmation mail to customer
    my $Shop = $Customer->getSite();
    my $MailType = GetShopMailType( $Shop, 'UserByMerchantWelcome', 1 );

    return unless defined $MailType;

    #Create Header
    my %MailParams = (
        'TO'          => $User->get('EMail'),
        'HTMLAllowed' => $User->get('IsHtmlEMailAllowed'),
    );

    my %Vars = (
        'Customer'      => $Customer->tleHash,
        'Shop'          => $Shop->tleHash,
        'URL'           => GetUserNewPasswordUrl($Shop, $User),
        'User'          => $User->tleHash,
    );

    my $LocaleID = defined $User->get('LocaleID') ? $User->get('LocaleID') : $Servlet->params('Locale');
    my $Locale = $Shop->siteLocale($LocaleID);
    my $hLanguage = $Locale->get('Language');

    my %RegionalParams = (
        Locale      => $LocaleID,
        Language    => $hLanguage->{'Code2'},
    );

    eval {$MailType->send( $Shop, \%MailParams, \%RegionalParams, \%Vars );};

    if (ExistsError()) {
        my $Error = GetError();
        LogError($Error);
        $Servlet->params('SEND_MAIL_FAILED',1);
    }

    return;
}

sub NewUserLogin {
    my $self = shift;
    my ($Servlet) = @_;

    my $Customer = $Servlet->object;

    # load form and test mandatory fields like UserEMail
    my $hForm = $Servlet->form->form($Customer->getSite(), 'NewUserLogin');

    # insert new user with email as login (set alias and email to email address)
    my $UserEMail = $hForm->{'UserEMail'};
    die 'no user email' unless (defined $UserEMail and $UserEMail =~ /\S/);

    my $Shop = $Customer->getSite();
    $Servlet->vars('ObjectID', $Shop->id);
    my $User;
    eval {
        my $Form = $Servlet->form;
        #skip form errors, to show all form errors together
        my $hUser = $Form->form($Shop, 'SaveUser', 1);

        # test for existing aliases or email addresses with the given mail
        my $aUsers = GetUsersByEMail($UserEMail,$Shop);
        if( scalar(@$aUsers) ) {
            $Servlet->vars('LOGIN_EXISTS',1);
            $Servlet->form->executeFormError({
                'Reason'=> 'LOGIN_EXISTS',
                'Name'  => 'UserEMail',
                'Form'  => 'SaveRegistration',
            });
        }

        my %AttributeValues;
        $AttributeValues{'User'}     = $self->attributeValues($Servlet, $Shop, LoadClassByAlias('User'));
        $AttributeValues{'Address'}  = $self->attributeValues($Servlet, $Shop, LoadClassByAlias('Address'));
        $AttributeValues{'Customer'} = $self->attributeValues($Servlet, $Shop, $Shop->get('CustomerClass'));

        my $hAddressValues = $AttributeValues{'Address'}{'Values'};
        if (not defined $hAddressValues->{'DisplayName'} ) {
            my $FirstName = $hAddressValues->{'FirstName'};
            my $LastName  = $hAddressValues->{'LastName'};
            if (defined $FirstName or defined $LastName ) {
                $hAddressValues->{'DisplayName'} = $FirstName if defined $FirstName;
                $hAddressValues->{'DisplayName'} .= ' '.  $LastName if defined $LastName;
            }
        }

        my $hUserData;
        foreach my $FormKey (keys %$hUser) {
            if (defined $hUser->{$FormKey}) {
                next unless defined $hUser->{$FormKey};
                $FormKey =~ s/^User//;
                $hUserData->{$FormKey} = $hUser->{'User'.$FormKey};
            }
        }

        # create the User object
        $User = InsertUserForCustomer( $Shop, $Customer, $hUserData, $AttributeValues{'Address'}{'Values'} );

        # send email to new user
        my $SendEmailToUser = $Servlet->form->value('SendEmailToUser');

        if ($SendEmailToUser and $UserEMail ne ''){
            $self->SendUserByMerchantWelcomeMail($Servlet, $Customer, $User);
            $Servlet->params('USER_CREATED_SEND_MAIL',1);
        }
        else{
            $Servlet->params('USER_CREATED_SEND_MAIL_FAILED',1);
        }
    };
    my $Error = ExistsError() ? GetError() : undef;
    $Servlet->vars('ObjectID', $Customer->id);
    Error($Error) if defined $Error;
    my $aOrders = $Customer->get('Orders');
    $_->set({'User' => $User}) foreach @$aOrders;
}

sub SaveComparedAddresses {
    my $self = shift;
    my ($Servlet) = @_;

    my $Customer = $Servlet->object;
    my $FormValues = $Servlet->form->form($Customer, 'SaveCompareAddresses');

    my $SelectedUserID = $Servlet->form->value('SelectedUserID');
    my $User = $Customer->getSite()->loadObject($SelectedUserID, 'User');

    Error('NOT_A_VALID_USER', {
        'CustomerPath'  => $Customer->get('Path'),
        'UserPath'      => $User->get('Path')}) unless ExistsCustomerUser($User->id, $Customer->id);

    my $aRegistrationSet = $Customer->getSite()->get('RegistrationSet');
    my %RegistrationSet;
    $RegistrationSet{$_->{'Attribute'}->alias} = $_ foreach @$aRegistrationSet;
    foreach my $AddressType (qw (Customer User)) {
        my %Address;

        # if country is united states, state is mandatory
        my $USCountry = 0;
        if ($FormValues->{$AddressType.'.CountryID'}) {
            $USCountry = 1 if GetInfoCountry($FormValues->{$AddressType.'.CountryID'})->{'Code2'} eq 'US';
        }

        foreach my $Key ( grep { /^$AddressType/ } keys %$FormValues) {
            my ($AttributeName) = $Key =~ /^$AddressType\.(.*)/;
            $Address{$AttributeName} = $FormValues->{$Key} if $RegistrationSet{$AttributeName}{'IsVisible'} || ($Key eq $AddressType.'.State' && $USCountry);
        }
        my $BillingAddress = $AddressType eq 'Customer' ? $Customer->get('BillingAddress') : $User->get('BillingAddress');
        $BillingAddress->set(\%Address);
    }

    return;
}

sub DeleteMandateID {
    my $self = shift;
    my ($Servlet) = @_;

    my $Customer = $Servlet->object;

    # delete existing mandat reference number
    $Customer->set({'MandateID' => undef});

}

sub UnityViewGeneral {
    my $self = shift;
    my ($Servlet) = @_;

    my $Customer = $Servlet->object;

    # here we made the decision which object is used for fetching addresses
    # on the customer general page:
    # 1. user and user address for registered customers
    # 2. order and order address for unregistered customers
    # 3. or customer and customer address if we have no user or orders for the customer

    my $ObjectForAddress;
    # 1. a registered customer has at least one user, we use this user object
    $ObjectForAddress = $Customer->get('Users')->[0] if scalar(@{$Customer->get('Users')});

    # 2. a unregistered customer has not an user
    # we use the address from the order or from the customer if we have no order
    $ObjectForAddress = $Customer->get('Orders')->[0] if (not defined $ObjectForAddress) and scalar(@{$Customer->get('Orders')});

    # 3. we use the customer object if we have no order or user informations
    $ObjectForAddress = $Customer unless defined $ObjectForAddress;

    $Servlet->params('ObjectForAddress', $ObjectForAddress->tleHash);


    # we allow editing unregistered customers until an order exists
    if ($ObjectForAddress->instanceOf('Customer') and not scalar(@{$ObjectForAddress->get('Orders')}) ) {
        $Servlet->params('UnRegisteredEditable',1);
    }else{
        $Servlet->params('UnRegisteredEditable',0);
    }

    # get the shipping address from the object if we have one
    my $ShippingAddressToUse = $ObjectForAddress->instanceOf('Customer') ? undef : $ObjectForAddress->get('ShippingAddress');

    # display ShippingAddress also for merchant if this address is not set as shipping address
    $Servlet->params('RegisteredUserShowShippingAddress',1);
    if ($ObjectForAddress->instanceOf('User') and not defined $ObjectForAddress->get('ShippingAddress') ){

        # use the Adress from the addressfolder which is not the billingaddress
        my $AddressFolder = $ObjectForAddress->folder('Addresses');
        my $BillingAddressID = $ObjectForAddress->get('BillingAddress')->id;
        my $ShippingAddress = undef;

        # look for an address wich is not the billing address
        foreach my $Address (@{$AddressFolder->children()}){
            next if $BillingAddressID eq $Address->id;
            $ShippingAddress = $Address;
            last;
        }

        # we can use a shipping address from the folder but non is saved on the object
        $Servlet->params('RegisteredUserShowShippingAddress',0);
        # we use the shipping address found in the folder
        $ShippingAddressToUse = $ShippingAddress if defined $ShippingAddress;

    }

    $Servlet->params('UsableShippingAddress', defined $ShippingAddressToUse ? $ShippingAddressToUse->tleHash : undef);

    # add form data for the maxlength validation
    my $hBillingFormData = $Servlet->form->formData($Customer->class, 'UnitySaveBillingAddress');
    my $hShippingFormData = $Servlet->form->formData($Customer->class, 'UnitySaveShippingAddress');
    my %DotFormFields = (%{$hBillingFormData->{'Fields'}}, %{$hShippingFormData->{'Fields'}});

    my $FormFields;
    foreach my $FormField (keys %DotFormFields) {
        my $NewFormField = $FormField;
        $NewFormField =~ s/\./\_/;
        $FormFields->{$NewFormField} = $DotFormFields{$FormField};
    }

    $Servlet->vars('FormFields', $FormFields);

    # check if we have both zipcode and city visible
    my $BillingAddressSetVisibleFields = $Customer->getSite()->get('BillingAddressSetVisibleFields');
    my @BillingAddressSetVisibleFieldsArray = split /,/, $BillingAddressSetVisibleFields;
    my $ZipCodeAndCity = 'Zipcode' ~~ @BillingAddressSetVisibleFieldsArray && 'City' ~~ @BillingAddressSetVisibleFieldsArray;
    $Servlet->vars('ZipCodeAndCity', $ZipCodeAndCity);

        # trigger hook to allow the adding of context sensitive apps
    TriggerHook('UI_Customer_UnityViewGeneral', { 'Servlet' => $Servlet });

    return;
}

1;
