#========================================================================================
# package      DE_EPAGES::Klarna::API::Cartridge
# base         DE_EPAGES::Payment::API::PaymentInstaller
# state        public
#----------------------------------------------------------------------------------------
# description  This is the main cartridge class for install/patch/uninstall.
#========================================================================================
package DE_EPAGES::Klarna::API::Cartridge;
use base DE_EPAGES::Payment::API::PaymentInstaller;

use strict;
use DE_EPAGES::Object::API::Factory      qw ( LoadClassByAlias ExistsClassByAlias );
use DE_EPAGES::Database::API::Connection qw ( Transaction );
use DE_EPAGES::Object::API::XML::ObjectDeleteHandler qw (DeleteObjects);
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw (ImportObjects);
use DE_EPAGES::Core::API::Error qw( HandleError );

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer = DE_EPAGES::Klarna::API::Cartridge->new(%options);
#----------------------------------------------------------------------------------------
# description  Creates a new installer object of DE_EPAGES::Klarna.
#----------------------------------------------------------------------------------------
# input        %options | options for un/install
#               <ul>
#                 <li>IsRecursive - recursive un/installation of required or dependent
#                     cartridges
#               </ul>
#               | hash
# return       $Installer | cartridge installer | object
#========================================================================================
sub new {
    my ($class, %options) = @_;

    my $self = __PACKAGE__->SUPER::new(
        %options,
        'CartridgeDirectory' => 'DE_EPAGES/Klarna',
        'Version'            => '7.37.0',
        'Patches'            => ['6.15.1','6.15.2','6.15.3','6.15.4','6.16.0','6.16.1','6.16.2','6.16.3','6.17.0','6.17.1','6.17.2','6.17.3','6.17.4','6.17.5','6.17.6','6.17.7','6.17.8','6.17.9','6.17.9.1','6.17.9.2','6.17.9.3','6.17.10','6.17.11','6.17.12','6.17.12.1','6.17.12.2','6.17.13','6.17.13.1','6.17.13.2','6.17.14','6.17.14.1','6.17.15','6.17.15.1','6.17.16','6.17.17','6.17.17.1','6.17.18','6.17.19','6.17.20','6.17.20.1','6.17.21','6.17.21.1','6.17.22','6.17.23','6.17.24','6.17.24.1','6.17.25','6.17.26','6.17.26.1','6.17.27','6.17.27.1','6.17.28','6.17.28.1','6.17.29','6.17.29.1','6.17.30','6.17.31','6.17.32','6.17.33','6.17.33.1','6.17.34','6.17.34.1','6.17.35','6.17.36','6.17.36.1','6.17.37','6.17.38','6.17.38.1','6.17.39','6.17.39.1','6.17.40','6.17.40.1','6.17.41','6.17.41.1','6.17.42','6.17.43','6.17.43.1','6.17.44','6.17.45','6.17.45.1','6.17.46','6.17.47','6.17.47.1','6.17.48','6.17.48.1','6.17.49','6.17.49.1','6.17.50','6.17.51','6.17.51.1','6.17.52','6.17.53','7.0.0','7.1.0','7.2.0','7.3.0','7.4.0','7.5.0','7.6.0','7.7.0','7.8.0','7.9.0','7.10.0','7.11.0','7.12.0','7.13.0','7.14.0','7.15.0','7.16.0','7.17.0','7.18.0','7.19.0','7.20.0','7.21.0','7.22.0','7.23.0','7.24.0','7.25.0','7.26.0','7.27.0','7.28.0','7.29.0','7.30.0','7.31.0','7.32.0','7.33.0','7.34.0','7.35.0','7.36.0'],

    );
    return bless $self, $class;
}

#========================================================================================
# function     install
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->install;
#----------------------------------------------------------------------------------------
# description  Installs the cartridge.
#========================================================================================
sub install {
    my $self = shift;

    $self->startInstall;
    $self->SUPER::install;
    ImportObjects($self->xmlDirectory .'/MailTypes.xml');
    $self->finishInstall;
    return;
}

#========================================================================================
# function     uninstall
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->uninstall;
#----------------------------------------------------------------------------------------
# description  Uninstalls the cartridge.
#========================================================================================
sub uninstall {
    my $self = shift;

    $self->startUninstall;
    DeleteObjects($self->xmlDirectory . '/MailTypes.xml');
    $self->SUPER::uninstall;
    $self->finishUninstall;
    return;
}

sub patch_6_15_1 {
    my $self = shift;

    #calling data changes from patch.xml file
    $self->importPatchDir('Patch_6_15_1', 'Patch');

    # delete leftover PaymentMethodKlarna objects if present
    if(ExistsClassByAlias('PaymentMethodKlarna')) {
        $self->logInfo('deleting remains of old PaymentMethodKlarna (EPG-27611)');
        LoadClassByAlias('PaymentMethodKlarna')->delete;
    }

    return '6.15.2';
}

#========================================================================================
# function     patchDBStructure_6_15_1
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->patchDBStructure_6_15_1;
#----------------------------------------------------------------------------------------
# description  Patches cartridge db structure in active store (database connection)
#               from version '6.15.1' to version '6.15.2'.
#----------------------------------------------------------------------------------------
# return       $Version | version of cartridge after patch function, '6.15.2' | string
#========================================================================================
sub patchDBStructure_6_15_1 {
    my $self = shift;

    my $dbi = $self->getCurrentDBHandle();

    if($dbi->driverName() eq 'mysql'){
        $self->_transformPatchDebris_6_15_1();

        #changes for file /Database/mysql/Tables/klarnapclass.sql
        $self->logInfo('change type of column klarnapclass.startfee');
        $dbi->modifyTableColumn('klarnapclass', 'startfee decimal(19,4) not null');

        $self->logInfo('change type of column klarnapclass.minpurchase');
        $dbi->modifyTableColumn('klarnapclass', 'minpurchase decimal(19,4) not null');

        $self->logInfo('change type of column klarnapclass.interestrate');
        $dbi->modifyTableColumn('klarnapclass', 'interestrate double not null');

        $self->logInfo('change type of column klarnapclass.currencyid');
        $dbi->modifyTableColumn('klarnapclass', 'currencyid char(3) not null');

        $self->logInfo('change type of column klarnapclass.description');
        $dbi->modifyTableColumn('klarnapclass', 'description varchar(500) not null');

        $self->logInfo('change type of column klarnapclass.invoicefee');
        $dbi->modifyTableColumn('klarnapclass', 'invoicefee decimal(19,4) not null');

        $self->logInfo('change type of column klarnapclass.months');
        $dbi->modifyTableColumn('klarnapclass', 'months integer not null');

        unless ($dbi->existsTableColumn('klarnapclass', 'countryid')) {
            $self->logInfo('add column klarnapclass.countryid after column klarnapclass.isaccount');
            $dbi->addTableColumnAfterColumn('klarnapclass', 'isaccount', 'countryid integer not null default 0');
            $dbi->modifyTableColumnDropDefault('klarnapclass','countryid');
        }
        unless ($dbi->existsTableColumn('klarnapclass', 'paymentmethodid')) {
            $self->logInfo('add column klarnapclass.paymentmethodid after column klarnapclass.countryid');
            $dbi->addTableColumnAfterColumn('klarnapclass', 'countryid', 'paymentmethodid integer not null default 0');
            $dbi->modifyTableColumnDropDefault('klarnapclass','paymentmethodid');
        }
        $self->logInfo('create index klarnapclass.i_klarna_pclass_country');
        $dbi->createIndex('klarnapclass', 'i_klarna_pclass_country', undef, ['countryid','paymentmethodid'])
            unless ($dbi->existsIndex('klarnapclass','i_klarna_pclass_country'));

        $self->logInfo('migrating Klarna Pclasses to 6.15.2');
        $self->_migratePclasses_6_15_1();

        if (not $dbi->existsForeignKey('fk_klarnapclass_countryid')) {
            $self->logInfo('add foreign key klarnapclass.fk_klarnapclass_countryid');
            $dbi->addForeignKey(
                'klarnapclass',
                'fk_klarnapclass_countryid',
                [
                    'countryid',
                ],
                'country',
                [
                    'countryid',
                ]
            );
        }
        if (not $dbi->existsForeignKey('fk_klarnapclass_paymentmethodid')) {
            $self->logInfo('add foreign key klarnapclass.fk_klarnapclass_paymentmethodid');
            $dbi->addForeignKey(
                'klarnapclass',
                'fk_klarnapclass_paymentmethodid',
                [
                    'paymentmethodid',
                ],
                'paymentmethod',
                [
                    'paymentmethodid',
                ]
            );
        }

    }

    return '6.15.2';
}

# called from patchDBStructure_6_15_1
# check for remains of PaymentMethodKlarna in DB (EPG-27611) and, if found, sets the package
# name to the payment base package name
sub _transformPatchDebris_6_15_1 {
    my $self = shift;
    $self->logInfo('checking for remains of old PaymentMethodKlarna objects (EPG-27611)');

    my $dbi = $self->getCurrentDBHandle();
    eval {
        my $ClassID = $self->getClassID('PaymentMethodKlarna');
        # prevent error "Can't compile package 'DE_EPAGES::Klarna::API::Object::PaymentMethodKlarna'"
        $dbi->do(<<'END_SQL', [$ClassID]);
            UPDATE class
            SET package = 'DE_EPAGES::Payment::API::Object::PaymentMethod'
            WHERE classid = ?
END_SQL
    };

    HandleError(OnError => sub {
        my ($Error) = @_;
        return 'FIXED' if($Error->code eq 'NoClassName');
    });

    return;
}

# called from patchDBStructure_6_15_1
sub _migratePclasses_6_15_1 {
    my $self = shift;

    my $dbi = $self->getCurrentDBHandle();

    my $NeedsMigration = $dbi->execute(<<'END_SQL', [0, 0])->[0]->[0] > 0;
        SELECT COUNT(*)
        FROM klarnapclass
        WHERE (countryid = ? OR paymentmethodid = ?)
END_SQL
    unless($NeedsMigration) {
        $self->logInfo('Pclasses already migrated. Skipping.');
        return;
    }

    $self->logInfo('fetching Shops with Paymentmethod Klarna');
    my $aPaymentMethods = $dbi->execute(<<'END_SQL', ['KlarnaFinancing', 'KlarnaHirePurchase']);
        SELECT shop.shopid, paymentmethodid
        FROM paymentmethod
        JOIN shop ON shop.shopid = paymentmethod.shopid
        JOIN paymenttype ON paymenttype.paymenttypeid = paymentmethod.paymenttypeid
        JOIN object ON object.objectid = paymentmethod.paymenttypeid
        WHERE (alias = ? OR alias = ?)
END_SQL

    foreach my $ResultRow (@$aPaymentMethods) {
        Transaction ( sub {
            my $ShopID = $ResultRow->[0];
            my $PaymentMethodID = $ResultRow->[1];
            $self->logInfo("determining pclasses for PaymentMethodID $PaymentMethodID");
            my $Pclasses = $dbi->execute(<<'END_SQL', [$PaymentMethodID, 'Kr_Pclasses']);
                SELECT objectid from object
                WHERE parentid IN (
                    SELECT objectid
                    FROM object
                    WHERE parentid = ? AND alias = ?
                )
END_SQL
            foreach my $Pclass (@$Pclasses) {
                my $PclassID = $Pclass->[0];
                $self->logInfo("migrating Pclass $PclassID for PaymentMethodID $PaymentMethodID of Shop $ShopID");
                $self->_migratePclass_6_15_1($ShopID, $PaymentMethodID, $PclassID);
            }
        });
    }
    return;

}

# called from _migratePclasses_6_15_1
sub _migratePclass_6_15_1 {
    my $self = shift;
    my $dbi = $self->getCurrentDBHandle();
    my ($ShopID, $PaymentMethodID, $PclassID) = @_;

    # no migration needed if countryid and paymentmethodid are already set
    my $NeedsMigration = $dbi->execute(<<'END_SQL', [$PclassID, 0, 0])->[0]->[0] > 0;
        SELECT COUNT(*)
        FROM klarnapclass
        WHERE klarnapclassid = ? AND (countryid = ? OR paymentmethodid = ?)
END_SQL

    unless($NeedsMigration) {
        $self->logInfo("Pclass $PclassID already migrated. Skipping.");
        return;
    }

    $self->logInfo("determining currency of klarna pclass $PclassID");
    my $PclassCurrency = $dbi->execute(<<'END_SQL', [$PclassID]);
        SELECT currencyid
        FROM klarnapclass
        WHERE klarnapclassid = ?
END_SQL
    my $CurrencyID = $PclassCurrency->[0]->[0];

    my $CurrencyToCountry = { 'DKK' => 'DK', 'NOK' => 'NO', 'SEK' => 'SE' };
    if($CurrencyID ~~ $CurrencyToCountry) {
        $self->logInfo("determining country for klarna pclass $PclassID");
        my $CountryIDResult = $dbi->execute(<<'END_SQL', [$CurrencyToCountry->{$CurrencyID}]);
            SELECT countryid
            FROM country
            WHERE code2 = ?
END_SQL
        my $CountryID = $CountryIDResult->[0]->[0];

        $self->logInfo("set country and paymentmethod of klarna pclass $PclassID to $CountryID, $PaymentMethodID");
        $dbi->do(<<'END_SQL', [$CountryID, $PaymentMethodID, $PclassID]);
            UPDATE klarnapclass
            SET countryid = ?, paymentmethodid = ?
            WHERE klarnapclassID = ?
END_SQL
    }
    else {
        # if currency is ambiguous, use the shop's country
        $self->logInfo("determining shop country for klarna pclass $PclassID");
        my $ShopCountryResult = $dbi->execute(<<'END_SQL', [$ShopID]);
            SELECT countryid
            FROM address
            JOIN shop ON shop.addressid = address.addressid
            WHERE shop.shopid = ?
END_SQL
        my $ShopCountryID = $ShopCountryResult->[0]->[0];

        $self->logInfo("set country and paymentmethod of klarna pclass $PclassID");
        $dbi->do(<<'END_SQL', [$ShopCountryID, $PaymentMethodID, $PclassID]);
            UPDATE klarnapclass
            SET countryid = ?, paymentmethodid = ?
            WHERE klarnapclassID = ?
END_SQL
    }
    return;
}

#========================================================================================
# function     patchDBStructure_6_15_2
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->patchDBStructure_6_15_2;
#----------------------------------------------------------------------------------------
# description  Patches cartridge db structure in active store (database connection)
#               from version '6.15.2' to version '6.15.3'.
#----------------------------------------------------------------------------------------
# return       $Version | version of cartridge after patch function, '6.15.3' | string
#========================================================================================
sub patchDBStructure_6_15_2 {
    my $self = shift;

    my $dbi = $self->getCurrentDBHandle();

    if($dbi->driverName() eq 'mysql'){

        #changes for file /Database/mysql/Tables/klarnapclass.sql
        $self->logInfo('create index klarnapclass.i_klarna_pclass_paymentmethod');
        $dbi->createIndex('klarnapclass', 'i_klarna_pclass_paymentmethod', undef, ['paymentmethodid'])
            unless ($dbi->existsIndex('klarnapclass','i_klarna_pclass_paymentmethod'));

    }

    return '6.15.3';
}

#========================================================================================
# function     encryptData
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->encryptData;
#----------------------------------------------------------------------------------------
# description  Encrpyts data of this cartridge.
#========================================================================================
sub encryptData {
    my $self = shift;

    $self->encryptDataByClass('PaymentMethodKlarnaFinancing');
    $self->encryptDataByClass('PaymentMethodKlarnaHirePurchase');
    $self->encryptDataByClass('PaymentMethodKlarnaInvoice');
}

#========================================================================================
# function     patchDBStructure_6_17_5
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->patchDBStructure_6_17_5;
#----------------------------------------------------------------------------------------
# description  Patches cartridge db structure in active store (database connection)
#               from version '6.17.5' to version '6.17.6'.
#----------------------------------------------------------------------------------------
# return       $Version | version of cartridge after patch function, '6.17.6' | string
#========================================================================================
sub patchDBStructure_6_17_5 {
    my $self = shift;
    #dont use API function, to avoid inconistencies between perl modules and database content

    my $dbi = $self->getCurrentDBHandle();

    if($dbi->driverName() eq 'mysql'){

        #changes for file /Database/mysql/Tables/lineitempaymentklarna.sql
        unless ($dbi->existsTableColumn('lineitempaymentklarna', 'isreservationoutofdate')) {
            $self->logInfo('add column lineitempaymentklarna.isreservationoutofdate');
            $dbi->addTableColumnAfterColumn('lineitempaymentklarna', 'kr_cellphone', 'isreservationoutofdate bool null');
        }
        unless ($dbi->existsTableColumn('lineitempaymentklarna', 'reservationnumber')) {
            $self->logInfo('add column lineitempaymentklarna.reservationnumber');
            $dbi->addTableColumnAfterColumn('lineitempaymentklarna', 'isreservationoutofdate', 'reservationnumber varchar(255) null');
        }
        unless ($dbi->existsTableColumn('lineitempaymentklarna', 'protocol')) {
            $self->logInfo('add column lineitempaymentklarna.protocol');
            $dbi->addTableColumnAfterColumn('lineitempaymentklarna', 'reservationnumber', 'protocol varchar(255) null');
        }
    }

    return '6.17.6';
}

1;
