#========================================================================================
# package      DE_EPAGES::Klarna::API::Object::PaymentMethodKlarnaInvoice
# base         DE_EPAGES::Klarna::API::Object::PaymentMethodKlarnaBase
# state        public
# description  object interface for PaymentMethodKlarnaInvoice
#========================================================================================
package DE_EPAGES::Klarna::API::Object::PaymentMethodKlarnaInvoice;
use base DE_EPAGES::Klarna::API::Object::PaymentMethodKlarnaBase;

use strict;
use DE_EPAGES::Klarna::API::Constants qw (KLARNA_INVOICE KLARNA_ACCOUNT @KLARNA_AVAILABLE_COUNTRIES %KLARNA_COUNTRIES_CURRENCIES);
use DE_EPAGES::Core::API::PerlTools qw (isValid);
use DE_EPAGES::Core::API::Log qw (LogDebug);
use DE_EPAGES::Klarna::API::KlarnaCalc qw ( GetEffectiveInvoiceFee );

use constant DEFAULT_CURRENCY    => 'EUR';
use constant DEFAULT_INVOICE_FEE => 1.95;


#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Object = $Package->insert($hInfo);
# example      my $Class = LoadClassByAlias('PaymentMethodPaynet');
#               my $Object = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Overrides and calls PaymentMethodKlarnaBase->insert to add a Template Path
#----------------------------------------------------------------------------------------
# input        $Package | class package | string
# input        $hInfo | attributes of new object
# return       $Object | new object | object
#========================================================================================
sub insert {
    my ($Class, $hInfo) = @_;

    my $InitialNoteTemplatePath = '/DE_EPAGES/Klarna/Templates/InitialNoteContent.html';
    my $self = $Class->SUPER::insert($hInfo, $InitialNoteTemplatePath);
    return $self;
}

#========================================================================================
# function     canAddToBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       if ($PaymentMethod->canAddToBasket($hVars));
#----------------------------------------------------------------------------------------
# description  Tests if a payment method can be added to the basket with the given currency.
#               Overrides function <function DE_EPAGES::Order::API::Object::ContainerPaymentMethod::canAddToBasket>
#               The payment method can be added to the container if the CurrencyID
#               attribute matches the currency of the container.
#----------------------------------------------------------------------------------------
# input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price    - int
#               </ul> | ref.hash
# return       $canAddToBasket | true if payment method can be used for the given currency | boolean
#========================================================================================
sub canAddToBasket {
    my $self = shift;
    my ($hVars) = @_;

    return 0 unless $self->checkGeneral($hVars);

    # at least invoice is available
    return 1;
}

#========================================================================================
# function     checkGeneral
# state        private
#----------------------------------------------------------------------------------------
# syntax       if ($PaymentMethod->checkGeneral($hVars));
#----------------------------------------------------------------------------------------
# description  Tests if a payment method can be added to the basket with the given currency.
#               Overrides function <function DE_EPAGES::Order::API::Object::ContainerPaymentMethod::canAddToBasket>
#               The payment method can be added to the container if the CurrencyID
#               attribute matches the currency of the container.
#----------------------------------------------------------------------------------------
# input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price    - int
#               </ul> | ref.hash
# return       $canAddToBasket | true if payment method can be used for the given currency | boolean
#========================================================================================
sub checkGeneral {
    my $self = shift;
    my ($hVars) = @_;

    if ($hVars->{'CurrencyID'} ne $self->get('Kr_CurrencyID')) {
        LogDebug('Klarna', 'Klarna currency different from Shop currency, payment denied');
        return 0;
    }

    return $self->SUPER::checkGeneral($hVars);
}

#========================================================================================
# function      recalculateLineItem
# state         public
#----------------------------------------------------------------------------------------
# syntax        $Package->recalculateLineItem($LineItem);
#----------------------------------------------------------------------------------------
# description   Recalculates lineitem. Saves default Klarna payment type if
#                it is yet missing.
#----------------------------------------------------------------------------------------
# input         $LineItem | line item | object
# return        $object | created object | object
#========================================================================================
sub recalculateLineItem {
    my $self = shift;
    my ($LineItem) = @_;

    $self->SUPER::recalculateLineItem($LineItem);

    my $InvoiceNo = $LineItem->get('Kr_InvoiceNo');
    if (isValid($InvoiceNo)) {
        return;
    }

    if (!($LineItem->get('InternalMethod'))) {
        my $SelectedMethod = KLARNA_INVOICE;
        $LineItem->set({ 'InternalMethod' => $SelectedMethod });
    }
    if ($LineItem->get('InternalMethod') <= KLARNA_ACCOUNT) { #Invoice or account is selected so campaign has to be removed.
        $LineItem->set({ 'KlarnaPeriodicCost' => undef });
    }
    return;
}

#========================================================================================
# function      price
# state         public
#----------------------------------------------------------------------------------------
# syntax        $self->price( {
#                   'CurrencyID' => $CurrencyID,
#                   'TaxModel' => $TaxModel,
#                   'Container'  => $Container,} );
#----------------------------------------------------------------------------------------
# description   Returns KlarnaInvoiceFee Gross
#               Overrides price method from Base Package (ContainerPaymentMethod)
#----------------------------------------------------------------------------------------
# input         $hValues | attribute name => value hash | ref.hash
# return        $price | InvoiceFee | float
#========================================================================================
sub price {
    my $self = shift;
    my ($hValues) = @_;

    my $PaymentMethod = $self;
    my $Container = $hValues->{'Container'};

    my $PriceAttributeName = $PaymentMethod->get('Kr_SubTotalAttrName') // 'LineItemsSubTotal';

    my $InvoiceFee = GetEffectiveInvoiceFee($PaymentMethod, $Container->get($PriceAttributeName));
    return if(!defined($InvoiceFee) || $InvoiceFee == 0);

    my $hInfo = $Container->get(['CurrencyID', 'TaxMatrix', 'TaxModel', 'TaxArea']);

    # saved price is always net
    my $Price = $hInfo->{'TaxMatrix'}->net2gross(
        $hInfo->{'CurrencyID'},
        $InvoiceFee, {
        'TaxModel' => $hInfo->{'TaxModel'},
        'TaxClass' => $hInfo->{'TaxMatrix'}->defaultTaxClass,
        'TaxArea'  => $hInfo->{'TaxArea'}
    },
        0
    );

    return $Price;
}

#========================================================================================
# function      default
# state         private
#----------------------------------------------------------------------------------------
# syntax        $hValues = DE_EPAGES::Klarna::API::Object::PaymentMethodKlarna->default( $Parent, $hInfo );
# example       $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description   Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#               sets default fee,currency, testmode
#----------------------------------------------------------------------------------------
# input         $Parent | parent object of the new object | object
# input         $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return        $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;

    return $class->SUPER::default($Parent, $hInfo, DEFAULT_INVOICE_FEE, DEFAULT_CURRENCY);
}

1;
