#========================================================================================
# package      DE_EPAGES::MRW::API::Object::ShippingFulfillmentMethodMRW
# base         DE_EPAGES::ExternalShipping::API::Object::ShippingFulfillmentMethod
# state        public
#----------------------------------------------------------------------------------------
# description  Class for shipping fulfillment connection
#========================================================================================
package DE_EPAGES::MRW::API::Object::ShippingFulfillmentMethodMRW;
use base DE_EPAGES::ExternalShipping::API::Object::ShippingFulfillmentMethod;

use strict;

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );

use DE_EPAGES::MRW::API::RemoteMRWBarcode;
use DE_EPAGES::MRW::API::Constants qw ( %MRW_LABELSTATUS );
use DE_EPAGES::Core::API::Log qw (LogDebug);

#========================================================================================
# function     featureName
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FeatureName = $ShippingFulfillmentMethod->featureName;
#----------------------------------------------------------------------------------------
# description   Returns the feature name 'MRW'.
#----------------------------------------------------------------------------------------
# return       $FeatureName | feature name (undef means no feature count) | String
#========================================================================================
sub featureName () { return 'MRW'; }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingFulfillmentMethodMRW = DE_EPAGES::MRW::API::Object::ShippingFulfillmentMethodMRW->insert($hInfo);
# example      my $Class = LoadClassByAlias('ShippingFulfillmentMethodMRW');
#               my $ShippingFulfillmentMethodMRW = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertShippingFulfillmentMethodMRW' is triggered.
#----------------------------------------------------------------------------------------
# input        $Package | class package | string
# input        $hInfo | attributes of new object | ref.hash
# return       $ShippingFulfillmentMethodMRW | new object | object
# hook         OBJ_InsertShippingFulfillmentMethodMRW | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingFulfillmentMethodMRWID - shipping label id - int
#               </ul>
#========================================================================================
sub insert {
    my ($Package, $hInfo) = @_;

    my $self = $Package->SUPER::insert($hInfo);
    TriggerHook('OBJ_InsertShippingFulfillmentMethodMRW', {'Object' => $self, 'ShippingFulfillmentMethodMRWID' => $self->id });
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingFulfillmentMethodMRW->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteShippingFulfillmentMethodMRW' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteShippingFulfillmentMethodMRW | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingFulfillmentMethodMRWID - shipping label id - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;
    TriggerHook('OBJ_DeleteShippingFulfillmentMethodMRW', {'Object' => $self, 'ShippingFulfillmentMethodMRWID' => $self->id });
    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingFulfillmentMethodMRW->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateShippingFulfillmentMethodMRW' and 'OBJ_AfterUpdateShippingFulfillmentMethodMRW'.
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateShippingFulfillmentMethodMRW | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingFulfillmentMethodMRWID - shipping label id - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateShippingFulfillmentMethodMRW | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingFulfillmentMethodMRWID - shipping label id - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;
    TriggerHook('OBJ_BeforeUpdateShippingFulfillmentMethodMRW', {'Object'=>$self, 'Values'=> $hValues, 'ShippingFulfillmentMethodMRWID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateShippingFulfillmentMethodMRW',  {'Object'=>$self, 'Values'=> $hValues, 'ShippingFulfillmentMethodMRWID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}
#========================================================================================
# function     webserviceValidateLabel
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShippingFulfillmentMethod->webserviceValidateLabel( $ShippingLabel );
#----------------------------------------------------------------------------------------
# description  Calls validation webservice for one transaction. 
#               Has to be implemented in child class, if 
#               ShippingFulfillmentType.IntegrationType=Webservice
#----------------------------------------------------------------------------------------
# input        $ShippingLabel | shipping label | object
#========================================================================================
sub webserviceValidateLabel {
    my $self = shift;
    my ($Transaction) = @_;
    Error('NotImplemented');
}

#========================================================================================
# function     webserviceValidateTransaction
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShippingFulfillmentMethod->webserviceValidateTransaction( $Transaction );
#----------------------------------------------------------------------------------------
# description  Calls validation webservice for one transaction. 
#               Has to be implemented in child class, if 
#               ShippingFulfillmentType.IntegrationType=Webservice
#----------------------------------------------------------------------------------------
# input        $Transaction | shipping transaction | object
#========================================================================================
sub webserviceValidateTransaction {
    my $self = shift;
    my ($Transaction) = @_;
    Error('NotImplemented');
}

#========================================================================================
# function     webserviceDoTransaction
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShippingFulfillmentMethod->webserviceDoTransaction( $Transaction );
#----------------------------------------------------------------------------------------
# description  Calls execution webservice for one transaction. 
#               Has to be implemented in child class, if
#               ShippingFulfillmentType.IntegrationType=Webservice
#----------------------------------------------------------------------------------------
# input        $Transaction | shipping transaction | object
#========================================================================================
sub webserviceDoTransaction {
    my $self = shift;
    my ($Transaction,$LanguageID) = @_;

    my $Remote = DE_EPAGES::MRW::API::RemoteMRWBarcode->new();
    my $errCnt = 0;
    my $ErrorsMessages = '';
    # do loop over all labels in transaction
    foreach my $ShippingLabel (@{$Transaction->get('ShippingLabels')}) {

        my $result = $Remote->processShipmentRequest($self,$ShippingLabel,$LanguageID);
        my $ResultStatus = $result->{'requestStatus'};
        my $Errors = $result->{'BarCodeError'};
        my $TrackingNr = $result->{'TrackingNumber'};
        my $ShippingLabelName = $ShippingLabel->get('NameOrAlias');
        LogDebug('MyLabel',$ShippingLabelName);
        if (!$ResultStatus || !$TrackingNr){
                $errCnt++;
                $ErrorsMessages = "$ErrorsMessages | $ShippingLabelName -- $Errors";
                #overwrite state if also a warning occurs
#                $ShippingLabel->set({   'LabelStatus' => $MRW_LABELSTATUS{'ERROR_NO_BARCODE'},
#                                        'BarCodeError' => $result->{'BarCodeError'},
#                });
        }
        else {
            # if there is no error and the trackingNr has been generated, we update the shipping label with $result ()a hash containing trackinnr and shippingLabelLinkMRW, ...)
             $ShippingLabel->set({   'LabelStatus' => $MRW_LABELSTATUS{'BAR_CODE_CREATED'},
                                      'Issued' => 1,
                                      'BarCodeError' => $result->{'BarCodeError'},
                                      'TrackingID' => $result->{'TrackingNumber'},
                                      'ShippingLabelLinkMRW' => $result->{'Label'}});
        }
    }
    my $error='';
    my $TransStatus = 'STATUS_EXPORTED';
    if ( $errCnt == scalar(@{$Transaction->get('ShippingLabels')} ) ) {
        $TransStatus = 'STATUS_EXPORT_FAILED';
        $error='ALL_EXPORTED_FAILED';
    }
    elsif ( $errCnt != 0 ){
        $TransStatus = 'STATUS_EXPORTED_WITH_ERRORS';
        $error='SOME_EXPORTED_FAILED';
    }
    $Transaction->set({'TransStatus' => $TransStatus});
    
    return "$error $ErrorsMessages";
}

#========================================================================================
# function     isUsable
# state        public
#----------------------------------------------------------------------------------------
# syntax       isUsable( );
#----------------------------------------------------------------------------------------
# description  Checks if ShippingFulfillmentMethod is usable
#----------------------------------------------------------------------------------------
# return       $IsUsable | True if ShippingFulfillmentMethod is usable | boolean
#========================================================================================
sub isUsable {
    my $self = shift;

    return  $self->get('WebserviceLocationCode') &&
            $self->get('WebserviceAccountNumber') &&
            $self->get('WebserviceUsername') &&
            $self->get('WebservicePassword') &&
            $self->SUPER::isUsable();
}
1;