#========================================================================================
# package      DE_EPAGES::Order::API::LineItemData
# state        public
#----------------------------------------------------------------------------------------
# description  Gets the data of LineItemContainer and its line items.
#========================================================================================
package DE_EPAGES::Order::API::LineItemsData;

use DE_EPAGES::Shop::API::Resource::ShopRequestContext;
use DE_EPAGES::Shop::API::Resource::Formatter;
use DE_EPAGES::Order::API::LineItemPriceCalculator;

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( $BasketOrOrder );
#----------------------------------------------------------------------------------------
# description  constuctor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Order::API::LineItemData
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

sub getLineItems {
    my $self = shift;
    my ($Container, $Formatter, $LanguageID, $IncludeTaxInLineItem) = @_;

    my $CurrencyID = $Container->get('CurrencyID');
    my $TaxModel = $Container->get('TaxModel');

    my $Context = $self->_createShopRequestContext($Container->getSite);
    return [
        map { $self->_LineItem($_, $CurrencyID, $TaxModel, $Formatter, $Context, $LanguageID, $IncludeTaxInLineItem) }
        @{ $Container->get('LineItems') }
    ];
}

sub _LineItem {
    my $self = shift;
    my ($LineItem, $CurrencyID, $TaxModel, $Formatter, $Context, $LanguageID, $IncludeTaxInLineItem) = @_;

    return unless $LineItem->instanceOf('LineItemProduct');

    my @ItemAttributes = qw(Quantity OrderUnit GUID SKU Name BasePrice LineItemPrice LocalizedVariationString CouponDiscount);
    my $hInLineItem = $LineItem->get(\@ItemAttributes, $LanguageID);

    my $hProductSpecificInformation = $self->_getProductSpecificInformation($LineItem, $LanguageID, $Context);

    my $PriceCalculator = $self->_createPriceCalculator();
    my $hLineItemAttributes = $PriceCalculator->calculate($LineItem, $CurrencyID, $IncludeTaxInLineItem);
    my $BasePrice = $hLineItemAttributes->{'BasePrice'};
    my $LineItemPrice = $hLineItemAttributes->{'LineItemPrice'};
    my $CouponDiscount = $hLineItemAttributes->{'CouponDiscount'};
    my $TaxModelLineItem = $hLineItemAttributes->{'TaxModelLineItem'};

    return {
        lineItemId => $hInLineItem->{'GUID'},
        sku => $hInLineItem->{'SKU'},
        name => $hProductSpecificInformation->{'Name'} // $hInLineItem->{'Name'},
        variationString => $hInLineItem->{'LocalizedVariationString'},
        productId => $hProductSpecificInformation->{'ProductId'},
        quantity  => $Formatter->quantity($hInLineItem->{'Quantity'}, $hInLineItem->{'OrderUnit'}),
        singleItemPrice => $Formatter->price($BasePrice, $CurrencyID, $TaxModelLineItem),
        lineItemPrice => $Formatter->price($LineItemPrice, $CurrencyID, $TaxModelLineItem),
        lineItemCouponDiscount => $Formatter->price($CouponDiscount, $CurrencyID, $TaxModelLineItem),
        taxClass => $self->_createTaxClass($LineItem, $LanguageID),
        images => $hProductSpecificInformation->{'Image'},
        links => $hProductSpecificInformation->{'Links'} // [],
        energyLabelsString => $hProductSpecificInformation->{'EnergyLabelsString'},
        energyLabelsMaxEfficiencyString => $hProductSpecificInformation->{'EnergyLabelHighestEfficiencyClass'},
        energyLabelSourceFile => $hProductSpecificInformation->{'EnergyLabelSourceFilePath'},
        essentialFeatures => $hProductSpecificInformation->{'EssentialFeatures'},
        deliveryWeight => $self->_calculateLineItemProductWeight($LineItem, $Formatter),
    };
}

sub _calculateLineItemProductWeight {
    my $self = shift;
    my ($LineItemProduct, $Formatter) = @_;

    return unless $LineItemProduct->instanceOf('LineItemProduct');

    my @LineitemAttributes = qw(Quantity Weight WeightUnit WeightQuantity);
    my $hLineItemAttributes = $LineItemProduct->get(\@LineitemAttributes);

    my $Quantity = $hLineItemAttributes->{'Quantity'};
    my $ProductWeight = $hLineItemAttributes->{'Weight'};
    my $ProductWeightUnit = $hLineItemAttributes->{'WeightUnit'};
    my $WeightQuantity = $hLineItemAttributes->{'WeightQuantity'};

    if(not defined $WeightQuantity){
        my $Product = $LineItemProduct->get('Product');
        $WeightQuantity = $Product->getWithParent('PriceQuantity') if defined $Product;
        $WeightQuantity = 1 unless $WeightQuantity;
    }

    if (defined $ProductWeight and defined $ProductWeightUnit and $WeightQuantity != 0) {
         my $TotalWeight = $ProductWeight * $Quantity / $WeightQuantity;
         return $Formatter->quantity( $TotalWeight, $hLineItemAttributes->{'WeightUnit'});
    }
    return;
}

sub _getProductSpecificInformation {
    my $self = shift;
    my ($LineItem, $LanguageID, $Context) = @_;
    return unless $LineItem->instanceOf('LineItemProduct');

    my $Product = $LineItem->get('Product', $LanguageID);
    return unless defined $Product;

    my (@Links, $hProduct, $ImageBaseURL);
    push @Links, {
       rel  => 'product',
       href => $Context->apiBaseUrl . '/api/products/' . $Product->guid()
    };
    my $TaxClass = $LineItem->get('TaxClass');
    if(defined $TaxClass) {
       push @Links, {
           rel  => 'taxClass',
           href => $Context->apiBaseUrl . '/api/tax-classes/' . $TaxClass->guid()
       };
    }

    my @ProductAttributes = qw(
        ImageHotDeal
        ImageMedium
        ImageLarge
        ImageSmall
        ImageThumbnail
        EnergyLabelsString
        EnergyLabelHighestEfficiencyClass
        EnergyLabelSourceFile);
    $hProduct = $Product->getWithParent(\@ProductAttributes);
    $ImageBaseURL = $Context->imageBaseUrl().$Product->get('WebPath');

    my $EssentialFeatures = $Product->getWithParent('MainCharacteristics', $LanguageID);
    my $hProductSpecificInformation = {
        'EssentialFeatures' => $EssentialFeatures,
        'Image' => $self->_getImages($hProduct, $ImageBaseURL),
        'Links' => \@Links,
        'Name'  => $Product->get('Name', $LanguageID),
        'ProductId' => $Product->guid(),
        'EnergyLabelsString' => $hProduct->{'EnergyLabelSourceFile'} ? $hProduct->{'EnergyLabelsString'} : undef,
        'EnergyLabelHighestEfficiencyClass' => $hProduct->{'EnergyLabelSourceFile'} ? $hProduct->{'EnergyLabelHighestEfficiencyClass'} : undef,
        'EnergyLabelSourceFilePath' => $hProduct->{'EnergyLabelSourceFile'} ? $ImageBaseURL .'/'. $hProduct->{'EnergyLabelSourceFile'} : undef,
    };

    return $hProductSpecificInformation;
}

sub _createTaxClass{
    my $self = shift;
    my ($LineItem, $LanguageID) = @_;

    my $TaxClass = $LineItem->get('TaxClassDigital') // $LineItem->get('TaxClass');
    return unless defined $TaxClass;

    return {
        'taxClassId' => $TaxClass->guid(),
        'name' => $TaxClass->get('Name', $LanguageID),
        'percentage' => $LineItem->get('TaxRate') * 100,
    };
}

sub _getImages {
    my $self = shift;
    my ($hProduct, $BaseURL) = @_;

    return [
        map { $self->_getImage($_, $hProduct->{$_}, $BaseURL) }
        grep { defined $hProduct->{$_} }
        qw ( ImageHotDeal ImageLarge ImageMedium ImageSmall ImageThumbnail )
    ];
}

sub _getImage {
    my $self = shift;
    my ($Classifier, $Value, $BaseUrl) = @_;

    # Remove 'Image'-prefix (ImageHotdeal -> Hotdeal)
    $Classifier =~ s/^Image//;
    return {
        'classifier' => $Classifier,
        'url' => $BaseUrl . '/' . $Value,
    };
}

sub _createPriceCalculator {
    my $self = shift;
    return DE_EPAGES::Order::API::LineItemPriceCalculator->new();
}

sub _createShopRequestContext {
    my $self = shift;
    my ($Shop) = @_;
    return DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Shop);
}

1;