#========================================================================================
# package      DE_EPAGES::Product::API::Resource::ProductVariation
# state        public
# base         DE_EPAGES::WebService::API::Resource::BaseObjectResource
# description  implements the REST resource for getting variation attributes and sub
#               products from a master product.
#========================================================================================
package DE_EPAGES::Product::API::Resource::ProductVariation;
use base DE_EPAGES::WebService::API::Resource::BaseObjectResource;

use DE_EPAGES::Core::API::Error qw (Error);
use DE_EPAGES::Core::API::Url qw (ParseQueryString);
use DE_EPAGES::Object::API::Factory qw (LoadObjectByPath);
use DE_EPAGES::Object::API::Object::Locale;
use DE_EPAGES::WebService::API::ResourcePager;
use DE_EPAGES::Shop::API::Resource::ShopRequestContext;
use DE_EPAGES::Product::API::ProductVariationService;
use feature 'switch';

use strict;

use constant DEFAULT_PAGE_SIZE => 100;
use constant MAX_PAGE_SIZE => 1000;



sub service {
    return LoadObjectByPath('/RestServices/ProductVariation');
}



sub class {
    return 'Product';
}



sub producedContentTypes {
    shift;
    my ($RequestMethod) = @_;

    return {
        GET => [
            ['get', 'application/vnd.epages.v1+json'],
            ['get', 'application/json'],
        ],
    }->{$RequestMethod};
}



sub match {
    my $self = shift;
    my ($Request) = @_;

    return unless $Request->uri->path =~ m|^/api/products/([^/]+)/variations/?$|;

    return $1;
}



sub href {
    my $self = shift;
    my ($APIBaseURL, $ProductGUID) = @_;
    return "$APIBaseURL/api/products/$ProductGUID/variations";
}



sub params {
    my $self = shift;
    my ($ActionName, $Object, $Request) = @_;

    given( $ActionName ) {
        when( 'get' ) {
            my $hQuery = ParseQueryString( $Request->uri->query, 'utf-8' );
            my $LocaleID = $hQuery->{'locale'}->[0];
            my $Page = $hQuery->{'page'}->[0];
            my $ResultsPerPage = $hQuery->{'resultsPerPage'}->[0];
            my $IncludeInvisible = $hQuery->{'includeInvisible'}->[0];
            return [ $Object, $LocaleID, $Page, $ResultsPerPage, $IncludeInvisible ];
        }
        default {
            return [ $Object ];
        }
    }
}



sub get {
    my $self = shift;
    my ($MasterProduct, $LocaleID, $Page, $PageSize, $IncludeInvisible) = @_;

    $self->_checkIfProductIsMasterProduct($MasterProduct);

    my $Shop = $MasterProduct->getSite();
    my $Context = DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Shop);
    $LocaleID = $Context->validLocale($LocaleID);
    my $Locale= DE_EPAGES::Object::API::Object::Locale->load($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    my $Pager = DE_EPAGES::WebService::API::ResourcePager->new({
        'Page' => $Page,
        'PageSize' => $PageSize,
        'DefaultPageSize' => DEFAULT_PAGE_SIZE,
        'MaxPageSize' => MAX_PAGE_SIZE,
    });

    my $ProductVariationService = DE_EPAGES::Product::API::ProductVariationService->new($MasterProduct, $LanguageID);

    my $aSubProducts = $ProductVariationService->getSubProducts($IncludeInvisible);
    my $aVariationAttributes = $ProductVariationService->getVariationAttributes();

    my $APIBaseURL = $Context->apiBaseUrl();
    my $hPage = $Pager->pagedItems($aSubProducts, sub {
        return $ProductVariationService->getVariationsForSubproduct($_, $APIBaseURL);
    });
    $hPage->{'variationAttributes'} = $aVariationAttributes;
    $hPage->{'links'} = [];

    return $hPage;
}



sub _checkIfProductIsMasterProduct {
    my $self = shift;
    my ($Product) = @_;

    my $aVariationTypeValues = $Product->get('VariationTypeValues');
    Error('ResourceNotFound', {
        MasterProductGUID => $Product->guid }
    ) if 0 == scalar @$aVariationTypeValues;

    return;
}

1;
