#========================================================================================
# package      DE_EPAGES::Product::DAL::Product
# state        private
# description  database interface for tables:
#               <ul>
#                   <li>productmanufacturer
#                   <li>product
#                   <li>productavailabilityemail
#                   <li>productprepayment
#                   <li>productprice
#                   <li>productshippingmethod
#               </ul>
#========================================================================================
package DE_EPAGES::Product::DAL::Product;
use base qw (Exporter);

use strict;
our @EXPORT = qw (
    DAL_InsertProductManufacturer
    DAL_UpdateProductManufacturer
    DAL_DeleteProductManufacturer
    DAL_GetAllPKeyProductManufacturers
    DAL_GetAllInfoProductManufacturers
    DAL_GetInfoProductManufacturer
    DAL_GetPKeyProductManufacturersBySiteID
    DAL_GetInfoProductManufacturersBySiteID

    DAL_InsertProductAvailabilityEmail
    DAL_UpdateProductAvailabilityEmail
    DAL_DeleteProductAvailabilityEmail
    DAL_UpdateProductAvailabilityEmailLocaleID
    DAL_GetAllPKeyProductAvailabilityEmails
    DAL_GetAllInfoProductAvailabilityEmails
    DAL_GetInfoProductAvailabilityEmail
    DAL_GetPKeyProductAvailabilityEmailsByProductID
    DAL_GetInfoProductAvailabilityEmailsByProductID
    DAL_GetPKeyProductAvailabilityEmailsByEMail
    DAL_GetInfoProductAvailabilityEmailsByEMail
    DAL_GetPKeyProductAvailabilityEmailsByLocaleID
    DAL_GetInfoProductAvailabilityEmailsByLocaleID

    DAL_InsertProductPrepayment
    DAL_UpdateProductPrepayment
    DAL_DeleteProductPrepayment
    DAL_GetAllPKeyProductPrepayments
    DAL_GetAllInfoProductPrepayments
    DAL_GetInfoProductPrepayment
    DAL_GetPKeyProductPrepaymentsByProductID
    DAL_GetInfoProductPrepaymentsByProductID
    DAL_GetPKeyProductPrepaymentsByCurrencyID
    DAL_GetInfoProductPrepaymentsByCurrencyID

    DAL_InsertProductPrice
    DAL_UpdateProductPrice
    DAL_DeleteProductPrice
    DAL_GetAllPKeyProductPrices
    DAL_GetAllInfoProductPrices
    DAL_GetInfoProductPrice
    DAL_GetPKeyProductPricesByProductID
    DAL_GetInfoProductPricesByProductID
    DAL_GetPKeyProductPricesByCurrencyID
    DAL_GetInfoProductPricesByCurrencyID

    DAL_InsertProductShippingMethod
    DAL_DeleteProductShippingMethod
    DAL_GetAllPKeyProductShippingMethods
    DAL_GetAllInfoProductShippingMethods
    DAL_GetInfoProductShippingMethod
    DAL_GetPKeyProductShippingMethodsByProductID
    DAL_GetInfoProductShippingMethodsByProductID
    DAL_GetPKeyProductShippingMethodsByShippingMethodID
    DAL_GetInfoProductShippingMethodsByShippingMethodID

    DAL_GetPKeyProductsByAlias
    DAL_CountFeatureProducts

    DAL_GetSmartSearchProductIDs

    DAL_GetSubProductIDsBySuperProductID
    DAL_GetVariationAttributesBySuperProductID
    DAL_GetPKeyPreDefAttributesByProductID

    DAL_GetAllProductIDsAliasOrderedByShopID
    DAL_GetSubProductsOrderedByPrice
    DAL_GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice

    DAL_GetUpdatedProducts
    DAL_GetWatchedProducts

    DAL_UpdateStockLevel
);

use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::PerlTools qw ( UsePackage );
use DE_EPAGES::Database::API::Connection qw ( GetCurrentDBHandle );
use DE_EPAGES::Database::API::DAL;
use DE_EPAGES::DataCache::API::Cache qw ( GetCache UpdateCache GetCacheSize );

my $DAL_PACKAGE = 'DE_EPAGES::Database::API::DAL';
my $DO_CACHE = 1;

sub _CallDriver {
    my ($FunctionName, @Parameter) = @_;
    my $dbi = GetCurrentDBHandle();
    my $Package = 'DE_EPAGES::Product::DAL::'.$dbi->driverName.'::Product';
    UsePackage($Package);
    GetLog->debug("call $Package->$FunctionName()");
    return $Package->$FunctionName($dbi, @Parameter);
}

sub _DoFillInfoCache ($$) {
    my ($CacheName, $CountRows) = @_;
    my $dbi = GetCurrentDBHandle();
    my $CacheSize = GetCacheSize($dbi->cacheKey, $CacheName);
    return $CountRows < $CacheSize ? 1 : 0;
}

#========================================================================================
# function     DAL_InsertProductManufacturer
#----------------------------------------------------------------------------------------
# syntax       $ManufacturerID = DAL_InsertProductManufacturer($hParameter);
#----------------------------------------------------------------------------------------
# description  insert entry in table productmanufacturer
#----------------------------------------------------------------------------------------
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ManufacturerID - manufacturer identifier (primary key)(optional)- int
#                  <li>SiteID        - site id              - int
#                  <li>Name          - manufacturer name    - varchar(100)
#               </ul> | reference to hash
# return       $ManufacturerID | manufacturer identifier (primary key) | int
#========================================================================================
sub DAL_InsertProductManufacturer ($) {
    my ($hParameter) = @_;
    my $ManufacturerID = _CallDriver('InsertProductManufacturer', $hParameter);
    if ($DO_CACHE) {
        my $dbi = GetCurrentDBHandle();
        foreach my $key (keys %$hParameter) {
            $hParameter->{$key} = $dbi->encodeSpaces($hParameter->{$key});
        }
        UpdateCache($dbi->cacheKey,'AllProductManufacturer');
        UpdateCache($dbi->cacheKey,'InfoProductManufacturer', $ManufacturerID);
        UpdateCache($dbi->cacheKey,'ProductManufacturerBySiteID');
    }
    return $ManufacturerID;
}

#========================================================================================
# function     DAL_UpdateProductManufacturer
#----------------------------------------------------------------------------------------
# syntax       DAL_UpdateProductManufacturer($ManufacturerID, $hParameter);
#----------------------------------------------------------------------------------------
# description  updates an entry of table productmanufacturer
#----------------------------------------------------------------------------------------
# input        $ManufacturerID | manufacturer identifier (primary key) | int
# input        $hParameter | with following keys:
#               <ul>
#                   <li>SiteID        - site id              - int
#                   <li>Name          - manufacturer name    - varchar(100)
#               </ul>|reference to hash
#========================================================================================
sub DAL_UpdateProductManufacturer ($$) {
    my ($ManufacturerID, $hParameter) = @_;
    my $hInfo = DAL_GetInfoProductManufacturer($ManufacturerID);
    my $hChangedValues = {};
    my $dbi = GetCurrentDBHandle();
    foreach my $ChangedKey (qw ( SiteID Name )) {
        next unless exists $hParameter->{$ChangedKey};
        next if not defined $hInfo->{$ChangedKey} and not defined $hParameter->{$ChangedKey};
        if ( defined $hParameter->{$ChangedKey}
             and not defined $hInfo->{$ChangedKey}
             or
             not defined $hParameter->{$ChangedKey}
             and defined $hInfo->{$ChangedKey}
        ) {
            $hChangedValues->{$ChangedKey} = $dbi->encodeSpaces($hParameter->{$ChangedKey});
            next;
        }
        next if $hInfo->{$ChangedKey} eq $hParameter->{$ChangedKey};
        $hChangedValues->{$ChangedKey} = $dbi->encodeSpaces($hParameter->{$ChangedKey})
    }
    return 0 unless scalar keys %$hChangedValues;

    _CallDriver('UpdateProductManufacturer', $ManufacturerID, $hChangedValues);
    if($DO_CACHE) {
        UpdateCache($dbi->cacheKey,'InfoProductManufacturer', $ManufacturerID);
        UpdateCache($dbi->cacheKey,'ProductManufacturerBySiteID');
        $hInfo->{$_} = $hChangedValues->{$_} foreach keys %$hChangedValues;
        my $InfoCache = GetCache($dbi->cacheKey,'InfoProductManufacturer');
        $InfoCache->{$ManufacturerID} = $hInfo;
    }
    return 1;
}

#========================================================================================
# function     DAL_DeleteProductManufacturer
#----------------------------------------------------------------------------------------
# syntax       DAL_DeleteProductManufacturer($ManufacturerID);
#----------------------------------------------------------------------------------------
# description  removes an entry of table productmanufacturer
#----------------------------------------------------------------------------------------
# input        $ManufacturerID | manufacturer identifier (primary key) | int
#========================================================================================
sub DAL_DeleteProductManufacturer ($) {
    my ($ManufacturerID) = @_;
    my $cacheKey = GetCurrentDBHandle()->cacheKey;
    my $hInfo = undef;
    if ($DO_CACHE) {
        my $InfoCache = GetCache($cacheKey,'InfoProductManufacturer');
        $hInfo = $InfoCache->{$ManufacturerID};
    }
    _CallDriver('DeleteProductManufacturer', $ManufacturerID);
    if ($DO_CACHE) {
        UpdateCache($cacheKey,'InfoProductManufacturer', $ManufacturerID);
        UpdateCache($cacheKey,'AllProductManufacturer');
        if (defined $hInfo) {
            UpdateCache($cacheKey,'ProductManufacturerBySiteID');
        } else {
            UpdateCache($cacheKey,'ProductManufacturerBySiteID');
        }
    }
    return;
}

#========================================================================================
# function     DAL_GetAllPKeyProductManufacturers
#----------------------------------------------------------------------------------------
# syntax       $raProductManufacturerIDs = DAL_GetAllPKeyProductManufacturers();
#----------------------------------------------------------------------------------------
# description  returns all primary keys of table productmanufacturer
#----------------------------------------------------------------------------------------
# return       $raProductManufacturerIDs | manufacturer identifier | reference to array of int
#========================================================================================
sub DAL_GetAllPKeyProductManufacturers () {
    return _CallDriver('GetAllPKeyProductManufacturers') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductManufacturer');
    $CacheRef->{'1'} = _CallDriver('GetAllPKeyProductManufacturers') unless exists $CacheRef->{'1'};
    return [ @{$CacheRef->{'1'}} ];
}

#========================================================================================
# function     DAL_GetAllInfoProductManufacturers
#----------------------------------------------------------------------------------------
# syntax       $rahProductManufacturerIDs = DAL_GetAllInfoProductManufacturers();
#----------------------------------------------------------------------------------------
# description  returns all rows of table productmanufacturer
#----------------------------------------------------------------------------------------
# return       $rahProductManufacturerIDs | with following keys:
#               <ul>
#                   <li>ManufacturerID - manufacturer identifier (primary key)- int
#                   <li>SiteID        - site id              - int
#                   <li>Name          - manufacturer name    - varchar(100)
#               </ul>| reference to array of hashes
#========================================================================================
sub DAL_GetAllInfoProductManufacturers ();
sub DAL_GetAllInfoProductManufacturers () {
    return _CallDriver('GetAllInfoProductManufacturers') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductManufacturer');
    if (not exists $CacheRef->{'1'}) {
        my $rResultSet = _CallDriver('GetAllInfoProductManufacturers');
        # make copy to primary key cache
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            push @PrimaryKeySet, $hRow->{'ManufacturerID'};
        }
        $CacheRef->{'1'} = \@PrimaryKeySet;
        # make copy to info cache
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductManufacturer');
        if (_DoFillInfoCache('InfoProductManufacturer', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString = $Row->{'ManufacturerID'};
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductManufacturer', @{$CacheRef->{'1'}})) {
        my @ResultSet;
        foreach my $PrimaryKey (@{$CacheRef->{'1'}}) {
            my $Info = DAL_GetInfoProductManufacturer($PrimaryKey);
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{'1'};
    return DAL_GetAllInfoProductManufacturers();
}

#========================================================================================
# function     DAL_GetInfoProductManufacturer
#----------------------------------------------------------------------------------------
# syntax       $hResult = DAL_GetInfoProductManufacturer($ManufacturerID);
#----------------------------------------------------------------------------------------
# description  returns one dataset
#----------------------------------------------------------------------------------------
# input        $ManufacturerID | manufacturer identifier (primary key) | int
# return       $hResult | with following keys:
#               <ul>
#                   <li>ManufacturerID - manufacturer identifier (primary key)- int
#                   <li>SiteID        - site id              - int
#                   <li>Name          - manufacturer name    - varchar(100)
#               </ul>| reference to hash
#========================================================================================
sub DAL_GetInfoProductManufacturer ($) {
    my ($ManufacturerID) = @_;
    return _CallDriver('GetInfoProductManufacturer', $ManufacturerID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef    = GetCache($dbi->cacheKey,'InfoProductManufacturer');
    my $CacheString = $ManufacturerID;
    $CacheRef->{$CacheString} = _CallDriver('GetInfoProductManufacturer', $ManufacturerID) unless exists $CacheRef->{$CacheString};
    return undef unless defined $CacheRef->{$CacheString};
    return {%{$CacheRef->{$CacheString}}};
}

#========================================================================================
# function     DAL_GetPKeyProductManufacturersBySiteID
#----------------------------------------------------------------------------------------
# syntax       $raResult = DAL_GetPKeyProductManufacturersBySiteID($SiteID, $Name);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productmanufacturer by index i_siteid_name
#----------------------------------------------------------------------------------------
# input        $SiteID | site id | int
# input        $Name | manufacturer name (optional) | varchar(100)
# return       $raManufacturerIDs | manufacturer identifier (primary key) | reference to array of int
#========================================================================================
sub DAL_GetPKeyProductManufacturersBySiteID ($;$) {
    my ($SiteID, $Name) = @_;
    return _CallDriver('GetPKeyProductManufacturersBySiteID', $SiteID, $Name) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductManufacturerBySiteID');
    my $CacheString =  join ('<||>', (map {defined $_ ? $_ : ''} $SiteID, $Name));
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductManufacturersBySiteID', $SiteID, $Name) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductManufacturersBySiteID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductManufacturersBySiteID($SiteID, $Name);
#----------------------------------------------------------------------------------------
# description  get full row of table productmanufacturer by index i_siteid_name
#----------------------------------------------------------------------------------------
# input        $SiteID | site id | int
# input        $Name | manufacturer name (optional) | varchar(100)
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ManufacturerID - manufacturer identifier (primary key)- int
#                   <li>SiteID        - site id              - int
#                   <li>Name          - manufacturer name    - varchar(100)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductManufacturersBySiteID ($;$);
sub DAL_GetInfoProductManufacturersBySiteID ($;$) {
    my ($SiteID, $Name) = @_;
    return _CallDriver('GetInfoProductManufacturersBySiteID', $SiteID, $Name) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductManufacturerBySiteID');
    my $CacheString =  join ('<||>', (map {defined $_ ? $_ : ''} $SiteID, $Name));
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductManufacturersBySiteID', $SiteID, $Name);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            push @PrimaryKeySet, $hRow->{'ManufacturerID'};
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductManufacturer');
        if (_DoFillInfoCache('InfoProductManufacturer', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString = $Row->{'ManufacturerID'};
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductManufacturer', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductManufacturer($PrimaryKey);
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductManufacturersBySiteID($SiteID, $Name);
}

#========================================================================================
# function     DAL_InsertProductAvailabilityEmail
#----------------------------------------------------------------------------------------
# syntax       DAL_InsertProductAvailabilityEmail($hParameter);
#----------------------------------------------------------------------------------------
# description  insert entry in table productavailabilityemail
#----------------------------------------------------------------------------------------
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ProductID     - product identifier   (part of primary key)- int
#                  <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#                  <li>LocaleID      - locale id, example en_GB - char(5)
#               </ul> | reference to hash
#========================================================================================
sub DAL_InsertProductAvailabilityEmail ($) {
    my ($hParameter) = @_;
    _CallDriver('InsertProductAvailabilityEmail', $hParameter);
    if ($DO_CACHE) {
        my $dbi = GetCurrentDBHandle();
        foreach my $key (keys %$hParameter) {
            $hParameter->{$key} = $dbi->encodeSpaces($hParameter->{$key});
        }
        UpdateCache($dbi->cacheKey,'AllProductAvailabilityEmail');
        UpdateCache($dbi->cacheKey,'InfoProductAvailabilityEmail', join('<||>', @$hParameter{qw (ProductID EMail)}),$hParameter );
        UpdateCache($dbi->cacheKey,'ProductAvailabilityEmailByProductID', $hParameter->{'ProductID'});
        UpdateCache($dbi->cacheKey,'ProductAvailabilityEmailByEMail', $hParameter->{'EMail'});
        UpdateCache($dbi->cacheKey,'ProductAvailabilityEmailByLocaleID', $hParameter->{'LocaleID'});
    }
    return;
}

#========================================================================================
# function     DAL_UpdateProductAvailabilityEmail
#----------------------------------------------------------------------------------------
# syntax       DAL_UpdateProductAvailabilityEmail($ProductID, $EMail, $hParameter);
#----------------------------------------------------------------------------------------
# description  updates an entry of table productavailabilityemail
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $EMail         | email to send availablity information to (part of primary key) | varchar(150)
# input        $hParameter | with following keys:
#               <ul>
#                   <li>LocaleID      - locale id, example en_GB - char(5)
#               </ul>|reference to hash
#========================================================================================
sub DAL_UpdateProductAvailabilityEmail ($$$) {
    my ($ProductID, $EMail, $hParameter) = @_;
    my $hInfo = DAL_GetInfoProductAvailabilityEmail($ProductID, $EMail);
    my $hChangedValues = {};
    my $dbi = GetCurrentDBHandle();
    foreach my $ChangedKey (qw ( LocaleID )) {
        next unless exists $hParameter->{$ChangedKey};
        next if not defined $hInfo->{$ChangedKey} and not defined $hParameter->{$ChangedKey};
        if ( defined $hParameter->{$ChangedKey}
             and not defined $hInfo->{$ChangedKey}
             or
             not defined $hParameter->{$ChangedKey}
             and defined $hInfo->{$ChangedKey}
        ) {
            $hChangedValues->{$ChangedKey} = $dbi->encodeSpaces($hParameter->{$ChangedKey});
            next;
        }
        next if $hInfo->{$ChangedKey} eq $hParameter->{$ChangedKey};
        $hChangedValues->{$ChangedKey} = $dbi->encodeSpaces($hParameter->{$ChangedKey})
    }
    return 0 unless scalar keys %$hChangedValues;

    _CallDriver('UpdateProductAvailabilityEmail', $ProductID, $EMail, $hChangedValues);
    if($DO_CACHE) {
        # Copy old values and updates ones to a new hash which is used to update the cache.
        my $hMergedCacheInfo = { %$hInfo, %$hChangedValues };
        UpdateCache($dbi->cacheKey,'InfoProductAvailabilityEmail', join('<||>',$ProductID, $EMail), $hMergedCacheInfo);
        if (exists $hChangedValues->{'EMail'}) {
            UpdateCache($dbi->cacheKey,'ProductAvailabilityEmailByEMail', $hInfo->{'EMail'});
            UpdateCache($dbi->cacheKey,'ProductAvailabilityEmailByEMail', $hChangedValues->{'EMail'});
        }
        if (exists $hChangedValues->{'LocaleID'}) {
            UpdateCache($dbi->cacheKey,'ProductAvailabilityEmailByLocaleID', $hInfo->{'LocaleID'});
            UpdateCache($dbi->cacheKey,'ProductAvailabilityEmailByLocaleID', $hChangedValues->{'LocaleID'});
        }
    }
    return 1;
}

#========================================================================================
# function     DAL_UpdateProductAvailabilityEmailLocaleID
#----------------------------------------------------------------------------------------
# syntax       DAL_UpdateProductAvailabilityEmailLocaleID($LocaleIDOld, $LocaleIDNew);
#----------------------------------------------------------------------------------------
# description  updates an entry of table productavailabilityemail
#----------------------------------------------------------------------------------------
# input        $LocaleIDOld | old locale id | String
# input        $LocaleIDNew | new locale id | String
#========================================================================================
sub DAL_UpdateProductAvailabilityEmailLocaleID ($$) {
    my ($LocaleIDOld, $LocaleIDNew) = @_;
    my $dbi = GetCurrentDBHandle();
    _CallDriver('UpdateProductAvailabilityEmailLocaleID', $LocaleIDOld, $LocaleIDNew);
    if($DO_CACHE) {
        UpdateCache($dbi->cacheKey,'InfoProductAvailabilityEmail');
        UpdateCache($dbi->cacheKey,'ProductAvailabilityEmailByLocaleID', $LocaleIDOld);
        UpdateCache($dbi->cacheKey,'ProductAvailabilityEmailByLocaleID', $LocaleIDNew);

    }
    return 1;
}

#========================================================================================
# function     DAL_DeleteProductAvailabilityEmail
#----------------------------------------------------------------------------------------
# syntax       DAL_DeleteProductAvailabilityEmail($ProductID, $EMail);
#----------------------------------------------------------------------------------------
# description  removes an entry of table productavailabilityemail
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $EMail         | email to send availablity information to (part of primary key) | varchar(150)
#========================================================================================
sub DAL_DeleteProductAvailabilityEmail ($$) {
    my ($ProductID, $EMail) = @_;
    my $cacheKey = GetCurrentDBHandle()->cacheKey;
    my $hInfo = undef;
    if ($DO_CACHE) {
        my $InfoCache = GetCache($cacheKey,'InfoProductAvailabilityEmail');
        $hInfo = $InfoCache->{ join('<||>',$ProductID, $EMail)};
    }
    _CallDriver('DeleteProductAvailabilityEmail', $ProductID, $EMail);
    if ($DO_CACHE) {
        UpdateCache($cacheKey,'InfoProductAvailabilityEmail',  join('<||>',$ProductID, $EMail));
        UpdateCache($cacheKey,'AllProductAvailabilityEmail');
        UpdateCache($cacheKey,'ProductAvailabilityEmailByProductID');
        if (defined $hInfo) {
            UpdateCache($cacheKey,'ProductAvailabilityEmailByEMail', $hInfo->{'EMail'});
            UpdateCache($cacheKey,'ProductAvailabilityEmailByLocaleID', $hInfo->{'LocaleID'});
        } else {
            UpdateCache($cacheKey,'ProductAvailabilityEmailByEMail');
            UpdateCache($cacheKey,'ProductAvailabilityEmailByLocaleID');
        }
    }
    return;
}

#========================================================================================
# function     DAL_GetAllPKeyProductAvailabilityEmails
#----------------------------------------------------------------------------------------
# syntax       $rahProductAvailabilityEmailIDs = DAL_GetAllPKeyProductAvailabilityEmails();
#----------------------------------------------------------------------------------------
# description  returns all primary keys of table productavailabilityemail
#----------------------------------------------------------------------------------------
# return       $rahProductAvailabilityEmailIDs | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#               </ul>| reference to array of hashes
#========================================================================================
sub DAL_GetAllPKeyProductAvailabilityEmails () {
    return _CallDriver('GetAllPKeyProductAvailabilityEmails') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductAvailabilityEmail');
    $CacheRef->{'1'} = _CallDriver('GetAllPKeyProductAvailabilityEmails') unless exists $CacheRef->{'1'};
    return [ @{$CacheRef->{'1'}} ];
}

#========================================================================================
# function     DAL_GetAllInfoProductAvailabilityEmails
#----------------------------------------------------------------------------------------
# syntax       $rahProductAvailabilityEmailIDs = DAL_GetAllInfoProductAvailabilityEmails();
#----------------------------------------------------------------------------------------
# description  returns all rows of table productavailabilityemail
#----------------------------------------------------------------------------------------
# return       $rahProductAvailabilityEmailIDs | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#                   <li>LocaleID      - locale id, example en_GB - char(5)
#               </ul>| reference to array of hashes
#========================================================================================
sub DAL_GetAllInfoProductAvailabilityEmails ();
sub DAL_GetAllInfoProductAvailabilityEmails () {
    return _CallDriver('GetAllInfoProductAvailabilityEmails') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductAvailabilityEmail');
    if (not exists $CacheRef->{'1'}) {
        my $rResultSet = _CallDriver('GetAllInfoProductAvailabilityEmails');
        # make copy to primary key cache
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID EMail )} = @$hRow{qw (ProductID EMail )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{'1'} = \@PrimaryKeySet;
        # make copy to info cache
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductAvailabilityEmail');
        if (_DoFillInfoCache('InfoProductAvailabilityEmail', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join('<||>', @$Row{qw (ProductID EMail )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductAvailabilityEmail', @{$CacheRef->{'1'}})) {
        my @ResultSet;
        foreach my $PrimaryKey (@{$CacheRef->{'1'}}) {
            my $Info = DAL_GetInfoProductAvailabilityEmail($PrimaryKey->{'ProductID'}, $PrimaryKey->{'EMail'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{'1'};
    return DAL_GetAllInfoProductAvailabilityEmails();
}

#========================================================================================
# function     DAL_GetInfoProductAvailabilityEmail
#----------------------------------------------------------------------------------------
# syntax       $hResult = DAL_GetInfoProductAvailabilityEmail($ProductID, $EMail);
#----------------------------------------------------------------------------------------
# description  returns one dataset
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $EMail         | email to send availablity information to (part of primary key) | varchar(150)
# return       $hResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#                   <li>LocaleID      - locale id, example en_GB - char(5)
#               </ul>| reference to hash
#========================================================================================
sub DAL_GetInfoProductAvailabilityEmail ($$) {
    my ($ProductID, $EMail) = @_;
    return _CallDriver('GetInfoProductAvailabilityEmail', $ProductID, $EMail) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef    = GetCache($dbi->cacheKey,'InfoProductAvailabilityEmail');
    my $CacheString =  join('<||>',$ProductID, $EMail);
    $CacheRef->{$CacheString} = _CallDriver('GetInfoProductAvailabilityEmail', $ProductID, $EMail) unless exists $CacheRef->{$CacheString};
    return undef unless defined $CacheRef->{$CacheString};
    return {%{$CacheRef->{$CacheString}}};
}

#========================================================================================
# function     DAL_GetPKeyProductAvailabilityEmailsByProductID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetPKeyProductAvailabilityEmailsByProductID($ProductID);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productavailabilityemail by primary key
#----------------------------------------------------------------------------------------
# input        $ProductID | product identifier | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetPKeyProductAvailabilityEmailsByProductID ($) {
    my ($ProductID) = @_;
    return _CallDriver('GetPKeyProductAvailabilityEmailsByProductID', $ProductID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductAvailabilityEmailByProductID');
    my $CacheString = $ProductID;
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductAvailabilityEmailsByProductID', $ProductID) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductAvailabilityEmailsByProductID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductAvailabilityEmailsByProductID($ProductID);
#----------------------------------------------------------------------------------------
# description  get full row of table productavailabilityemail by primary key
#----------------------------------------------------------------------------------------
# input        $ProductID | product identifier | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#                   <li>LocaleID      - locale id, example en_GB - char(5)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductAvailabilityEmailsByProductID ($);
sub DAL_GetInfoProductAvailabilityEmailsByProductID ($) {
    my ($ProductID) = @_;
    return _CallDriver('GetInfoProductAvailabilityEmailsByProductID', $ProductID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductAvailabilityEmailByProductID');
    my $CacheString = $ProductID;
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductAvailabilityEmailsByProductID', $ProductID);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID EMail )} = @$hRow{qw (ProductID EMail )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductAvailabilityEmail');
        if (_DoFillInfoCache('InfoProductAvailabilityEmail', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join ('<||>', @$Row{qw (ProductID EMail )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductAvailabilityEmail', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductAvailabilityEmail($PrimaryKey->{'ProductID'}, $PrimaryKey->{'EMail'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductAvailabilityEmailsByProductID($ProductID);
}

#========================================================================================
# function     DAL_GetPKeyProductAvailabilityEmailsByEMail
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetPKeyProductAvailabilityEmailsByEMail($EMail);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productavailabilityemail by index i_productavailability
#----------------------------------------------------------------------------------------
# input        $EMail | email to send availablity information to | varchar(150)
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetPKeyProductAvailabilityEmailsByEMail ($) {
    my ($EMail) = @_;
    return _CallDriver('GetPKeyProductAvailabilityEmailsByEMail', $EMail) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductAvailabilityEmailByEMail');
    my $CacheString = $EMail;
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductAvailabilityEmailsByEMail', $EMail) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductAvailabilityEmailsByEMail
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductAvailabilityEmailsByEMail($EMail);
#----------------------------------------------------------------------------------------
# description  get full row of table productavailabilityemail by index i_productavailability
#----------------------------------------------------------------------------------------
# input        $EMail | email to send availablity information to | varchar(150)
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#                   <li>LocaleID      - locale id, example en_GB - char(5)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductAvailabilityEmailsByEMail ($);
sub DAL_GetInfoProductAvailabilityEmailsByEMail ($) {
    my ($EMail) = @_;
    return _CallDriver('GetInfoProductAvailabilityEmailsByEMail', $EMail) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductAvailabilityEmailByEMail');
    my $CacheString = $EMail;
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductAvailabilityEmailsByEMail', $EMail);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID EMail )} = @$hRow{qw (ProductID EMail )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductAvailabilityEmail');
        if (_DoFillInfoCache('InfoProductAvailabilityEmail', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join ('<||>', @$Row{qw (ProductID EMail )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductAvailabilityEmail', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductAvailabilityEmail($PrimaryKey->{'ProductID'}, $PrimaryKey->{'EMail'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductAvailabilityEmailsByEMail($EMail);
}

#========================================================================================
# function     DAL_GetPKeyProductAvailabilityEmailsByLocaleID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetPKeyProductAvailabilityEmailsByLocaleID($LocaleID);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productavailabilityemail by index i_productavailabilityemail_locale
#----------------------------------------------------------------------------------------
# input        $LocaleID | locale id, example en_GB | char(5)
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetPKeyProductAvailabilityEmailsByLocaleID ($) {
    my ($LocaleID) = @_;
    return _CallDriver('GetPKeyProductAvailabilityEmailsByLocaleID', $LocaleID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductAvailabilityEmailByLocaleID');
    my $CacheString = $LocaleID;
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductAvailabilityEmailsByLocaleID', $LocaleID) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductAvailabilityEmailsByLocaleID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductAvailabilityEmailsByLocaleID($LocaleID);
#----------------------------------------------------------------------------------------
# description  get full row of table productavailabilityemail by index i_productavailabilityemail_locale
#----------------------------------------------------------------------------------------
# input        $LocaleID | locale id, example en_GB | char(5)
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#                   <li>LocaleID      - locale id, example en_GB - char(5)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductAvailabilityEmailsByLocaleID ($);
sub DAL_GetInfoProductAvailabilityEmailsByLocaleID ($) {
    my ($LocaleID) = @_;
    return _CallDriver('GetInfoProductAvailabilityEmailsByLocaleID', $LocaleID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductAvailabilityEmailByLocaleID');
    my $CacheString = $LocaleID;
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductAvailabilityEmailsByLocaleID', $LocaleID);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID EMail )} = @$hRow{qw (ProductID EMail )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductAvailabilityEmail');
        if (_DoFillInfoCache('InfoProductAvailabilityEmail', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join ('<||>', @$Row{qw (ProductID EMail )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductAvailabilityEmail', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductAvailabilityEmail($PrimaryKey->{'ProductID'}, $PrimaryKey->{'EMail'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductAvailabilityEmailsByLocaleID($LocaleID);
}

#========================================================================================
# function     DAL_InsertProductPrepayment
#----------------------------------------------------------------------------------------
# syntax       DAL_InsertProductPrepayment($hParameter);
#----------------------------------------------------------------------------------------
# description  insert entry in table productprepayment
#----------------------------------------------------------------------------------------
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ProductID     - product identifier   (part of primary key)- int
#                  <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                  <li>Value         -                      - money
#               </ul> | reference to hash
#========================================================================================
sub DAL_InsertProductPrepayment ($) {
    my ($hParameter) = @_;
    _CallDriver('InsertProductPrepayment', $hParameter);
    if ($DO_CACHE) {
        my $dbi = GetCurrentDBHandle();
        foreach my $key (keys %$hParameter) {
            $hParameter->{$key} = $dbi->encodeSpaces($hParameter->{$key});
        }
        UpdateCache($dbi->cacheKey,'AllProductPrepayment');
        UpdateCache($dbi->cacheKey,'InfoProductPrepayment', join('<||>', @$hParameter{qw (ProductID CurrencyID)}),$hParameter );
        UpdateCache($dbi->cacheKey,'ProductPrepaymentByProductID', $hParameter->{'ProductID'});
        UpdateCache($dbi->cacheKey,'ProductPrepaymentByCurrencyID', $hParameter->{'CurrencyID'});
    }
    return;
}

#========================================================================================
# function     DAL_UpdateProductPrepayment
#----------------------------------------------------------------------------------------
# syntax       DAL_UpdateProductPrepayment($ProductID, $CurrencyID, $hParameter);
#----------------------------------------------------------------------------------------
# description  updates an entry of table productprepayment
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $CurrencyID    | alpha currency code (ISO 4217) (part of primary key) | char(3)
# input        $hParameter | with following keys:
#               <ul>
#                   <li>Value         -                      - money
#               </ul>|reference to hash
#========================================================================================
sub DAL_UpdateProductPrepayment ($$$) {
    my ($ProductID, $CurrencyID, $hParameter) = @_;
    my $hInfo = DAL_GetInfoProductPrepayment($ProductID, $CurrencyID);
    my $hChangedValues = {};
    my $dbi = GetCurrentDBHandle();
    foreach my $ChangedKey (qw ( Value )) {
        next unless exists $hParameter->{$ChangedKey};
        next if not defined $hInfo->{$ChangedKey} and not defined $hParameter->{$ChangedKey};
        if ( defined $hParameter->{$ChangedKey}
             and not defined $hInfo->{$ChangedKey}
             or
             not defined $hParameter->{$ChangedKey}
             and defined $hInfo->{$ChangedKey}
        ) {
            $hChangedValues->{$ChangedKey} = $dbi->encodeSpaces($hParameter->{$ChangedKey});
            next;
        }
        next if $hInfo->{$ChangedKey} eq $hParameter->{$ChangedKey};
        $hChangedValues->{$ChangedKey} = $dbi->encodeSpaces($hParameter->{$ChangedKey})
    }
    return 0 unless scalar keys %$hChangedValues;

    _CallDriver('UpdateProductPrepayment', $ProductID, $CurrencyID, $hChangedValues);
    if($DO_CACHE) {
        # Copy old values and updates ones to a new hash which is used to update the cache.
        my $hMergedCacheInfo = { %$hInfo, %$hChangedValues };
        UpdateCache($dbi->cacheKey,'InfoProductPrepayment', join('<||>',$ProductID, $CurrencyID), $hMergedCacheInfo);
        if (exists $hChangedValues->{'CurrencyID'}) {
            UpdateCache($dbi->cacheKey,'ProductPrepaymentByCurrencyID', $hInfo->{'CurrencyID'});
            UpdateCache($dbi->cacheKey,'ProductPrepaymentByCurrencyID', $hChangedValues->{'CurrencyID'});
        }
    }
    return 1;
}

#========================================================================================
# function     DAL_DeleteProductPrepayment
#----------------------------------------------------------------------------------------
# syntax       DAL_DeleteProductPrepayment($ProductID, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  removes an entry of table productprepayment
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $CurrencyID    | alpha currency code (ISO 4217) (part of primary key) | char(3)
#========================================================================================
sub DAL_DeleteProductPrepayment ($$) {
    my ($ProductID, $CurrencyID) = @_;
    my $cacheKey = GetCurrentDBHandle()->cacheKey;
    my $hInfo = undef;
    if ($DO_CACHE) {
        my $InfoCache = GetCache($cacheKey,'InfoProductPrepayment');
        $hInfo = $InfoCache->{ join('<||>',$ProductID, $CurrencyID)};
    }
    _CallDriver('DeleteProductPrepayment', $ProductID, $CurrencyID);
    if ($DO_CACHE) {
        UpdateCache($cacheKey,'InfoProductPrepayment',  join('<||>',$ProductID, $CurrencyID));
        UpdateCache($cacheKey,'AllProductPrepayment');
        UpdateCache($cacheKey,'ProductPrepaymentByProductID');
        if (defined $hInfo) {
            UpdateCache($cacheKey,'ProductPrepaymentByCurrencyID', $hInfo->{'CurrencyID'});
        } else {
            UpdateCache($cacheKey,'ProductPrepaymentByCurrencyID');
        }
    }
    return;
}

#========================================================================================
# function     DAL_GetAllPKeyProductPrepayments
#----------------------------------------------------------------------------------------
# syntax       $rahProductPrepaymentIDs = DAL_GetAllPKeyProductPrepayments();
#----------------------------------------------------------------------------------------
# description  returns all primary keys of table productprepayment
#----------------------------------------------------------------------------------------
# return       $rahProductPrepaymentIDs | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#               </ul>| reference to array of hashes
#========================================================================================
sub DAL_GetAllPKeyProductPrepayments () {
    return _CallDriver('GetAllPKeyProductPrepayments') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductPrepayment');
    $CacheRef->{'1'} = _CallDriver('GetAllPKeyProductPrepayments') unless exists $CacheRef->{'1'};
    return [ @{$CacheRef->{'1'}} ];
}

#========================================================================================
# function     DAL_GetAllInfoProductPrepayments
#----------------------------------------------------------------------------------------
# syntax       $rahProductPrepaymentIDs = DAL_GetAllInfoProductPrepayments();
#----------------------------------------------------------------------------------------
# description  returns all rows of table productprepayment
#----------------------------------------------------------------------------------------
# return       $rahProductPrepaymentIDs | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                   <li>Value         -                      - money
#               </ul>| reference to array of hashes
#========================================================================================
sub DAL_GetAllInfoProductPrepayments ();
sub DAL_GetAllInfoProductPrepayments () {
    return _CallDriver('GetAllInfoProductPrepayments') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductPrepayment');
    if (not exists $CacheRef->{'1'}) {
        my $rResultSet = _CallDriver('GetAllInfoProductPrepayments');
        # make copy to primary key cache
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID CurrencyID )} = @$hRow{qw (ProductID CurrencyID )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{'1'} = \@PrimaryKeySet;
        # make copy to info cache
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductPrepayment');
        if (_DoFillInfoCache('InfoProductPrepayment', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join('<||>', @$Row{qw (ProductID CurrencyID )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductPrepayment', @{$CacheRef->{'1'}})) {
        my @ResultSet;
        foreach my $PrimaryKey (@{$CacheRef->{'1'}}) {
            my $Info = DAL_GetInfoProductPrepayment($PrimaryKey->{'ProductID'}, $PrimaryKey->{'CurrencyID'});
            push @ResultSet, $Info if defined $Info;

        }
        return \@ResultSet;
    }
    delete $CacheRef->{'1'};
    return DAL_GetAllInfoProductPrepayments();
}

#========================================================================================
# function     DAL_GetInfoProductPrepayment
#----------------------------------------------------------------------------------------
# syntax       $hResult = DAL_GetInfoProductPrepayment($ProductID, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  returns one dataset
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $CurrencyID    | alpha currency code (ISO 4217) (part of primary key) | char(3)
# return       $hResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                   <li>Value         -                      - money
#               </ul>| reference to hash
#========================================================================================
sub DAL_GetInfoProductPrepayment ($$) {
    my ($ProductID, $CurrencyID) = @_;
    return _CallDriver('GetInfoProductPrepayment', $ProductID, $CurrencyID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef    = GetCache($dbi->cacheKey,'InfoProductPrepayment');
    my $CacheString =  join('<||>',$ProductID, $CurrencyID);
    $CacheRef->{$CacheString} = _CallDriver('GetInfoProductPrepayment', $ProductID, $CurrencyID) unless exists $CacheRef->{$CacheString};
    return undef unless defined $CacheRef->{$CacheString};
    return {%{$CacheRef->{$CacheString}}};
}

#========================================================================================
# function     DAL_GetPKeyProductPrepaymentsByProductID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetPKeyProductPrepaymentsByProductID($ProductID);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productprepayment by primary key
#----------------------------------------------------------------------------------------
# input        $ProductID | product identifier | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetPKeyProductPrepaymentsByProductID ($) {
    my ($ProductID) = @_;
    return _CallDriver('GetPKeyProductPrepaymentsByProductID', $ProductID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductPrepaymentByProductID');
    my $CacheString = $ProductID;
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductPrepaymentsByProductID', $ProductID) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductPrepaymentsByProductID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductPrepaymentsByProductID($ProductID);
#----------------------------------------------------------------------------------------
# description  get full row of table productprepayment by primary key
#----------------------------------------------------------------------------------------
# input        $ProductID | product identifier | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                   <li>Value         -                      - money
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductPrepaymentsByProductID ($);
sub DAL_GetInfoProductPrepaymentsByProductID ($) {
    my ($ProductID) = @_;
    return _CallDriver('GetInfoProductPrepaymentsByProductID', $ProductID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductPrepaymentByProductID');
    my $CacheString = $ProductID;
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductPrepaymentsByProductID', $ProductID);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID CurrencyID )} = @$hRow{qw (ProductID CurrencyID )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductPrepayment');
        if (_DoFillInfoCache('InfoProductPrepayment', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join ('<||>', @$Row{qw (ProductID CurrencyID )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductPrepayment', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductPrepayment($PrimaryKey->{'ProductID'}, $PrimaryKey->{'CurrencyID'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductPrepaymentsByProductID($ProductID);
}

#========================================================================================
# function     DAL_GetPKeyProductPrepaymentsByCurrencyID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetPKeyProductPrepaymentsByCurrencyID($CurrencyID);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productprepayment by index i_productprepayment_currency
#----------------------------------------------------------------------------------------
# input        $CurrencyID | alpha currency code (ISO 4217) | char(3)
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetPKeyProductPrepaymentsByCurrencyID ($) {
    my ($CurrencyID) = @_;
    return _CallDriver('GetPKeyProductPrepaymentsByCurrencyID', $CurrencyID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductPrepaymentByCurrencyID');
    my $CacheString = $CurrencyID;
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductPrepaymentsByCurrencyID', $CurrencyID) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductPrepaymentsByCurrencyID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductPrepaymentsByCurrencyID($CurrencyID);
#----------------------------------------------------------------------------------------
# description  get full row of table productprepayment by index i_productprepayment_currency
#----------------------------------------------------------------------------------------
# input        $CurrencyID | alpha currency code (ISO 4217) | char(3)
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                   <li>Value         -                      - money
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductPrepaymentsByCurrencyID ($);
sub DAL_GetInfoProductPrepaymentsByCurrencyID ($) {
    my ($CurrencyID) = @_;
    return _CallDriver('GetInfoProductPrepaymentsByCurrencyID', $CurrencyID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductPrepaymentByCurrencyID');
    my $CacheString = $CurrencyID;
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductPrepaymentsByCurrencyID', $CurrencyID);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID CurrencyID )} = @$hRow{qw (ProductID CurrencyID )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductPrepayment');
        if (_DoFillInfoCache('InfoProductPrepayment', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join ('<||>', @$Row{qw (ProductID CurrencyID )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductPrepayment', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductPrepayment($PrimaryKey->{'ProductID'}, $PrimaryKey->{'CurrencyID'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductPrepaymentsByCurrencyID($CurrencyID);
}

#========================================================================================
# function     DAL_InsertProductPrice
#----------------------------------------------------------------------------------------
# syntax       DAL_InsertProductPrice($hParameter);
#----------------------------------------------------------------------------------------
# description  insert entry in table productprice
#----------------------------------------------------------------------------------------
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ProductID     - product identifier   (part of primary key)- int
#                  <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                  <li>ListPrice     - price amount         (optional)- money
#                  <li>ManufacturerPrice -                      (optional)- money
#                  <li>EcoParticipation - Additional fee for electronic products in France (optional)- money
#                  <li>Deposit       - Deposit (e.g. for bottles) (optional)- money
#               </ul> | reference to hash
#========================================================================================
sub DAL_InsertProductPrice ($) {
    my ($hParameter) = @_;
    _CallDriver('InsertProductPrice', $hParameter);
    if ($DO_CACHE) {
        my $dbi = GetCurrentDBHandle();
        foreach my $key (keys %$hParameter) {
            $hParameter->{$key} = $dbi->encodeSpaces($hParameter->{$key});
        }
        UpdateCache($dbi->cacheKey,'AllProductPrice');
        UpdateCache($dbi->cacheKey,'InfoProductPrice', join('<||>', @$hParameter{qw (ProductID CurrencyID)}),$hParameter );
        UpdateCache($dbi->cacheKey,'ProductPriceByProductID', $hParameter->{'ProductID'});
        UpdateCache($dbi->cacheKey,'ProductPriceByCurrencyID', $hParameter->{'CurrencyID'});
    }
    return;
}

#========================================================================================
# function     DAL_UpdateProductPrice
#----------------------------------------------------------------------------------------
# syntax       DAL_UpdateProductPrice($ProductID, $CurrencyID, $hParameter);
#----------------------------------------------------------------------------------------
# description  updates an entry of table productprice
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $CurrencyID    | alpha currency code (ISO 4217) (part of primary key) | char(3)
# input        $hParameter | with following keys:
#               <ul>
#                   <li>ListPrice     - price amount         (optional)- money
#                   <li>ManufacturerPrice -                      (optional)- money
#                   <li>EcoParticipation - Additional fee for electronic products in France (optional)- money
#                   <li>Deposit       - Deposit (e.g. for bottles) (optional)- money
#               </ul>|reference to hash
#========================================================================================
sub DAL_UpdateProductPrice ($$$) {
    my ($ProductID, $CurrencyID, $hParameter) = @_;
    my $hInfo = DAL_GetInfoProductPrice($ProductID, $CurrencyID);
    my $hChangedValues = {};
    my $dbi = GetCurrentDBHandle();
    foreach my $ChangedKey (qw ( ListPrice ManufacturerPrice EcoParticipation Deposit )) {
        next unless exists $hParameter->{$ChangedKey};
        next if not defined $hInfo->{$ChangedKey} and not defined $hParameter->{$ChangedKey};
        if ( defined $hParameter->{$ChangedKey}
             and not defined $hInfo->{$ChangedKey}
             or
             not defined $hParameter->{$ChangedKey}
             and defined $hInfo->{$ChangedKey}
        ) {
            $hChangedValues->{$ChangedKey} = $dbi->encodeSpaces($hParameter->{$ChangedKey});
            next;
        }
        next if $hInfo->{$ChangedKey} eq $hParameter->{$ChangedKey};
        $hChangedValues->{$ChangedKey} = $dbi->encodeSpaces($hParameter->{$ChangedKey})
    }
    return 0 unless scalar keys %$hChangedValues;

    _CallDriver('UpdateProductPrice', $ProductID, $CurrencyID, $hChangedValues);
    if($DO_CACHE) {
        # Copy old values and updates ones to a new hash which is used to update the cache.
        my $hMergedCacheInfo = { %$hInfo, %$hChangedValues };
        UpdateCache($dbi->cacheKey,'InfoProductPrice', join('<||>',$ProductID, $CurrencyID), $hMergedCacheInfo);
        if (exists $hChangedValues->{'CurrencyID'}) {
            UpdateCache($dbi->cacheKey,'ProductPriceByCurrencyID', $hInfo->{'CurrencyID'});
            UpdateCache($dbi->cacheKey,'ProductPriceByCurrencyID', $hChangedValues->{'CurrencyID'});
        }
    }
    return 1;
}

#========================================================================================
# function     DAL_DeleteProductPrice
#----------------------------------------------------------------------------------------
# syntax       DAL_DeleteProductPrice($ProductID, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  removes an entry of table productprice
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $CurrencyID    | alpha currency code (ISO 4217) (part of primary key) | char(3)
#========================================================================================
sub DAL_DeleteProductPrice ($$) {
    my ($ProductID, $CurrencyID) = @_;
    my $cacheKey = GetCurrentDBHandle()->cacheKey;
    my $hInfo = undef;
    if ($DO_CACHE) {
        my $InfoCache = GetCache($cacheKey,'InfoProductPrice');
        $hInfo = $InfoCache->{ join('<||>',$ProductID, $CurrencyID)};
    }
    _CallDriver('DeleteProductPrice', $ProductID, $CurrencyID);
    if ($DO_CACHE) {
        UpdateCache($cacheKey,'InfoProductPrice',  join('<||>',$ProductID, $CurrencyID));
        UpdateCache($cacheKey,'AllProductPrice');
        UpdateCache($cacheKey,'ProductPriceByProductID');
        if (defined $hInfo) {
            UpdateCache($cacheKey,'ProductPriceByCurrencyID', $hInfo->{'CurrencyID'});
        } else {
            UpdateCache($cacheKey,'ProductPriceByCurrencyID');
        }
    }
    return;
}

#========================================================================================
# function     DAL_GetAllPKeyProductPrices
#----------------------------------------------------------------------------------------
# syntax       $rahProductPriceIDs = DAL_GetAllPKeyProductPrices();
#----------------------------------------------------------------------------------------
# description  returns all primary keys of table productprice
#----------------------------------------------------------------------------------------
# return       $rahProductPriceIDs | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#               </ul>| reference to array of hashes
#========================================================================================
sub DAL_GetAllPKeyProductPrices () {
    return _CallDriver('GetAllPKeyProductPrices') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductPrice');
    $CacheRef->{'1'} = _CallDriver('GetAllPKeyProductPrices') unless exists $CacheRef->{'1'};
    return [ @{$CacheRef->{'1'}} ];
}

#========================================================================================
# function     DAL_GetAllInfoProductPrices
#----------------------------------------------------------------------------------------
# syntax       $rahProductPriceIDs = DAL_GetAllInfoProductPrices();
#----------------------------------------------------------------------------------------
# description  returns all rows of table productprice
#----------------------------------------------------------------------------------------
# return       $rahProductPriceIDs | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                   <li>ListPrice     - price amount         (optional)- money
#                   <li>ManufacturerPrice -                      (optional)- money
#                   <li>EcoParticipation - Additional fee for electronic products in France (optional)- money
#                   <li>Deposit       - Deposit (e.g. for bottles) (optional)- money
#               </ul>| reference to array of hashes
#========================================================================================
sub DAL_GetAllInfoProductPrices ();
sub DAL_GetAllInfoProductPrices () {
    return _CallDriver('GetAllInfoProductPrices') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductPrice');
    if (not exists $CacheRef->{'1'}) {
        my $rResultSet = _CallDriver('GetAllInfoProductPrices');
        # make copy to primary key cache
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID CurrencyID )} = @$hRow{qw (ProductID CurrencyID )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{'1'} = \@PrimaryKeySet;
        # make copy to info cache
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductPrice');
        if (_DoFillInfoCache('InfoProductPrice', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join('<||>', @$Row{qw (ProductID CurrencyID )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductPrice', @{$CacheRef->{'1'}})) {
        my @ResultSet;
        foreach my $PrimaryKey (@{$CacheRef->{'1'}}) {
            my $Info = DAL_GetInfoProductPrice($PrimaryKey->{'ProductID'}, $PrimaryKey->{'CurrencyID'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{'1'};
    return DAL_GetAllInfoProductPrices();
}

#========================================================================================
# function     DAL_GetInfoProductPrice
#----------------------------------------------------------------------------------------
# syntax       $hResult = DAL_GetInfoProductPrice($ProductID, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  returns one dataset
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $CurrencyID    | alpha currency code (ISO 4217) (part of primary key) | char(3)
# return       $hResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                   <li>ListPrice     - price amount         (optional)- money
#                   <li>ManufacturerPrice -                      (optional)- money
#                   <li>EcoParticipation - Additional fee for electronic products in France (optional)- money
#                   <li>Deposit       - Deposit (e.g. for bottles) (optional)- money
#               </ul>| reference to hash
#========================================================================================
sub DAL_GetInfoProductPrice ($$) {
    my ($ProductID, $CurrencyID) = @_;
    return _CallDriver('GetInfoProductPrice', $ProductID, $CurrencyID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef    = GetCache($dbi->cacheKey,'InfoProductPrice');
    my $CacheString =  join('<||>',$ProductID, $CurrencyID);
    $CacheRef->{$CacheString} = _CallDriver('GetInfoProductPrice', $ProductID, $CurrencyID) unless exists $CacheRef->{$CacheString};
    return undef unless defined $CacheRef->{$CacheString};
    return {%{$CacheRef->{$CacheString}}};
}

#========================================================================================
# function     DAL_GetPKeyProductPricesByProductID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetPKeyProductPricesByProductID($ProductID);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productprice by primary key
#----------------------------------------------------------------------------------------
# input        $ProductID | product identifier | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetPKeyProductPricesByProductID ($) {
    my ($ProductID) = @_;
    return _CallDriver('GetPKeyProductPricesByProductID', $ProductID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductPriceByProductID');
    my $CacheString = $ProductID;
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductPricesByProductID', $ProductID) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductPricesByProductID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductPricesByProductID($ProductID);
#----------------------------------------------------------------------------------------
# description  get full row of table productprice by primary key
#----------------------------------------------------------------------------------------
# input        $ProductID | product identifier | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                   <li>ListPrice     - price amount         (optional)- money
#                   <li>ManufacturerPrice -                      (optional)- money
#                   <li>EcoParticipation - Additional fee for electronic products in France (optional)- money
#                   <li>Deposit       - Deposit (e.g. for bottles) (optional)- money
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductPricesByProductID ($);
sub DAL_GetInfoProductPricesByProductID ($) {
    my ($ProductID) = @_;
    return _CallDriver('GetInfoProductPricesByProductID', $ProductID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductPriceByProductID');
    my $CacheString = $ProductID;
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductPricesByProductID', $ProductID);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID CurrencyID )} = @$hRow{qw (ProductID CurrencyID )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductPrice');
        if (_DoFillInfoCache('InfoProductPrice', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join ('<||>', @$Row{qw (ProductID CurrencyID )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductPrice', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductPrice($PrimaryKey->{'ProductID'}, $PrimaryKey->{'CurrencyID'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductPricesByProductID($ProductID);
}

#========================================================================================
# function     DAL_GetPKeyProductPricesByCurrencyID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetPKeyProductPricesByCurrencyID($CurrencyID);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productprice by index i_productprice_currency
#----------------------------------------------------------------------------------------
# input        $CurrencyID | alpha currency code (ISO 4217) | char(3)
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetPKeyProductPricesByCurrencyID ($) {
    my ($CurrencyID) = @_;
    return _CallDriver('GetPKeyProductPricesByCurrencyID', $CurrencyID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductPriceByCurrencyID');
    my $CacheString = $CurrencyID;
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductPricesByCurrencyID', $CurrencyID) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductPricesByCurrencyID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductPricesByCurrencyID($CurrencyID);
#----------------------------------------------------------------------------------------
# description  get full row of table productprice by index i_productprice_currency
#----------------------------------------------------------------------------------------
# input        $CurrencyID | alpha currency code (ISO 4217) | char(3)
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                   <li>ListPrice     - price amount         (optional)- money
#                   <li>ManufacturerPrice -                      (optional)- money
#                   <li>EcoParticipation - Additional fee for electronic products in France (optional)- money
#                   <li>Deposit       - Deposit (e.g. for bottles) (optional)- money
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductPricesByCurrencyID ($);
sub DAL_GetInfoProductPricesByCurrencyID ($) {
    my ($CurrencyID) = @_;
    return _CallDriver('GetInfoProductPricesByCurrencyID', $CurrencyID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductPriceByCurrencyID');
    my $CacheString = $CurrencyID;
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductPricesByCurrencyID', $CurrencyID);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID CurrencyID )} = @$hRow{qw (ProductID CurrencyID )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductPrice');
        if (_DoFillInfoCache('InfoProductPrice', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join ('<||>', @$Row{qw (ProductID CurrencyID )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductPrice', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductPrice($PrimaryKey->{'ProductID'}, $PrimaryKey->{'CurrencyID'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductPricesByCurrencyID($CurrencyID);
}

#========================================================================================
# function     DAL_InsertProductShippingMethod
#----------------------------------------------------------------------------------------
# syntax       DAL_InsertProductShippingMethod($hParameter);
#----------------------------------------------------------------------------------------
# description  insert entry in table productshippingmethod
#----------------------------------------------------------------------------------------
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ProductID     - product identifier   (part of primary key)- int
#                  <li>ShippingMethodID - shipping method id   (part of primary key)- int
#               </ul> | reference to hash
#========================================================================================
sub DAL_InsertProductShippingMethod ($) {
    my ($hParameter) = @_;
    _CallDriver('InsertProductShippingMethod', $hParameter);
    if ($DO_CACHE) {
        my $dbi = GetCurrentDBHandle();
        foreach my $key (keys %$hParameter) {
            $hParameter->{$key} = $dbi->encodeSpaces($hParameter->{$key});
        }
        UpdateCache($dbi->cacheKey,'AllProductShippingMethod');
        UpdateCache($dbi->cacheKey,'InfoProductShippingMethod', join('<||>', @$hParameter{qw (ProductID ShippingMethodID)}),$hParameter );
        UpdateCache($dbi->cacheKey,'ProductShippingMethodByProductID', $hParameter->{'ProductID'});
        UpdateCache($dbi->cacheKey,'ProductShippingMethodByShippingMethodID', $hParameter->{'ShippingMethodID'});
    }
    return;
}

#========================================================================================
# function     DAL_DeleteProductShippingMethod
#----------------------------------------------------------------------------------------
# syntax       DAL_DeleteProductShippingMethod($ProductID, $ShippingMethodID);
#----------------------------------------------------------------------------------------
# description  removes an entry of table productshippingmethod
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $ShippingMethodID | shipping method id   (part of primary key) | int
#========================================================================================
sub DAL_DeleteProductShippingMethod ($$) {
    my ($ProductID, $ShippingMethodID) = @_;
    my $cacheKey = GetCurrentDBHandle()->cacheKey;
    my $hInfo = undef;
    if ($DO_CACHE) {
        my $InfoCache = GetCache($cacheKey,'InfoProductShippingMethod');
        $hInfo = $InfoCache->{ join('<||>',$ProductID, $ShippingMethodID)};
    }
    _CallDriver('DeleteProductShippingMethod', $ProductID, $ShippingMethodID);
    if ($DO_CACHE) {
        UpdateCache($cacheKey,'InfoProductShippingMethod',  join('<||>',$ProductID, $ShippingMethodID));
        UpdateCache($cacheKey,'AllProductShippingMethod');
        UpdateCache($cacheKey,'ProductShippingMethodByProductID');
        if (defined $hInfo) {
            UpdateCache($cacheKey,'ProductShippingMethodByShippingMethodID', $hInfo->{'ShippingMethodID'});
        } else {
            UpdateCache($cacheKey,'ProductShippingMethodByShippingMethodID');
        }
    }
    return;
}

#========================================================================================
# function     DAL_GetAllPKeyProductShippingMethods
#----------------------------------------------------------------------------------------
# syntax       $rahProductShippingMethodIDs = DAL_GetAllPKeyProductShippingMethods();
#----------------------------------------------------------------------------------------
# description  returns all primary keys of table productshippingmethod
#----------------------------------------------------------------------------------------
# return       $rahProductShippingMethodIDs | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>ShippingMethodID - shipping method id   (part of primary key)- int
#               </ul>| reference to array of hashes
#========================================================================================
sub DAL_GetAllPKeyProductShippingMethods () {
    return _CallDriver('GetAllPKeyProductShippingMethods') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductShippingMethod');
    $CacheRef->{'1'} = _CallDriver('GetAllPKeyProductShippingMethods') unless exists $CacheRef->{'1'};
    return [ @{$CacheRef->{'1'}} ];
}

#========================================================================================
# function     DAL_GetAllInfoProductShippingMethods
#----------------------------------------------------------------------------------------
# syntax       $rahProductShippingMethodIDs = DAL_GetAllInfoProductShippingMethods();
#----------------------------------------------------------------------------------------
# description  returns all rows of table productshippingmethod
#----------------------------------------------------------------------------------------
# return       $rahProductShippingMethodIDs | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>ShippingMethodID - shipping method id   (part of primary key)- int
#               </ul>| reference to array of hashes
#========================================================================================
sub DAL_GetAllInfoProductShippingMethods ();
sub DAL_GetAllInfoProductShippingMethods () {
    return _CallDriver('GetAllInfoProductShippingMethods') unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'AllProductShippingMethod');
    if (not exists $CacheRef->{'1'}) {
        my $rResultSet = _CallDriver('GetAllInfoProductShippingMethods');
        # make copy to primary key cache
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID ShippingMethodID )} = @$hRow{qw (ProductID ShippingMethodID )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{'1'} = \@PrimaryKeySet;
        # make copy to info cache
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductShippingMethod');
        if (_DoFillInfoCache('InfoProductShippingMethod', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join('<||>', @$Row{qw (ProductID ShippingMethodID )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductShippingMethod', @{$CacheRef->{'1'}})) {
        my @ResultSet;
        foreach my $PrimaryKey (@{$CacheRef->{'1'}}) {
            my $Info = DAL_GetInfoProductShippingMethod($PrimaryKey->{'ProductID'}, $PrimaryKey->{'ShippingMethodID'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{'1'};
    return DAL_GetAllInfoProductShippingMethods();
}

#========================================================================================
# function     DAL_GetInfoProductShippingMethod
#----------------------------------------------------------------------------------------
# syntax       $hResult = DAL_GetInfoProductShippingMethod($ProductID, $ShippingMethodID);
#----------------------------------------------------------------------------------------
# description  returns one dataset
#----------------------------------------------------------------------------------------
# input        $ProductID     | product identifier   (part of primary key) | int
# input        $ShippingMethodID | shipping method id   (part of primary key) | int
# return       $hResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>ShippingMethodID - shipping method id   (part of primary key)- int
#               </ul>| reference to hash
#========================================================================================
sub DAL_GetInfoProductShippingMethod ($$) {
    my ($ProductID, $ShippingMethodID) = @_;
    return _CallDriver('GetInfoProductShippingMethod', $ProductID, $ShippingMethodID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef    = GetCache($dbi->cacheKey,'InfoProductShippingMethod');
    my $CacheString =  join('<||>',$ProductID, $ShippingMethodID);
    $CacheRef->{$CacheString} = _CallDriver('GetInfoProductShippingMethod', $ProductID, $ShippingMethodID) unless exists $CacheRef->{$CacheString};
    return undef unless defined $CacheRef->{$CacheString};
    return {%{$CacheRef->{$CacheString}}};
}

#========================================================================================
# function     DAL_GetPKeyProductShippingMethodsByProductID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetPKeyProductShippingMethodsByProductID($ProductID);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productshippingmethod by primary key
#----------------------------------------------------------------------------------------
# input        $ProductID | product identifier | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>ShippingMethodID - shipping method id   (part of primary key)- int
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetPKeyProductShippingMethodsByProductID ($) {
    my ($ProductID) = @_;
    return _CallDriver('GetPKeyProductShippingMethodsByProductID', $ProductID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductShippingMethodByProductID');
    my $CacheString = $ProductID;
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductShippingMethodsByProductID', $ProductID) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductShippingMethodsByProductID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductShippingMethodsByProductID($ProductID);
#----------------------------------------------------------------------------------------
# description  get full row of table productshippingmethod by primary key
#----------------------------------------------------------------------------------------
# input        $ProductID | product identifier | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>ShippingMethodID - shipping method id   (part of primary key)- int
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductShippingMethodsByProductID ($);
sub DAL_GetInfoProductShippingMethodsByProductID ($) {
    my ($ProductID) = @_;
    return _CallDriver('GetInfoProductShippingMethodsByProductID', $ProductID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductShippingMethodByProductID');
    my $CacheString = $ProductID;
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductShippingMethodsByProductID', $ProductID);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID ShippingMethodID )} = @$hRow{qw (ProductID ShippingMethodID )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductShippingMethod');
        if (_DoFillInfoCache('InfoProductShippingMethod', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join ('<||>', @$Row{qw (ProductID ShippingMethodID )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductShippingMethod', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductShippingMethod($PrimaryKey->{'ProductID'}, $PrimaryKey->{'ShippingMethodID'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductShippingMethodsByProductID($ProductID);
}

#========================================================================================
# function     DAL_GetPKeyProductShippingMethodsByShippingMethodID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetPKeyProductShippingMethodsByShippingMethodID($ShippingMethodID);
#----------------------------------------------------------------------------------------
# description  get primary keys of table productshippingmethod by index i_pshippmethod_smethod
#----------------------------------------------------------------------------------------
# input        $ShippingMethodID | shipping method id | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>ShippingMethodID - shipping method id   (part of primary key)- int
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetPKeyProductShippingMethodsByShippingMethodID ($) {
    my ($ShippingMethodID) = @_;
    return _CallDriver('GetPKeyProductShippingMethodsByShippingMethodID', $ShippingMethodID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductShippingMethodByShippingMethodID');
    my $CacheString = $ShippingMethodID;
    $CacheRef->{$CacheString} = _CallDriver('GetPKeyProductShippingMethodsByShippingMethodID', $ShippingMethodID) unless exists $CacheRef->{$CacheString};
    return [ @{$CacheRef->{$CacheString}} ];
}

#========================================================================================
# function     DAL_GetInfoProductShippingMethodsByShippingMethodID
#----------------------------------------------------------------------------------------
# syntax       $rahResult = DAL_GetInfoProductShippingMethodsByShippingMethodID($ShippingMethodID);
#----------------------------------------------------------------------------------------
# description  get full row of table productshippingmethod by index i_pshippmethod_smethod
#----------------------------------------------------------------------------------------
# input        $ShippingMethodID | shipping method id | int
# return       $rahResult | with following keys:
#               <ul>
#                   <li>ProductID     - product identifier   (part of primary key)- int
#                   <li>ShippingMethodID - shipping method id   (part of primary key)- int
#               </ul>|reference to array of hashes
#========================================================================================
sub DAL_GetInfoProductShippingMethodsByShippingMethodID ($);
sub DAL_GetInfoProductShippingMethodsByShippingMethodID ($) {
    my ($ShippingMethodID) = @_;
    return _CallDriver('GetInfoProductShippingMethodsByShippingMethodID', $ShippingMethodID) unless $DO_CACHE;

    my $dbi = GetCurrentDBHandle();
    my $CacheRef = GetCache($dbi->cacheKey,'ProductShippingMethodByShippingMethodID');
    my $CacheString = $ShippingMethodID;
    my $CacheContent = $CacheRef->{$CacheString};
    if (not defined $CacheContent) {
        my $rResultSet = _CallDriver('GetInfoProductShippingMethodsByShippingMethodID', $ShippingMethodID);
        my (@PrimaryKeySet);
        foreach my $hRow (@$rResultSet) {
            my %hPrimaryKey;
            @hPrimaryKey{qw (ProductID ShippingMethodID )} = @$hRow{qw (ProductID ShippingMethodID )};
            push @PrimaryKeySet, \%hPrimaryKey;
        }
        $CacheRef->{$CacheString} = \@PrimaryKeySet;
        my $InfoCacheRef = GetCache($dbi->cacheKey,'InfoProductShippingMethod');
        if (_DoFillInfoCache('InfoProductShippingMethod', scalar @$rResultSet)) {
            foreach my $Row (@$rResultSet) {
                my $InfoCacheString =  join ('<||>', @$Row{qw (ProductID ShippingMethodID )});
                $InfoCacheRef->{$InfoCacheString} = { %$Row };
            }
        }
        return $rResultSet;
    } elsif (_DoFillInfoCache('InfoProductShippingMethod', scalar @$CacheContent)) {
        my @ResultSet;
        foreach my $PrimaryKey (@$CacheContent) {
            my $Info = DAL_GetInfoProductShippingMethod($PrimaryKey->{'ProductID'}, $PrimaryKey->{'ShippingMethodID'});
            push @ResultSet, $Info if defined $Info;
        }
        return \@ResultSet;
    }
    delete $CacheRef->{$CacheString};
    return DAL_GetInfoProductShippingMethodsByShippingMethodID($ShippingMethodID);
}

#========================================================================================
# function     DAL_GetPKeyProductsByAlias
#----------------------------------------------------------------------------------------
# syntax       $aResult = DAL_GetPKeyProductsByAlias($ShopID, $Alias);
#----------------------------------------------------------------------------------------
# description  Get list of ProductIDs by product alias for a specified shop.
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# input        $Alias | product alias | string
# return       $aResult | list of product ids | array
#========================================================================================
sub DAL_GetPKeyProductsByAlias ($$) {
    my ($ShopID, $Alias) = @_;

    return _CallDriver('GetPKeyProductsByAlias', $ShopID, $Alias);
}

#========================================================================================
# function     DAL_CountFeatureProducts
#----------------------------------------------------------------------------------------
# syntax       $aResult = DAL_CountFeatureProducts($ShopID);
#----------------------------------------------------------------------------------------
# description  Counts products in shop.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop identifier | int
# return       $FeatureCount | used feature count | int
#========================================================================================
sub DAL_CountFeatureProducts ($) {
    my ($ShopID) = @_;
    return _CallDriver('CountFeatureProducts', $ShopID);
}


#========================================================================================
# function     DAL_GetSmartSearchProductIDs
#----------------------------------------------------------------------------------------
# syntax       $aProductIDs = DAL_GetSmartSearchProductIDs($ProductFolderID, $SearchString);
#----------------------------------------------------------------------------------------
# description  Returns products like the given alias.
#----------------------------------------------------------------------------------------
# input        $ProductFolderID | object id of the products folder of the shops | integer
# input        $SearchString | search string | string
# return       $aProductIDs | list of product aliases | ref.array.string
#========================================================================================
sub DAL_GetSmartSearchProductIDs ($$) {
    my ($ProductFolderID, $SearchString) = @_;

    return _CallDriver('GetSmartSearchProductIDs', $ProductFolderID, $SearchString);
}

#========================================================================================
# function     DAL_GetSubProductIDsBySuperProductID
#----------------------------------------------------------------------------------------
# syntax       $aSubProductIDs = DAL_GetSubProductIDsBySuperProductID($SuperProductID, $aVariations);
#----------------------------------------------------------------------------------------
# description  Returns the subproduct ids by their super product id by means of the
#               tables product and selectpredefattribute.
#----------------------------------------------------------------------------------------
# input        $ProductID | product type id (part of primary key) | int
# input        $aVariations | array of PreDefAttribute object ids of the attribute values | ref.array.int
# return       $aSubProductIDs | array of SubProductIDs | ref.array.int
#========================================================================================
sub DAL_GetSubProductIDsBySuperProductID ($$) {
    my ($SuperProductID, $aVariations) = @_;

    return _CallDriver('GetSubProductIDsBySuperProductID', $SuperProductID, $aVariations);
}

#========================================================================================
# function     DAL_GetVariationAttributesBySuperProductID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aaResult = DAL_GetVariationAttributesBySuperProductID(
#                   $SuperProductID,
#                   $aVariationAttributeIDs
#               );
#----------------------------------------------------------------------------------------
# description  Returns all sub product ids and all respective predef attribute id's
#               for the given super product id.
#----------------------------------------------------------------------------------------
# input        $dbi | database handle | object
# input        $SuperProductID | super product identifier | int
# input        $aVariationAttributeIDs | array of all variation attribute id's of the
#                   product | ref.array.int
# return       $aaResult | array of arrays containing sub product id and variation
#                   attribute value id's | ref.array.array.int
#========================================================================================
sub DAL_GetVariationAttributesBySuperProductID ($$) {
    my ($SuperProductID, $aVariationAttributeIDs) = @_;

    return _CallDriver('GetVariationAttributesBySuperProductID', $SuperProductID, $aVariationAttributeIDs);
}

#========================================================================================
# function     DAL_GetAllProductIDsAliasOrderedByShopID
#----------------------------------------------------------------------------------------
# syntax       $aProductIDs = DAL_GetAllProductIDsAliasOrderedByShopID($ShopID);
#----------------------------------------------------------------------------------------
# description  Returns an iterator over the IDs of all shop products ordered by their alias
#               with the restriction that super products have to be in front of their subs.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id of the products | integer
# return       $Iterator | object iterator | DE_EPAGES::Core::API::Iterator
#========================================================================================
sub DAL_GetAllProductIDsAliasOrderedByShopID ($) {
    my ($ShopID) = @_;

    return _CallDriver('GetAllProductIDsAliasOrderedByShopID', $ShopID);
}

#========================================================================================
# function     DAL_GetSubProductsOrderedByPrice
#----------------------------------------------------------------------------------------
# syntax       $Iterator = DAL_GetSubProductsOrderedByPrice($SuperProductID, $CurrencyID, $SortOrder);
#----------------------------------------------------------------------------------------
# description  Returns an iterator of product ids for all sub produsts ordered by price.
#----------------------------------------------------------------------------------------
# input        $SuperProductID | id of the super product | integer
# input        $CurrencyID | currency id, e.g. 'USD' | string
# input        $SortOrder | (optional default: ASC) sort order 'ASC' or 'DESC' | string
# return       $Iterator | iterator of product ids | DE_EPAGES::Core::API::Iterator
#========================================================================================
sub DAL_GetSubProductsOrderedByPrice {
    my ($SuperProductID, $CurrencyID, $SortOrder) = @_;

    return _CallDriver('GetSubProductsOrderedByPrice', $SuperProductID, $CurrencyID, $SortOrder);
}

#========================================================================================
# function     DAL_GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice
#----------------------------------------------------------------------------------------
# syntax       $Iterator = DAL_GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice(
#                       $SuperProductID, $CurrencyID, $SortOrder
#               );
#----------------------------------------------------------------------------------------
# description  Returns an iterator of product ids for all sub produsts ordered by price.
#----------------------------------------------------------------------------------------
# input        $SuperProductID | id of the super product | integer
# input        $CurrencyID | currency id, e.g. 'USD' | string
# input        $SortOrder | (optional default: ASC) sort order 'ASC' or 'DESC' | string
# return       $Iterator | iterator of product ids | DE_EPAGES::Core::API::Iterator
#========================================================================================
sub DAL_GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice {
    my ($SuperProductID, $CurrencyID, $SortOrder) = @_;

    return _CallDriver(
            'GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice',
            $SuperProductID,
            $CurrencyID,
            $SortOrder
    );
}

#========================================================================================
# function     DAL_GetUpdatedProducts
#----------------------------------------------------------------------------------------
# syntax       my @ProductIDs = DAL_GetUpdatedProducts($ShopId, $AttributeSet, $UpdatedAfter);
#----------------------------------------------------------------------------------------
# description  Returns an array of ids of products which where updated.
#               Set can be 'RestApiResponseAttributes' or 'RestApiStockLevel'.
#----------------------------------------------------------------------------------------
# input        $ShopId | object id of the shop | int
# input        $AttributeSet | name of the LastUpdateProfile | string
# input        $UpdatedAfter | time after which the update happened | datetime
# return       array of arrays of productID and lastChangeDate | array ref
#========================================================================================
sub DAL_GetUpdatedProducts {
    my ($ShopId, $AttributeSet, $UpdatedAfter) = @_;

    return _CallDriver('GetUpdatedProducts', $ShopId, $AttributeSet, $UpdatedAfter);
}

#========================================================================================
# function     DAL_GetWatchedProducts
#----------------------------------------------------------------------------------------
# syntax       my @ProductIDs = DAL_GetWatchedProducts($ShopId);
#----------------------------------------------------------------------------------------
# description  Returns an array of ids of watched products and their number of watchers.
#----------------------------------------------------------------------------------------
# input        $ShopId | object id of the shop | int
# return       array of arrays of productID and number of watchers | array ref
#========================================================================================
sub DAL_GetWatchedProducts {
    my ($ShopId) = @_;

    return _CallDriver('GetWatchedProducts', $ShopId);
}

#========================================================================================
# function     DAL_UpdateStockLevel
#----------------------------------------------------------------------------------------
# syntax       DAL_UpdateStockLevel($ObjectID, $amount);
#----------------------------------------------------------------------------------------
# description  Change the stocklevel relative of a given product.
#----------------------------------------------------------------------------------------
# input        $ObjectID | object id of product | int
# input        $Amount   | The amount if items, which should change the stock level | float
#========================================================================================
sub DAL_UpdateStockLevel {
    my ($ObjectID, $Amount) = @_;

    return _CallDriver('UpdateStockLevel', $ObjectID, $Amount);
}

1;
