#========================================================================================
# package      DE_EPAGES::Product::DAL::mysql::Product
# state        private
# description  MySQL database interface for tables:
#               <ul>
#                   <li>productmanufacturer
#                   <li>product
#                   <li>productavailabilityemail
#                   <li>productprepayment
#                   <li>productprice
#                   <li>productshippingmethod
#               </ul>
#========================================================================================
package DE_EPAGES::Product::DAL::mysql::Product;
use base qw( DE_EPAGES::Product::DAL::SQL::Product );

use strict;
use DE_EPAGES::Product::API::Constants qw ( SMART_PRODUCT_SEARCH_RESULT_COUNT );

#========================================================================================
# function     InsertProductManufacturer
#----------------------------------------------------------------------------------------
# syntax       $ManufacturerID = $Package->InsertProductManufacturer($dbh, $hParameter);
#----------------------------------------------------------------------------------------
# description  Inserts an entry into table productmanufacturer.
#----------------------------------------------------------------------------------------
# input        $dbi | database handle | DE_EPAGES::Database::API::DBI
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ManufacturerID - manufacturer identifier (primary key)(optional)- int
#                  <li>SiteID        - site id              - int
#                  <li>Name          - manufacturer name    - varchar(100)
#               </ul> | reference to hash
# return       $ManufacturerID | manufacturer identifier (primary key) | int
#========================================================================================
sub InsertProductManufacturer {
    my $self = shift;
    my ($dbi, $hParameter) = @_;
    $dbi->do('INSERT INTO productmanufacturer (siteid, name) VALUES(?, ?)',
         [ $hParameter->{'SiteID'}, $hParameter->{'Name'} ]
    );
    my $ManufacturerID = $dbi->dbh->{'mysql_insertid'}; # auto-generated key
    die 'Failed to create primary key for ProductManufacturer' unless defined $ManufacturerID;
    return $ManufacturerID;
}

#========================================================================================
# function     InsertProductAvailabilityEmail
#----------------------------------------------------------------------------------------
# syntax       $Package->InsertProductAvailabilityEmail($dbh, $hParameter);
#----------------------------------------------------------------------------------------
# description  Inserts an entry into table productavailabilityemail.
#----------------------------------------------------------------------------------------
# input        $dbi | database handle | DE_EPAGES::Database::API::DBI
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ProductID     - product identifier   (part of primary key)- int
#                  <li>EMail         - email to send availablity information to (part of primary key)- varchar(150)
#                  <li>LocaleID      - locale id, example en_GB - char(5)
#               </ul> | reference to hash
#========================================================================================
sub InsertProductAvailabilityEmail {
    my $self = shift;
    my ($dbi, $hParameter) = @_;
    $dbi->do('INSERT INTO productavailabilityemail (productid, email, localeid) VALUES (?, ?, ?)',
        [ @{$hParameter}{qw( ProductID EMail LocaleID )} ]
    );
}

#========================================================================================
# function     InsertProductPrepayment
#----------------------------------------------------------------------------------------
# syntax       $Package->InsertProductPrepayment($dbh, $hParameter);
#----------------------------------------------------------------------------------------
# description  Inserts an entry into table productprepayment.
#----------------------------------------------------------------------------------------
# input        $dbi | database handle | DE_EPAGES::Database::API::DBI
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ProductID     - product identifier   (part of primary key)- int
#                  <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                  <li>Value         -                      - money
#               </ul> | reference to hash
#========================================================================================
sub InsertProductPrepayment {
    my $self = shift;
    my ($dbi, $hParameter) = @_;
    $dbi->do('INSERT INTO productprepayment (productid, currencyid, value) VALUES (?, ?, ?)',
        [ @{$hParameter}{qw( ProductID CurrencyID Value )} ]
    );
}

#========================================================================================
# function     InsertProductPrice
#----------------------------------------------------------------------------------------
# syntax       $Package->InsertProductPrice($dbh, $hParameter);
#----------------------------------------------------------------------------------------
# description  Inserts an entry into table productprice.
#----------------------------------------------------------------------------------------
# input        $dbi | database handle | DE_EPAGES::Database::API::DBI
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ProductID     - product identifier   (part of primary key)- int
#                  <li>CurrencyID    - alpha currency code (ISO 4217) (part of primary key)- char(3)
#                  <li>ListPrice     - price amount         (optional)- money
#                  <li>ManufacturerPrice -                      (optional)- money
#                  <li>EcoParticipation - Additional fee for electronic products in France (optional)- money
#                  <li>Deposit       - Deposit (e.g. for bottles) (optional)- money
#               </ul> | reference to hash
#========================================================================================
sub InsertProductPrice {
    my $self = shift;
    my ($dbi, $hParameter) = @_;

    foreach my $key (keys %$hParameter){
        $hParameter->{$key} = undef if ( (defined $hParameter->{$key}) and ($hParameter->{$key} eq '') );
    }

    $dbi->do('INSERT INTO productprice (productid, currencyid, listprice, manufacturerprice, ecoparticipation, deposit) VALUES (?, ?, ?, ?, ?, ?)',
        [ @{$hParameter}{qw( ProductID CurrencyID ListPrice ManufacturerPrice EcoParticipation Deposit )} ]
    );
}

#========================================================================================
# function     InsertProductShippingMethod
#----------------------------------------------------------------------------------------
# syntax       $Package->InsertProductShippingMethod($dbh, $hParameter);
#----------------------------------------------------------------------------------------
# description  Inserts an entry into table productshippingmethod.
#----------------------------------------------------------------------------------------
# input        $dbi | database handle | DE_EPAGES::Database::API::DBI
# input        $hParameter | with following keys:
#               <ul>
#                  <li>ProductID     - product identifier   (part of primary key)- int
#                  <li>ShippingMethodID - shipping method id   (part of primary key)- int
#               </ul> | reference to hash
#========================================================================================
sub InsertProductShippingMethod {
    my $self = shift;
    my ($dbi, $hParameter) = @_;
    $dbi->do('INSERT INTO productshippingmethod (productid, shippingmethodid) VALUES (?, ?)',
        [ @{$hParameter}{qw( ProductID ShippingMethodID )} ]
    );
}

#========================================================================================
# function     GetSmartSearchProductIDs
#----------------------------------------------------------------------------------------
# syntax       $aProductIDs = $Package->GetSmartSearchProductIDs($dbh, $ProductFolderID, $SearchString);
#----------------------------------------------------------------------------------------
# description  Search for a product alias in a specified product folder of the shop and
#               return a list of product ids.
#----------------------------------------------------------------------------------------
# input        $dbi | database handle | object
# input        $ProductFolderID | id of the product folder of the shop | integer
# input        $SearchString | search string | string
# return       $aProductIDs | list of product ids | ref.array.integer
#========================================================================================
sub GetSmartSearchProductIDs {
    my $self = shift;
    my ($dbi, $ProductFolderID, $SearchString) = @_;

    my $sql = 'SELECT objectid FROM object WHERE parentid = ? AND alias LIKE ? LIMIT ?';
    my $aRows = $dbi->execute($sql, [$ProductFolderID, $SearchString, SMART_PRODUCT_SEARCH_RESULT_COUNT], ['Integer']);
    return [ map { $_->[0] } @$aRows];
}

#========================================================================================
# function     GetSubProductsOrderedByPrice
#----------------------------------------------------------------------------------------
# syntax       $Iterator = $Package->GetSubProductsOrderedByPrice($dbh, $SuperProductID, $CurrencyID, $SortOrder);
#----------------------------------------------------------------------------------------
# description  Returns an iterator of product ids for all sub produsts ordered by price.
#----------------------------------------------------------------------------------------
# input        $dbi | database handle | object
# input        $SuperProductID | id of the super product | integer
# input        $CurrencyID | currency id, e.g. 'USD' | string
# input        $SortOrder | (optional default: ASC) sort order 'ASC' or 'DESC' | string
# return       $Iterator | iterator of product ids | DE_EPAGES::Core::API::Iterator
#========================================================================================
sub GetSubProductsOrderedByPrice {
    my $self = shift;
    my ($dbi, $SuperProductID, $CurrencyID, $SortOrder) = @_;

    # prevent SQL injection (we can not use order by with a prepared statement parameter)
    my $Direction = 'ASC'; 
    if (defined $SortOrder and $SortOrder eq 'DESC'){
        $Direction = 'DESC';    
    }  

    return DE_EPAGES::Core::API::Iterator->wrap(
        Handle => sub { $_[1]->[0] },
        Source => $dbi->iterator( <<'END_SQL' . $Direction , [$CurrencyID, $CurrencyID, $SuperProductID] ) );
SELECT p.productid
FROM product p
LEFT JOIN productprice pp ON pp.productid = p.productid AND pp.currencyid = ?
LEFT JOIN productprice mpp ON mpp.productid = p.superproductid AND mpp.currencyid = ?
WHERE p.superproductid = ?
  AND IF(p.hassubownprices <> 0, pp.listprice, mpp.listprice) IS NOT NULL
ORDER BY IF(p.hassubownprices <> 0, pp.listprice, mpp.listprice)
END_SQL
}

#========================================================================================
# function     GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice
#----------------------------------------------------------------------------------------
# syntax       $Iterator = $Package->GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice(
#                       $dbh, $SuperProductID, $CurrencyID, $SortOrder
#               );
#----------------------------------------------------------------------------------------
# description  Returns an iterator of product ids for specific sub products ordered by price.
#----------------------------------------------------------------------------------------
# input        $dbi | database handle | object
# input        $SuperProductID | id of the super product | integer
# input        $CurrencyID | currency id, e.g. 'USD' | string
# input        $SortOrder | (optional default: ASC) sort order 'ASC' or 'DESC' | string
# return       $Iterator | iterator of product ids | DE_EPAGES::Core::API::Iterator
#========================================================================================
sub GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice {
    my $self = shift;
    my ($dbi, $SuperProductID, $CurrencyID, $SortOrder) = @_;

    # prevent SQL injection (we can not use order by with a prepared statement parameter)
    my $Direction = 'ASC'; 
    if (defined $SortOrder and $SortOrder eq 'DESC'){
        $Direction = 'DESC';    
    }  

    return DE_EPAGES::Core::API::Iterator->wrap(
        Handle => sub { $_[1]->[0] },
        Source => $dbi->iterator( <<'END_SQL' . $Direction , [$CurrencyID, $SuperProductID] ) );
SELECT p.productid
FROM product p
LEFT JOIN productprice pp ON pp.productid = p.productid AND pp.currencyid = ?
LEFT JOIN contentobject co ON co.contentobjectid = p.productid
WHERE p.superproductid = ?
      AND p.hassubownprices = 1
      AND p.isavailable = 1
      AND pp.listprice IS NOT NULL
      AND co.isvisible = 1
ORDER BY pp.listprice
END_SQL
}

#========================================================================================
# function     UpdateStockLevel
#----------------------------------------------------------------------------------------
# syntax       UpdateStockLevel($dbi, $ObjectID, $Amount);
#----------------------------------------------------------------------------------------
# description  Update the stocklevel of a product with a relative value.
#----------------------------------------------------------------------------------------
# input        $dbi      | database handle | object
# input        $ObjectID | object id of the product | int
# input        $Amount   | amount of stock to increase | float
#========================================================================================
sub UpdateStockLevel {
    my $self = shift;
    my ($dbi, $ObjectID, $Amount) = @_;

    my $StockLevel;

    $dbi->transaction( sub {
        $dbi->do('UPDATE product SET stocklevel = stocklevel + ? WHERE productid = ?',
            [ $Amount, $ObjectID ]);

        $StockLevel = $dbi->execute('SELECT stocklevel FROM product WHERE productid = ?', [ $ObjectID ])->[0]->[0];
    } );

    return $StockLevel;
}

1;
