package DE_EPAGES::Product::t::Product;
use base DE_EPAGES::Shop::API::Test;

use Module::Spy;
use Test::More;
use Test::Deep;
use Time::HiRes qw (time);
use strict;
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Core::API::File qw ( CopyFile DeleteFile WriteFile CompareMD5TextFile GetFileMD5Digest);
use DE_EPAGES::Core::API::Object::DateTimeFormatter;
use DE_EPAGES::Core::API::PerlTools qw ( ElementOf );
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadRootObject LoadObjectByPath );
use DE_EPAGES::Object::API::Language qw ( GetPKeyLanguageByCode );
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw ( ImportObjects );
use DE_EPAGES::Product::API::CSV::ProductImportHandler;
use DE_EPAGES::Product::API::Product qw (
    ExistsSubProductByAttributeValues
    GetSubProductByAttributeValues
    InsertMultipleSubProducts
    SmartSearchProducts
    UsedVariationValues
    GetUnityPrettyUrl
    GetVariationAttributes
);
use DE_EPAGES::Shop::API::Constants qw ( TAX_MODEL_GROSS TAX_MODEL_NET );
use DE_EPAGES::ShopCSVExportImport::API::CSV::ShopExportImport qw ( ImportCSV ExportCSV );

## no critic qw( ProhibitMagicNumbers ProhibitLongLines )

sub testBundle {
    my $self = shift;

    my $Product = $self->shop->path('Products/Bundle');
    ok( $Product->get('IsBundleProduct'), 'Product.IsBundleProduct' );
    ok( !$Product->get('IsDownloadProduct'), 'not Product.IsDownloadProduct' );
    ok( !$Product->get('IsAppointment'), 'not Product.IsAppointment' );

    my $aBundledProducts = $Product->get('BundledProducts');
    is( scalar @$aBundledProducts, 2, 'Product.BundledProducts' );

    my $Product1 = $aBundledProducts->[0];
    my $aIncludedInBundles = $Product1->get('IncludedInBundles');
    is( scalar @$aIncludedInBundles, 1, 'Product.IncludedInBundles' );
    ok( $aIncludedInBundles->[0]->id == $Product->id, 'Product.IncludedInBundles' );
}

sub testDownload {
    my $self = shift;

    my $Product = $self->shop->path('Products/DownloadProduct');
    ok( $Product->get('IsDownloadProduct'), 'Product.IsDownloadProduct' );
    ok( !$Product->get('IsAppointment'), 'not Product.IsAppointment' );
    ok( !$Product->get('IsBundleProduct'), 'not Product.IsBundleProduct' );

    my $aDownloads = $Product->get('DownloadProductMaps');
    is( scalar @$aDownloads, 1, 'Product.DownloadProductMaps' );

    my $aDirectDownloads = $Product->get('DirectDownloadProductMaps');
    is( scalar @$aDownloads, 1, 'Product.DirectDownloadProductMaps' );

    my $ProductEN = $Product->child('SubProducts')->child('DownloadProduct-en');
    $aDownloads = $ProductEN->get('DownloadProductMaps');
    is( scalar @$aDownloads, 2, 'SubProducts.DownloadProductMaps' );
    is( $aDownloads->[0]->get('TargetUrl'), 'Product.exe', 'files of master product before sub product' );
    is( $aDownloads->[1]->get('TargetUrl'), 'Product-en.dll', 'files of sub product after master product' );

    $aDirectDownloads = $ProductEN->get('DirectDownloadProductMaps');
    is( scalar @$aDirectDownloads, 1, 'SubProduct.DirectDownloadProductMaps' );

    my $ProductDE = $Product->child('SubProducts')->child('DownloadProduct-de');
    $aDownloads = $ProductDE->get('DownloadProductMaps');
    is( scalar @$aDownloads, 4, 'DownloadProductMaps of DownloadProduct-de' );
    my $DownloadLocal = $aDownloads->[0];
    is( $DownloadLocal->get('IsExternal'), 0, 'DownloadProductMap.IsExternal' );
    is( $DownloadLocal->get('FileName'), 'Product.exe', 'DownloadProductMap.FileName for local file' );

    my $DownloadExtern = $aDownloads->[3];
    is( $DownloadExtern->get('IsExternal'), 1, 'DownloadProductMap.IsExternal' );
    is( $DownloadExtern->get('FileName'), 'epages7.exe', 'DownloadProductMap.FileName for external file' );
}

sub testAppointment {
    my $self = shift;

    my $Product = $self->shop->path('Appointments/Table4Person');
    ok( $Product->get('IsAppointment'), 'Product.IsAppointment' );
    ok( !$Product->get('IsDownloadProduct'), 'not Product.IsDownloadProduct' );
    ok( !$Product->get('IsBundleProduct'), 'not Product.IsBundleProduct' );

    my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new('locale'=>'de_DE');
    my $StartDate = $Formatter->parse_datetime('20.02.2008 16:20:00','datetime', 'medium');
    my $EndDate1 = $Formatter->parse_datetime('20.02.2009 16:13:30','datetime', 'medium');
    my $EndDate2 = $Formatter->parse_datetime('20.02.2009 16:23:30','datetime', 'medium');
    my $EndDate3 = $Formatter->parse_datetime('20.02.2010 16:23:30','datetime', 'medium');
    my $EndDate4 = $Formatter->parse_datetime('20.02.2008 16:50:00','datetime', 'medium');
    my $EndDate5 = $Formatter->parse_datetime('20.02.2008 17:20:00','datetime', 'medium');
    my $EndDate6 = $Formatter->parse_datetime('20.02.2008 18:20:00','datetime', 'medium');

    is ($Product->timeFactor($StartDate,$EndDate1), 4391.945833, '2 hours without overlap');
    is ($Product->timeFactor($StartDate,$EndDate2), 4392.029167, '2 hours with overlap');
    is ($Product->timeFactor($StartDate,$EndDate4), 0.25, '1/2 hour');
    is ($Product->timeFactor($StartDate,$EndDate5), 0.5, '1 hour');
    is ($Product->timeFactor($StartDate,$EndDate6), 1, '2 hours');

    $Product->set({DurationUnit=>LoadObjectByPath('/Units/second/month'), DurationAmount=>1});

    is ($Product->timeFactor($StartDate,$EndDate1), 11.966667, 'month without overlap');
    is ($Product->timeFactor($StartDate,$EndDate2), 12.033333, 'month with overlap');

    $Product->set({DurationUnit=>LoadObjectByPath('/Units/second/month'), DurationAmount=>2});
    is ($Product->timeFactor($StartDate,$EndDate1), 5.983333, '2 month without overlap');
    is ($Product->timeFactor($StartDate,$EndDate2), 6.016667, '2 month with overlap');

    $Product->set({DurationUnit=>LoadObjectByPath('/Units/second/year'), DurationAmount=>1});
    is ($Product->timeFactor($StartDate,$EndDate1), 0.997222, 'year without overlap');
    is ($Product->timeFactor($StartDate,$EndDate2), 1.002778, 'year with overlap');
    is ($Product->timeFactor($StartDate,$EndDate3), 2.002778, 'year with more overlap');
    is ($Product->timeFactor($StartDate,$EndDate5), 0.002778, '1 hour for duration of 1 year');

    $Product->set({DurationUnit=>LoadObjectByPath('/Units/second/year'), DurationAmount=>2});
    is ($Product->timeFactor($StartDate,$EndDate1), 0.498611, '2 year without overlap');
    is ($Product->timeFactor($StartDate,$EndDate2), 0.501389, '2 years with overlap');
    is ($Product->timeFactor($StartDate,$EndDate3), 1.001389, '2 years with more overlap');
}

sub testSmartSearch {
    my $self = shift;

    my $Shop = $self->shop;
    my $aProducts = SmartSearchProducts($Shop, 'SmartSearch-');

    is (scalar(@$aProducts), 1, 'Smartsearch founds one');
    is ($aProducts->[0]->get('Alias'), 'SmartSearch-123', 'SmartSearch finds the right product');

    $aProducts = SmartSearchProducts($Shop, 'SmartSearch_');

    is (scalar(@$aProducts), 1, 'Smartsearch founds one');
    is ($aProducts->[0]->get('Alias'), 'SmartSearch_123', 'SmartSearch finds the right product');

    $aProducts = SmartSearchProducts($Shop, 'SmartSearch');
    is (scalar(@$aProducts), 2, 'Smartsearch founds one');
}

sub testGetSubProductByAttributeValues {
    my $self = shift;

    # check erroneously passed parameters
    eval {
        GetSubProductByAttributeValues(undef, undef);
    };
    ok( ExistsError() && GetError()->code eq 'PARAMETER_MISSING', 'missing parameters (2x undef) (GetSubProductByAttributeValues)' );

    eval {
        GetSubProductByAttributeValues('String', undef);
    };
    ok( ExistsError() && GetError()->code eq 'FORMAT_NOT_INTEGER', 'first parameter is no integer (GetSubProductByAttributeValues)' );

    eval {
        GetSubProductByAttributeValues(12345, 67890);
    };
    ok( ExistsError() && GetError()->code eq 'PARAMETER_NO_ARRAY', 'second parameter no array (GetSubProductByAttributeValues)' );

    # define test cases
    my $haExpectedResults = {
        'MyJacket' => [
            {
                'SubProduct' => 'JacketRed1',
                'Color'      => 'ExtremeRed',
                'Size'       => 'S',
            },
            {
                'SubProduct' => 'JacketBlack3',
                'Color'      => 'Black',
                'Size'       => 'L',
            }
        ],
    };

    # validate the correctness of GetSubProductByAttributeValues based on the given test shop
    foreach my $ProductAlias ( keys %{ $haExpectedResults } ) {
        my $SuperProduct = $self->shop->child('Products')->child($ProductAlias);
        my $ProductType = $self->shop->child('ProductTypes')->child($SuperProduct->class->alias);

        # match localized variation values with their ids in the database
        my %hProductAttributes;
        foreach my $ProductAttribute ( @{ $ProductType->directAttributes } ) {
            my $ahPreDefAttributes = $ProductAttribute->get('PreDefAttributes');
            foreach my $hPreDefAttribute ( @{ $ahPreDefAttributes } ) {
                $hProductAttributes{$ProductAttribute->alias}{$hPreDefAttribute->alias} = $hPreDefAttribute->id;
            }
        }

        foreach my $hEntry ( @{ $haExpectedResults->{$ProductAlias} } ) {
            my @aVariationString;
            push @aVariationString, $hProductAttributes{'Color'}{$hEntry->{'Color'}} if exists $hEntry->{'Color'};
            push @aVariationString, $hProductAttributes{'Size'}{$hEntry->{'Size'}}   if exists $hEntry->{'Size'};
            my $Exists = ExistsSubProductByAttributeValues($SuperProduct->id, \@aVariationString);
            ok( $Exists, 'sub product having the combination of variations exists' );
            my $SubProduct = GetSubProductByAttributeValues($SuperProduct->id, \@aVariationString);
            is( $SubProduct->alias, $hEntry->{'SubProduct'}, 'verify the sub products found for variations ' . join(',', @aVariationString) );
        }

        # error expected: invalid attribute id
        my $Exists = ExistsSubProductByAttributeValues($SuperProduct->id, [ qw( 12345 ) ]);
        ok( !$Exists, 'given variation does not exist' );

        eval {
            GetSubProductByAttributeValues($SuperProduct->id, [ qw( 12345 ) ]);
        };
        ok( ExistsError() && GetError()->code eq 'SUBPRODUCT_NOT_EXISTS', 'error: variation does not exist' );

        # error expected: too many sub products are matching
        eval {
            my @aVariationString;
            push @aVariationString, $hProductAttributes{'Color'}{'ExtremeRed'};
            GetSubProductByAttributeValues($SuperProduct->id, \@aVariationString);
        };
        ok( ExistsError() && GetError()->code eq 'TOO_MANY_MATCHING_SUBPRODUCTS', 'error: more than one subproduct found' );
    }

    return;
}
sub testGetVariationAttributes {
    my $self = shift;

    # define test cases
    my $hExpectedResults = [
            {
                'name' => 'Size',
                'displayName' => 'Size',
                'value' => 'S',
                'displayValue' => 'Size S',
            },
            {
                'name' => 'Color',
                'displayName' => 'Farbe',
                'value' => 'ExtremeRed',
                'displayValue' => 'Extrem Rot',
            }
        ];

        my $SubProduct1 = $self->shop->path('Products/MyJacket/SubProducts/JacketRed1');
        my $LanguageID = GetPKeyLanguageByCode('de');
        my $hResult = GetVariationAttributes($SubProduct1, $LanguageID);

        is_deeply($hResult, $hExpectedResults, 'Subproduct attributes are correctly retrieved by Language');

        my $SubProduct2 = $self->shop->path('Products/MyJacket/SubProducts/JacketRed2');
        my $SubProduct3 = $self->shop->path('Products/MyJacket/SubProducts/JacketRed3');
        my $SubProduct4 = $self->shop->path('Products/MyJacket/SubProducts/JacketRed4');
        my $SubProduct5 = $self->shop->path('Products/MyJacket/SubProducts/JacketRed5');
        # all five products have the same variation type values
        my $aVariationTypeValues = $SubProduct1->getWithParent('VariationTypeValues', $LanguageID);

        my $StartTime1 = time;
        foreach my $SubProduct ($SubProduct1, $SubProduct2, $SubProduct3, $SubProduct4, $SubProduct5) {
            GetVariationAttributes($SubProduct, $LanguageID, $aVariationTypeValues);
        }
        my $ElapsedTime1 = time - $StartTime1;
        ok($ElapsedTime1 < 0.02, 'Five subproduct attributes retrieve timing (variation type values provided): ' . sprintf("%.3f seconds", $ElapsedTime1));

        my $StartTime2 = time;
        foreach my $SubProduct ($SubProduct1, $SubProduct2, $SubProduct3, $SubProduct4, $SubProduct5) {
            GetVariationAttributes($SubProduct, $LanguageID);
        }
        my $ElapsedTime2 = time - $StartTime2;
        ok($ElapsedTime2 < 0.05, 'Five subproduct attributes retrieve timing (variation type values NOT provided): ' . sprintf("%.3f seconds", $ElapsedTime2));


    return;
}

# compare strings with normalized line breaks
sub _IsEqual($$$) {
    my ($String1, $String2, $Message) = @_;

    foreach ($String1, $String2) {
        s/\x{0D}?\x{0A}/\\x{0A}/gs if defined;
    }
    is( $String1, $String2, $Message );
}

sub testSubProductValueInheritance {
    my $self = shift;

    my $StartUnityProgressMonitorSpy = spy_on('DE_EPAGES::Product::API::Product', 'StartUnityProgressMonitor');
    my $UpdateUnityProgressMonitorSpy = spy_on('DE_EPAGES::Product::API::Product', 'UpdateUnityProgressMonitor');

    my $Shop = $self->shop;
    my $SuperProduct = $Shop->path('Products/MyJacket');
    ok( $SuperProduct->get('IsMaster'), 'SuperProduct.IsMaster' );
    is( $SuperProduct->get('DefaultSubProduct')->alias, 'JacketRed2', 'SuperProduct.DefaultSubProduct' );
    is( $SuperProduct->defaultObject->alias, 'JacketRed2', 'SuperProduct.defaultObject' );
    ok !$SuperProduct->get('IsDefault'), 'SuperProduct.IsDefault';
    ok $SuperProduct->isMaster, 'SuperProduct.isMaster()';
    ok !$SuperProduct->isSubProduct, 'SuperProduct.isSubProduct()';
    is $SuperProduct->countSubProducts, 10, 'SuperProduct.countSubProducts()';

    my $DefaultSubProduct = $SuperProduct->child('SubProducts')->child('JacketRed2');
    ok $DefaultSubProduct->get('IsDefault'), 'DefaultSubProduct.IsDefault';

    my $SubProduct = $SuperProduct->child('SubProducts')->child('JacketBlack3');
    ok( !$SubProduct->get('IsMaster'), 'SubProduct.IsMaster' );
    is( $SubProduct->get('DefaultSubProduct'), undef, 'SubProduct.DefaultSubProduct' );
    is( $SubProduct->defaultObject->id, $SubProduct->id, 'SubProduct.defaultObject' );
    ok !$SubProduct->get('IsDefault'), 'SubProduct.IsDefault';
    ok $SubProduct->isSubProduct, 'SubProduct.isSubProduct()';
    ok !$SubProduct->isMaster, 'SubProduct.isMaster()';
    is $SubProduct->countSubProducts, 0, 'SubProduct.countSubProducts()';

    my $LanguageID = $SuperProduct->getSite->get('LanguageID');
    is( $SuperProduct->get('Text', $LanguageID), undef, 'SuperProduct text is empty' );

    $SuperProduct->set({ 'Text' => "Line0\r\nLine1\nLine2" }, $LanguageID);
    _IsEqual $SuperProduct->get('Text', $LanguageID), "Line0\nLine1\nLine2", 'SuperProduct has multi-line text';
    _IsEqual $SubProduct->getWithParent('Text', $LanguageID), "Line0\nLine1\nLine2", 'SubProduct has inherited text';

    $SubProduct->set({ 'Text' => "Line0\nLine1\nLine2\nLine3" }, $LanguageID);
    _IsEqual $SubProduct->getWithParent('Text', $LanguageID), "Line0\nLine1\nLine2\nLine3", 'SubProduct has oveloaded text';

    $SubProduct->set({ 'Text' => undef }, $LanguageID);
    _IsEqual $SubProduct->getWithParent('Text', $LanguageID), "Line0\nLine1\nLine2", 'SubProduct has again inherited text';

    $SubProduct->set({ 'Text' => "Line0\r\nLine1\r\nLine2" }, $LanguageID); # EPG-31252: different line breaks
    _IsEqual $SubProduct->getWithParent('Text', $LanguageID), "Line0\nLine1\nLine2", 'SubProduct has still inherited text';

    $SuperProduct->set({ 'Text' => 'new text' }, $LanguageID);
    _IsEqual $SubProduct->getWithParent('Text', $LanguageID), 'new text', 'SubProduct uses changed text from master product';

    my $DirectGetColor = sub {
        my ($Product) = @_;
        my $ValueObject = $Product->get('Color');
        return defined $ValueObject ? $ValueObject->alias : 'undef';
    };

    # EPG-31615: sub products don't inherit the value of variation attributes from the master product
    my $AttributeColor = $SubProduct->class->attribute('Color');
    my $ValueBlack = $AttributeColor->preDefAttributeValue('Black');
    ok( $SuperProduct->isVariationAttribute('Color'), 'SuperProduct.isVariationAttribute' );
    ok( !$SubProduct->isVariationAttribute('Color'), 'SubProduct.isVariationAttribute' );
    $SuperProduct->set({ 'Color' => $ValueBlack });
    is( $DirectGetColor->($SuperProduct), 'Black' , 'SuperProduct.get variation attribute');
    is( $SubProduct->getWithParent('Color')->alias // 'undef', 'Black', 'SubProduct.getWithParent variation attribute');
    is( $DirectGetColor->( $SubProduct ), 'Black', 'SubProduct.get variation attribute');
    $SubProduct->set({ 'Color' => $ValueBlack });
    is( $DirectGetColor->( $SubProduct ), 'Black', 'SubProduct value is not deleted');

    # EPG-31745: new sub products don't inherit the value of variation attributes from the master product
    my $NewSuperProduct = $Shop->get('ProductFolder')->insertChild($SuperProduct->class, {
        Alias => 'NewJacket',
        Color => $ValueBlack,
    });
    my $ValueRed = $AttributeColor->preDefAttributeValue('ExtremeRed');
    my $ValueBlue = $AttributeColor->preDefAttributeValue('ExtremeBlue');
    $NewSuperProduct->set({ SelectedVariations => [ { SelectedID => $ValueBlue->id } ] });
    InsertMultipleSubProducts($NewSuperProduct, [{
        Attribute => $AttributeColor,
        Values => [ $ValueBlack, $ValueRed ],
    }], {
        Position => 500,
        Weight => 100.25,
    }, {
        AliasFromAttributeValueAlias => 1,
        Monitor => {
            Servlet => {}
        }
    });
    is ($StartUnityProgressMonitorSpy->calls_count(), 1, 'StartUnityProgressMonitor was called once');
    is (scalar @{$StartUnityProgressMonitorSpy->calls_most_recent()->[2]}, 2, 'StartUnityProgressMonitor was called with a list count of 2');
    is ($UpdateUnityProgressMonitorSpy->calls_count(), 2, 'UpdateUnityProgressMonitorSpy was called 2 times');
    my $SubProducts = $NewSuperProduct->child('SubProducts');
    is( scalar @{$SubProducts->children}, 2, 'InsertMultipleSubProducts created variations' );
    ok( $SubProducts->existsChild($_), "created variation $_" ) foreach qw( NewJacketBlack NewJacketExtremeRed );
    my $NewJacketBlack = $SubProducts->child('NewJacketBlack');
    is( $DirectGetColor->( $NewJacketBlack ), 'Black', 'SubProduct value is set directly');
    is $NewJacketBlack->get('Position'), 500, 'InsertMultipleSubProducts: uses Position attribute';
    is $NewJacketBlack->get('Weight'), 100.25, 'InsertMultipleSubProducts: uses Weight attribute';
    is scalar @{$NewSuperProduct->get('SelectedVariations')}, 3, 'Adding sub products does not remove selected values'; # should be Blue, Red and Black

    my $FindSubProductByValue = sub {
        my ($Master, $Value) = @_;
        my $Sub = eval{ GetSubProductByAttributeValues( $Master->id, [ $Value->id ] ) };
        Test::More::diag GetError()->message if ExistsError();
        return defined $Sub ? 1 : 0;
    };
    ok( $FindSubProductByValue->( $NewSuperProduct, $ValueRed ), 'find red' );
    ok( $FindSubProductByValue->( $NewSuperProduct, $ValueBlack ), 'find black' );

    # cleanup
    $NewSuperProduct->delete;
    $SuperProduct->set({ 'Color' => undef });
    $SuperProduct->set({ 'Text' => undef }, $LanguageID);
}

sub testUnitySubProductValuesInheritance {
    my $self = shift;

    my $Shop = $self->shop;
    $Shop->set({'IsUnityRegistered' => 1});

    my $SuperProduct = $Shop->path('Products/MyJacket');
    ok( $SuperProduct->get('IsMaster'), 'SuperProduct.IsMaster' );

    my $SubProduct = $SuperProduct->child('SubProducts')->child('JacketRed2');

    # write attribute to master and sub product and test if eq (no interitance)
    # see: Cartridges/DE_EPAGES/Product/API/Constants.pm @UNITY_ATTRIBUTES_FROM_MASTERPRODUCT_WRITTEN_ON_SUBS_ON_CREATE
    $SuperProduct->set({ 'Manufacturer' => "ePages" });
    _IsEqual $SubProduct->get('Manufacturer'), "ePages", 'SubProduct has correct non inherited Manufacturer';

    # set attribute on both products because we do not update this attribute
    # see: Cartridges/DE_EPAGES/Product/API/Constants.pm @UNITY_NOT_UPDATE_ATTRIBUTES_ON_SUBS_ON_UPDATE_MASTER
    $SuperProduct->set({ 'Weight' => "24" });
    $SubProduct->set({ 'Weight' => "24" });
    _IsEqual $SubProduct->get('Weight'), "24", 'SubProduct has correct non inherited Weight';

    $SuperProduct->set({ 'Weight' => "124" });
    ok( $SubProduct->get('Weight') != $SuperProduct->get('Weight') , 'SubProduct has different Weight' );


    # test inheritance of localized attributes
    my $LanguageID = $SuperProduct->getSite->get('LanguageID');
    is( $SuperProduct->get('Keywords', $LanguageID), undef, 'SuperProduct Keywords are empty' );
    is( $SubProduct->get('Keywords', $LanguageID), undef, 'SubProduct Keywords are empty' );

    $SuperProduct->set({ 'Keywords' => "Super Product Keywords" }, $LanguageID);
    _IsEqual $SubProduct->get('Keywords', $LanguageID), undef, 'SubProduct has correct non inherited Keywords';

    # reset flag for unity
    $Shop->set({'IsUnityRegistered' => 0});
}

sub testUsedVariationValues {
    my $self = shift;

    my $Shop = $self->shop;
    # check erroneously passed parameters
    eval {
        UsedVariationValues();
    };
    ok( ExistsError() && GetError()->code eq 'PARAMETER_MISSING', 'missing parameter (UsedVariationValues)' );

    eval {
        UsedVariationValues('String');
    };
    ok( ExistsError() && GetError()->code eq 'PARAMETER_NO_OBJECT', 'wrong parameter passed (UsedVariationValues)' );

    eval {
        UsedVariationValues(undef);
    };
    ok( ExistsError() && GetError()->code eq 'PARAMETER_MISSING', 'undefined parameter passed (UsedVariationValues)' );

    # validate the correctness of UsedVariationValues based on the given test shop
    my $SuperProduct = $Shop->path('Products/MyJacket');
    CompareUsedVariationValues($SuperProduct, [ qw( ExtremeRed Black S M L XL XXL ) ] );

    $SuperProduct = $Shop->path('Products/DownloadProduct');
    CompareUsedVariationValues($SuperProduct, [ qw( Deutsch Englisch French ) ] );

    $SuperProduct = $Shop->path('Products/MyPetrol');
    CompareUsedVariationValues($SuperProduct, [ () ] );

    return;
}

sub testDisplayTitle {
    my $self = shift;

    my $Shop = $self->shop;
    my $SuperProduct = $Shop->path('Products/MyJacket');
    my $SubProduct = $SuperProduct->child('SubProducts')->child('JacketBlack5');
    my $EN = GetPKeyLanguageByCode('en');
    my $OldShopName = $Shop->get('Name', $EN);

    $Shop->set({'Name' => 'Shop Name'}, $EN);

    # product without name and title
    is( $SuperProduct->getWithParent('Title', $EN), undef, 'Initial SuperProduct Title' );
    is( $SuperProduct->getWithParent('DisplayTitle', $EN), 'MyJacket - Shop Name', 'Initial SuperProduct DisplayTitle' );
    is( $SubProduct->getWithParent('Title', $EN), undef, 'Initial SubProduct Title' );
    is( $SubProduct->getWithParent('DisplayTitle', $EN), 'JacketBlack5 - Shop Name', 'Initial SubProduct DisplayTitle' );

    # super product with name, but no title
    $SuperProduct->set({'Name' => 'New SuperProduct Name'}, $EN);
    is( $SuperProduct->getWithParent('Title', $EN), undef, 'SuperProduct.Title (New SuperProduct Name)' );
    is( $SuperProduct->getWithParent('DisplayTitle', $EN), 'New SuperProduct Name - Shop Name', 'SuperProduct.DisplayTitle (New SuperProduct Name)' );
    is( $SubProduct->getWithParent('Title', $EN), undef, 'SubProduct.Title (New SuperProduct Name)' );
    is( $SubProduct->getWithParent('DisplayTitle', $EN), 'New SuperProduct Name - Shop Name', 'SubProduct.DisplayTitle (New SuperProduct Name)' );

    # super product with title, sub product without title
    $SuperProduct->set({'Title' => 'New SuperProduct Title'}, $EN);
    is( $SuperProduct->getWithParent('Title', $EN), 'New SuperProduct Title', 'SuperProduct.Title (New SuperProduct Title)' );
    is( $SuperProduct->getWithParent('DisplayTitle', $EN), 'New SuperProduct Title', 'SuperProduct.DisplayTitle (New SuperProduct Title)' );
    is( $SubProduct->getWithParent('Title', $EN), 'New SuperProduct Title', 'SubProduct.Title (New SuperProduct Title)' );
    is( $SubProduct->getWithParent('DisplayTitle', $EN), 'New SuperProduct Title', 'SubProduct.DisplayTitle (New SuperProduct Title)' );

    # sub product with title
    $SubProduct->set({'Title' => 'New SubProduct Title'}, $EN);
    is( $SuperProduct->getWithParent('DisplayTitle', $EN), 'New SuperProduct Title', 'SuperProduct.DisplayTitle (New SubProduct Title)' );
    is( $SubProduct->getWithParent('Title', $EN), 'New SubProduct Title', 'SubProduct.Title (New SuperProduct Title)' );
    is( $SubProduct->getWithParent('DisplayTitle', $EN), 'New SubProduct Title', 'SubProduct.DisplayTitle (New SuperProduct Title)' );

    # product name = shop name
    $SuperProduct->set({'Title' => undef}, $EN);
    $SubProduct->set({'Title' => undef, 'Name' => 'Shop Name'}, $EN);
    is( $SubProduct->getWithParent('DisplayTitle', $EN), 'Shop Name', 'Product Name = Shop Name' );

    # undefined shop name
    $Shop->set({'Name' => undef}, $EN);
    $SubProduct->set({'Name' => 'Sub Product Name'}, $EN);
    is( $SubProduct->getWithParent('DisplayTitle', $EN), 'Sub Product Name', 'Undefined Shop Name' );
    is( $SuperProduct->getWithParent('DisplayTitle', $EN), 'New SuperProduct Name', 'Undefined Shop Name' );

    # cleanup
    $Shop->set({'Name' => $OldShopName}, $EN);
    $SubProduct->set({'Title' => undef}, $EN);
    $SuperProduct->set({'Title' => undef}, $EN);
}

sub CompareUsedVariationValues {
    my ($SuperProduct, $aExpectedVariationValues) = @_;

    my $aGotPreDefAttributeIDs = UsedVariationValues($SuperProduct);
    my @aGotVariationValues;
    push @aGotVariationValues, LoadObject($_)->alias foreach @{ $aGotPreDefAttributeIDs };
    cmp_deeply( \@aGotVariationValues, bag(@{ $aExpectedVariationValues }), 'checked used variations for ' . $SuperProduct->alias );

    return;
}

sub testReferencePrice {
    my $self = shift;

    my $Shop = $self->shop;
    my $Product = $Shop->path('Products/ReferencePriceProduct');
    ok( !$Product->get('IsDownloadProduct'), 'not Product.IsDownloadProduct' );
    ok( !$Product->get('IsAppointment'), 'not Product.IsAppointment' );
    ok( !$Product->get('IsBundleProduct'), 'not Product.IsBundleProduct' );

    is ( $Product->getReferencePrice('EUR', 12), 2.4, 'verify reference price');
    is ( $Product->getReferencePrice('EUR', 100, 12), 17.6, 'verify reference price with discount');
    is ( $Product->getReferencePrice('EUR'), undef, 'verify reference price function');

    $Product = $Shop->path('Products/MyJacket');
    $Product->set({ RefAmount => undef , RefContentAmount => 10 });
    is $Product->getReferencePrice('EUR', 100), undef, 'no RefAmount';
    $Product->set({ RefAmount => 10 , RefContentAmount => undef });
    is $Product->getReferencePrice('EUR', 100), undef, 'no RefContentAmount';
    $Product->set({ RefAmount => 1 , RefContentAmount => 1 });
    is $Product->getReferencePrice('EUR', 100.123), 100.12, 'RefAmount == RefContentAmount';
    is $Product->getReferencePrice('BHD', 100.123), 100.123, 'Currency.MinorUnit=3';
    is $Product->getReferencePrice('BIF', 100.123), 100, 'Currency.MinorUnit=0';

    return;
}

sub testDisplayPrice {
    my $self = shift;

    my $Shop = $self->shop;
    my $SuperProduct = $Shop->path('Products/MyJacket');
    my $hPriceOptions = {
        CurrencyID => 'EUR',
        TaxModel => TAX_MODEL_GROSS,
    };
    is $SuperProduct->displayPrice($hPriceOptions), '199.95', 'displayPrice';
    is $SuperProduct->fromPrice($hPriceOptions), '189.95', 'fromPrice';
    is $SuperProduct->getListPrice($hPriceOptions->{'CurrencyID'}, TAX_MODEL_GROSS ), '199.95', 'getListPrice gross';
    is $SuperProduct->getListPrice($hPriceOptions->{'CurrencyID'}, TAX_MODEL_NET ), '168.03', 'getListPrice net';
    is $SuperProduct->getListPrice($hPriceOptions->{'CurrencyID'}, TAX_MODEL_NET, 1 ), '168.025', 'getListPrice net, extra precision';
    return;
}

sub testHasPrice {
    my $self = shift;

    my $CurrencyID = 'EUR';
    my $Shop = $self->shop;
    my $SuperProduct = $Shop->path('Products/Shirt');
    my $SuperListPrice = $SuperProduct->getDirectListPrice($CurrencyID, TAX_MODEL_NET, 1);
    my $SubProduct = $SuperProduct->child('SubProducts')->child('Shirt blue');
    my $SubListPrice = $SubProduct->getDirectListPrice($CurrencyID, TAX_MODEL_NET, 1);

    # check preconditions
    isnt( $SuperListPrice, undef, 'testHasPrice: Master product has a price');
    is( $SubListPrice, undef, 'testHasPrice: Sub product has no price');

    ok( $SuperProduct->hasPrice($CurrencyID), 'SuperProduct.hasPrice (price from master)' );
    ok( $SubProduct->hasPrice($CurrencyID), 'SubProduct.hasPrice (price from master)' );

    $SuperProduct->setListPrice($CurrencyID, TAX_MODEL_NET, undef);
    ok( !$SuperProduct->hasPrice($CurrencyID), 'SuperProduct.hasPrice (no price)' );
    ok( !$SubProduct->hasPrice($CurrencyID), 'SubProduct.hasPrice (no price)' );

    $SubProduct->setListPrice($CurrencyID, TAX_MODEL_NET, 9.99);
    $SubProduct->set({'HasSubOwnPrices' => 1});
    is( $SubProduct->getDirectListPrice($CurrencyID, TAX_MODEL_NET), 9.99, 'SubProduct.getDirectListPrice' );

    ok( $SubProduct->hasPrice($CurrencyID), 'SubProduct.hasPrice (price from sub)' );
    ok( $SuperProduct->hasPrice($CurrencyID), 'SuperProduct.hasPrice (price from sub)' );

    # cleanup
    $SubProduct->set({'HasSubOwnPrices' => 0});
    $SubProduct->setListPrice($CurrencyID, TAX_MODEL_NET, undef);
    $SuperProduct->setListPrice($CurrencyID, TAX_MODEL_NET, $SuperListPrice);
    return;
}

sub testSubProductWithLowestPrice {
    my $self = shift;

    my $Shop = $self->shop;
    my $hPriceOptions = {
        CurrencyID => 'GBP',
        TaxModel => TAX_MODEL_GROSS,
    };

    my $SuperProduct = $Shop->path('Products/Shirt');
    my $SubProduct = $SuperProduct->subProductWithLowestPrice($hPriceOptions);
    is $SubProduct, undef, 'no variations with own prices';

    $SuperProduct = $Shop->path('Products/MyJacket');
    $SubProduct = $SuperProduct->subProductWithLowestPrice($hPriceOptions);
    is $SubProduct->alias, 'JacketBlack4', 'subProductWithLowestPrice';
}

sub testLowestSubProductPrice {
    my $self = shift;

    my $Shop = $self->shop;
    my $hPriceOptions = {
        CurrencyID => 'GBP',
        TaxModel => TAX_MODEL_GROSS,
    };

    my $SuperProduct = $Shop->path('Products/Shirt');
    my $LowestPrice = $SuperProduct->lowestSubProductPrice(
            $hPriceOptions->{'CurrencyID'}, $hPriceOptions->{'TaxModel'}
    );
    is($LowestPrice, undef, 'undefined lowest price when no variations with own prices');

    $SuperProduct = $Shop->path('Products/MyJacket');
    $LowestPrice = $SuperProduct->lowestSubProductPrice(
            $hPriceOptions->{'CurrencyID'}, $hPriceOptions->{'TaxModel'}
    );
    is($LowestPrice, 175.00, 'lowest price gotten from variations (GBP)');

    $LowestPrice = $SuperProduct->lowestSubProductPrice('EUR', $hPriceOptions->{'TaxModel'});
    is($LowestPrice, 189.95, 'lowest price gotten from variations (EUR)');
}

sub testSlideshow {
    my $self = shift;

    my $Shop = $self->shop;
    my $SuperProduct = $Shop->path('Products/Shirt');
    my $SubProduct = $SuperProduct->child('SubProducts')->child('Shirt blue');

    # prepare some images
    my $PublicPath = $SuperProduct->get('PublicPath');
    my $Content = 'image';
    WriteFile("$PublicPath/green.png", \$Content, { create_directory => 1 });
    WriteFile("$PublicPath/blue.png", \$Content, { create_directory => 1 });
    WriteFile("$PublicPath/red.png", \$Content, { create_directory => 1 });

    ok( !$SuperProduct->get('UseSlideshow'), 'SuperProduct has no slideshow yet' );
    ok( !$SubProduct->get('UseSlideshow'), 'SubProduct has no slideshow yet' );

    $SuperProduct->set({'ImagesSlideShowString' => 'green.png;blue.png;grey.png'});
    is( $SuperProduct->get('ImagesSlideShowString'), 'green.png;blue.png;grey.png', 'SuperProduct.ImagesSlideShowString');
    is( scalar @{$SuperProduct->get('ImagesSlideShow')}, 3, 'number of slideshow images of the super product' );
    is_deeply( $SuperProduct->get('ImagesSlideShow'), [
          {
            'Position' => 0,
            'Sizes' => {
                         'MediumSmall' => undef,
                         'Thumbnail' => undef,
                         'MediumLarge' => undef,
                         'HotDeal' => undef,
                         'Small' => undef,
                         'Medium' => undef,
                         'Large' => 'green.png'
                       },
            'ImageName' => 'green.png'
          },
          {
            'Position' => 10,
            'Sizes' => {
                         'MediumSmall' => undef,
                         'Thumbnail' => undef,
                         'MediumLarge' => undef,
                         'HotDeal' => undef,
                         'Small' => undef,
                         'Medium' => undef,
                         'Large' => 'blue.png'
                       },
            'ImageName' => 'blue.png'
          },
          {
            'Position' => 20,
            'Sizes' => {
                         'MediumSmall' => undef,
                         'Thumbnail' => undef,
                         'MediumLarge' => undef,
                         'HotDeal' => undef,
                         'Small' => undef,
                         'Medium' => undef,
                         'Large' => undef
                       },
            'ImageName' => 'grey.png'
          }
        ], 'SuperProduct.ImagesSlideShow');
    ok( $SuperProduct->get('UseSlideshow'), 'SuperProduct now has a slideshow' );
    ok( $SubProduct->get('UseSlideshow'), 'SubProduct uses the slideshow of the super product' );
    is( scalar @{$SuperProduct->get('ImagesSlideShow')}, 3, 'number of slideshow images of the super product' );

    $SubProduct->set({'ImageLarge' => 'blue.png'});
    ok( $SuperProduct->get('UseSlideshow'), 'SuperProduct still has a slideshow' );
    ok( $SubProduct->get('UseSlideshow'), 'SubProduct use a slideshow' );
    is( $SubProduct->get('ImagesSlideShowString'), 'blue.png', 'correct slideshow string' );
    $SubProduct->set({'ImageLarge' => 'blue2.png'});
    is( $SubProduct->get('ImagesSlideShowString'), 'blue2.png;blue.png', 'correct slideshow string' );


    $SubProduct->set({'ImageLarge' => undef, 'ImageMedium' => 'blue.png'});
    ok( $SuperProduct->get('UseSlideshow'), 'SuperProduct still has a slideshow' );
    ok( $SubProduct->get('UseSlideshow'), 'SubProduct still use a slideshow' );

    $SubProduct->set({'ImagesSlideShowString' => 'blue.png;red.png'});
    is( scalar @{$SubProduct->get('ImagesSlideShow')}, 2, 'number of slideshow images of the sub product' );
    ok( $SuperProduct->get('UseSlideshow'), 'SuperProduct still has a slideshow' );
    ok( $SubProduct->get('UseSlideshow'), 'SubProduct now has its own slideshow' );

    # cleanup
    $SubProduct->set({'ImagesSlideShowString' => undef});
    $SubProduct->set({'ImageLarge' => undef, 'ImageMedium' => undef});
    $SuperProduct->set({'ImagesSlideShowString' => undef});
}

sub testImportImageAssignments {
    my $self = shift;

    my $Shop = $self->shop;
    # EPG-30821
    # test correct image assignment in CVS import by checking pixel colors
    my $ImagePath = $Shop->get('MediaGallery')->get('PublicPath') . '/Import';

    my @Images = (qw(ff0000.png 00ff00.png 0000ff.png ff00ff.png));
    foreach my $Image (@Images) {
        CopyFile($Image, "$ImagePath/$Image", { 'create_directory' => 1 } );
    }

    my ($aWarnings) = ImportCSV($Shop, 'images.csv', 'DE_EPAGES::Product::API::CSV::ProductExportImport',
        { 'sep_char' => ';' });
    ok(scalar(@$aWarnings) == 0, 'CSV import generates no warning');

    my $Products;
    my $ProductFolder = $Shop->folder('Products');
    my @ProductAliases = (qw(
        alle_bilder
        listenansicht
        detailseite
        vergroessert
        aktionsprodukte
        listenansicht_detailseite
        vergroessert_aktionsprodukte
        listenansicht_aktionsprodukte
    ));

    my $ProductIterator = $ProductFolder->childrenIterator();
    while (my $Product = <$ProductIterator>) {
        $Products->{$Product->alias} = $Product if ElementOf($Product->alias, @ProductAliases);
    }

    my $ExpectedColors = {
        'alle_bilder' => {
            'ImageSmall'       => [1,0,0], # Listenansicht
            'ImageMedium'      => [0,1,0], # Detailseite
            'ImageMediumSmall' => [0,1,0], # ??
            'ImageMediumLarge' => [0,1,0], # ??
            'ImageLarge'       => [0,0,1], # Vergrößert
            'ImageHotDeal'     => [1,0,1], # Aktionsprodukte
        },

        'listenansicht' => {
            'ImageSmall'       => [1,0,0],
            'ImageMedium'      => [1,0,0],
            'ImageMediumSmall' => [1,0,0],
            'ImageMediumLarge' => [1,0,0],
            'ImageLarge'       => [1,0,0],
            'ImageHotDeal'     => [1,0,0],
        },

        'detailseite' => {
            'ImageSmall'       => [0,1,0],
            'ImageMedium'      => [0,1,0],
            'ImageMediumSmall' => [0,1,0],
            'ImageMediumLarge' => [0,1,0],
            'ImageLarge'       => [0,1,0],
            'ImageHotDeal'     => [0,1,0],
        },

        'vergroessert' => {
            'ImageSmall'       => [0,0,1],
            'ImageMedium'      => [0,0,1],
            'ImageMediumSmall' => [0,0,1],
            'ImageMediumLarge' => [0,0,1],
            'ImageLarge'       => [0,0,1],
            'ImageHotDeal'     => [0,0,1],
        },

        'aktionsprodukte' => {
            'ImageSmall'       => [1,0,1],
            'ImageMedium'      => [1,0,1],
            'ImageMediumSmall' => [1,0,1],
            'ImageMediumLarge' => [1,0,1],
            'ImageLarge'       => [1,0,1],
            'ImageHotDeal'     => [1,0,1],
        },

        'listenansicht_detailseite' => {
            'ImageSmall'       => [1,0,0],
            'ImageMedium'      => [0,1,0],
            'ImageMediumSmall' => [0,1,0],
            'ImageMediumLarge' => [0,1,0],
            'ImageLarge'       => [0,1,0],
            'ImageHotDeal'     => [0,1,0],
        },

        'vergroessert_aktionsprodukte' => {
            'ImageSmall'       => [0,0,1],
            'ImageMedium'      => [0,0,1],
            'ImageMediumSmall' => [0,0,1],
            'ImageMediumLarge' => [0,0,1],
            'ImageLarge'       => [0,0,1],
            'ImageHotDeal'     => [1,0,1],
        },

        'listenansicht_aktionsprodukte' => {
            'ImageSmall'       => [1,0,0],
            'ImageMedium'      => [1,0,0],
            'ImageMediumSmall' => [1,0,0],
            'ImageMediumLarge' => [1,0,0],
            'ImageLarge'       => [1,0,0],
            'ImageHotDeal'     => [1,0,1],
        }
    };

    foreach my $ProductAlias (keys %$Products) {
        my $Product = $Products->{$ProductAlias};
        my $Colors = $ExpectedColors->{$ProductAlias};
        foreach my $AttrName (keys %$Colors) {
            if(defined($Colors->{$AttrName})) {
                die "$ProductAlias\[$AttrName\] is undefined" unless(defined($Product->get($AttrName)));
                my $ImageFile = $Product->get('PublicPath') . '/' . $Product->get($AttrName);

                my $Image = Image::Magick->new;
                my $ImageError = $Image->Read($ImageFile);
                die $ImageError if $ImageError;

                my ($r, $g, $b) = $Image->GetPixel(x => 15, y => 15); # center pixel is sufficient
                my ($er, $eg, $eb) = @{$Colors->{$AttrName}};
                ok($r == $er && $g == $eg && $b == $eb, "$ProductAlias\[$AttrName\] is correct image");
            }
        }
    }

    foreach my $Image (@Images) {
        DeleteFile("$ImagePath/$Image");
    }
    return;
}

sub testAvailabilityText {
    my $self = shift;

    my $Shop = $self->shop;
    my $Appointment = $Shop->path('Appointments/Table4Person');
    my $LanguageID = GetPKeyLanguageByCode('en');
    ok(!$Appointment->get('AvailabilityText', $LanguageID), 'AvailabilityText is null for appointments');

    my $Product = $Shop->path('Products/MyPetrol');
    $Product->set({'StockLevel' => 0 });

    is($Product->get('AvailabilityText', $LanguageID), 'StockLevel 0', 'getting availability value for out of stock.');
    is($Product->get('AvailabilityIcon', $LanguageID), 'ProductOutStockIcon', 'getting availability icon for out of stock.');
    return;
}

sub testResetMasterFlag {
    my $self = shift;

    my $Shop = $self->shop;
    my $MaxProducts = $Shop->featureMaxValue('Products');
    $Shop->featureMaxValue('Products', 25);

    my $ProductType = $Shop->get('ProductTypes')->child('Jacket');
    my $Product = $Shop->get('ProductFolder')->insertChild($ProductType, {
        Alias => 'TestMaster'
    });
    ok !$Product->isMaster, 'new product is not a master';

    my $Attribute = $ProductType->attribute('Color');
    my @VariationValues = map { $Attribute->preDefAttributeValue($_) } qw (Black ExtremeRed);
    $Product->set({ SelectedVariations => [ map { { 'SelectedID' => $_->id } } @VariationValues ] });
    ok $Product->isMaster, 'product with variation attributes is a master';
    is $Product->countSubProducts, 0, 'master has no sub produts yet';

    my $aVariationAttributes = $Product->variationAttributes;
    is scalar(@$aVariationAttributes), 1, 'Product.variationAttributes count';
    is $aVariationAttributes->[0]->alias, 'Color', 'Product.variationAttributes[0]';

    $Product->set({ SelectedVariations => [] });
    ok !$Product->isMaster, 'product without variation attributes is not a master';

    $Shop->featureMaxValue('Products', $MaxProducts);
    return;
}

sub testGetUnityPrettyUrl {
    my $self = shift;
    my $Shop = $self->shop;

    my $Product = $self->shop->path('Products/MyJacket');
    my $ShopLanguageID = $Shop->get('LanguageID');
    my $RelyOnStoredPrettyUrl = 1;

    # do this test for a base shop
    $Shop->set({'IsUnityRegistered' => 0});

    eval {
        GetUnityPrettyUrl($Product, $ShopLanguageID, $ShopLanguageID, $RelyOnStoredPrettyUrl);
    };
    ok( ExistsError() && GetError()->code eq 'NO_UNITY_SHOP', 'can not get pretty url from a non unity shop (GetUnityPrettyUrl)' );
}

sub testCSVImport {
    my $self = shift;
    my $Shop = $self->shop;

    # do this tests for an unity shop
    $Shop->set({'IsUnityRegistered' => 1});
    $Shop->featureMaxValue( 'VariationAttributeValues', 20 );

    # delete all products/categories/producttypes
    $Shop->folder('Products')->deleteChildren();
    $Shop->folder('Categories')->deleteChildren();
    $Shop->folder('ProductTypes')->deleteChildren();

    my $aWarnings = [];

    # import producttypes first
    ($aWarnings) = ImportCSV($Shop, 'ProductTypes_unity.csv', 'DE_EPAGES::Product::API::CSV::ProductTypeExportImport',
        { 'sep_char' => ';' });
    ok(scalar(@$aWarnings) == 0, 'CSV import producttypes generates no warning');

    # import products
    ($aWarnings) = ImportCSV($Shop, 'Products_unity.csv', 'DE_EPAGES::Product::API::CSV::ProductExportImport',
        { 'sep_char' => ';' });
    ok(scalar(@$aWarnings) == 0, 'CSV import products generates no warning');

    # export producttypes and products and compare export with import file
    ($aWarnings) = ExportCSV($Shop, 'ProductTypes_unity_export.csv', 'DE_EPAGES::Product::API::CSV::ProductTypeExportImport',
        { 'sep_char' => ';' }, $Shop->folder('ProductTypes')->children());
    ok(scalar(@$aWarnings) == 0, 'CSV export producttypes generates no warning');
    is(CompareMD5TextFile('ProductTypes_unity_export.csv',GetFileMD5Digest('ProductTypes_unity.csv')),1,'ProductTypes_unity.csv is equal to ProductTypes_unity_export.csv');
    DeleteFile('ProductTypes_unity_export.csv');

    ($aWarnings) = ExportCSV($Shop, 'Products_unity_export.csv', 'DE_EPAGES::Product::API::CSV::ProductExportImport',
        { 'sep_char' => ';' }, $Shop->folder('Products')->children());
    ok(scalar(@$aWarnings) == 0, 'CSV export products generates no warning');
    is(CompareMD5TextFile('Products_unity_export.csv',GetFileMD5Digest('Products_unity.csv')),1,'Products_unity.csv is equal to Products_unity_export.csv');
    DeleteFile('Products_unity_export.csv');

}


# be sure having reseted IsUnityRegistered
# otherwise we can not delete the shop because of remot connection to the SF does not exists
sub cleanup {
    my $self = shift;

    return if $ENV{'SKIP_CLEANUP'};

    my $System = LoadRootObject();
    if( $System->existsChild($self->testFolder) ) {
        my $Folder = $System->child($self->testFolder);
        $Folder->child($self->shopAlias)->set({'IsUnityRegistered' => 0}) if $Folder->existsChild($self->shopAlias);
    }

    # call the real cleanup
    $self->SUPER::cleanup;
}

sub init {
    my $self = shift;

    $self->SUPER::init;
    die 'Shop alias is not TestShop' unless $self->shop->alias eq 'TestShop';

    note 'importing import.xml';
    ImportObjects('import.xml', $self->shop->parent->pathString);
}

sub test {
    my $self = shift;

    $self->testSlideshow();
    $self->testSubProductWithLowestPrice();
    $self->testLowestSubProductPrice();
    $self->testHasPrice();
    $self->testDisplayPrice();
    $self->testSubProductValueInheritance();
    $self->testUnitySubProductValuesInheritance();
    $self->testDisplayTitle();
    $self->testBundle();
    $self->testDownload();
    $self->testAppointment();
    $self->testSmartSearch();
    $self->testGetSubProductByAttributeValues();
    $self->testGetVariationAttributes();
    $self->testUsedVariationValues();
    $self->testReferencePrice();
    $self->testImportImageAssignments();
    $self->testAvailabilityText();
    $self->testResetMasterFlag();
    $self->testGetUnityPrettyUrl();
    $self->testCSVImport();
    done_testing();
}

__PACKAGE__->runScript;