#========================================================================================
# package      DE_EPAGES::WebInterfacePoolDB::API::Cartridge
# state        public
#----------------------------------------------------------------------------------------
# description  main cartridge class for install/patch/uninstall
#========================================================================================
package DE_EPAGES::WebInterfacePoolDB::API::Cartridge;
use base DE_EPAGES::Trigger::API::TriggerInstaller;

use strict;
use DE_EPAGES::Core::API::Constants;
use DE_EPAGES::Core::API::EpagesConfig qw( EpagesConfig );
use DE_EPAGES::Core::API::IDN qw( EncodeDomain );
use DE_EPAGES::WebInterfacePoolDB::API::Table::Host qw( ExistsHostByName GetPKeyHostByName InsertHost );
use DE_EPAGES::WebInterfacePoolDB::API::Table::Pool qw( ExistsPoolByName GetPKeyPoolByName InsertPool );
use DE_EPAGES::WebInterfacePoolDB::API::Table::Request qw( ExistsRequestByHostID InsertRequest );
use DE_EPAGES::WebInterfacePoolDB::API::Table::Site qw( ExistsSiteByGUID GetPKeySiteByGUID InsertSite );
use DE_EPAGES::WebInterfacePoolDB::API::Table::Store qw( ExistsStoreByName GetPKeyStoreByName InsertStore );

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Cartridge = DE_EPAGES::Object::API::Cartridge->new(%options);
#----------------------------------------------------------------------------------------
# description  creates a new installer object of DE_EPAGES::Object
#----------------------------------------------------------------------------------------
# input        %options | options for un/install keys
#               <ul>
#                 <li>IsRecursive - recurive un/installation of required or dependent cartridges
#               </ul>
#               | hash
# return       $Cartridge | cartridge object | DE_EPAGES::WebInterfacePoolDB::API::Cartridge
#========================================================================================
sub new {
    my ($class, %options) = @_;
    my $self = __PACKAGE__->SUPER::new(
        %options,
        'CartridgeDirectory'=> 'DE_EPAGES/WebInterfacePoolDB',
        'Version'            => '7.37.0',
        'Patches'           => ['6.15.1','6.15.2','6.15.3','6.15.4','6.16.0','6.16.1','6.16.2','6.16.3','6.17.0','6.17.1','6.17.2','6.17.3','6.17.4','6.17.5','6.17.6','6.17.7','6.17.8','6.17.9','6.17.9.1','6.17.9.2','6.17.9.3','6.17.10','6.17.11','6.17.12','6.17.12.1','6.17.12.2','6.17.13','6.17.13.1','6.17.13.2','6.17.14','6.17.14.1','6.17.15','6.17.15.1','6.17.16','6.17.17','6.17.17.1','6.17.18','6.17.19','6.17.20','6.17.20.1','6.17.21','6.17.21.1','6.17.22','6.17.23','6.17.24','6.17.24.1','6.17.25','6.17.26','6.17.26.1','6.17.27','6.17.27.1','6.17.28','6.17.28.1','6.17.29','6.17.29.1','6.17.30','6.17.31','6.17.32','6.17.33','6.17.33.1','6.17.34','6.17.34.1','6.17.35','6.17.36','6.17.36.1','6.17.37','6.17.38','6.17.38.1','6.17.39','6.17.39.1','6.17.40','6.17.40.1','6.17.41','6.17.41.1','6.17.42','6.17.43','6.17.43.1','6.17.44','6.17.45','6.17.45.1','6.17.46','6.17.47','6.17.47.1','6.17.48','6.17.48.1','6.17.49','6.17.49.1','6.17.50','6.17.51','6.17.51.1','6.17.52','6.17.53','7.0.0','7.1.0','7.2.0','7.3.0','7.4.0','7.5.0','7.6.0','7.7.0','7.8.0','7.9.0','7.10.0','7.11.0','7.12.0','7.13.0','7.14.0','7.15.0','7.16.0','7.17.0','7.18.0','7.19.0','7.20.0','7.21.0','7.22.0','7.23.0','7.24.0','7.25.0','7.26.0','7.27.0','7.28.0','7.29.0','7.30.0','7.31.0','7.32.0','7.33.0','7.34.0','7.35.0','7.36.0'],
    );
    return bless $self, $class;
}

#========================================================================================
# function     install
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->install;
#----------------------------------------------------------------------------------------
# description  install the cartridge
#========================================================================================
sub install {
    my $self = shift;
    $self->startInstall;
    $self->SUPER::install;

    $self->_insertTestRequest;

    $self->finishInstall;
}

# EPG-17038: add initial data required for Test Request
sub _insertTestRequest {
    my $self = shift;

    my $HostName = EncodeDomain( EpagesConfig()->get('DE_EPAGES::Object', 'SystemDomainName', 'localhost') );
    my $HostID = ExistsHostByName( $HostName )
        ? GetPKeyHostByName( $HostName )
        : InsertHost( { 'Name' => $HostName } );

    my $PoolName = 'DefaultPool';
    my $PoolID = ExistsPoolByName( $PoolName )
        ? GetPKeyPoolByName( $PoolName )
        : InsertPool( { 'Name' => $PoolName } );

    my $StoreName = 'TestRequest';
    my $StoreID = ExistsStoreByName( $StoreName )
        ? GetPKeyStoreByName( $StoreName )
        : InsertStore( { 'Name' => $StoreName, 'PoolID' => $PoolID } );

    my $SiteGUID = 'TestRequest';
    my $SiteID = ExistsSiteByGUID( $SiteGUID )
        ? GetPKeySiteByGUID( $SiteGUID )
        : InsertSite( { 'GUID' => $SiteGUID, 'StoreID' => $StoreID, 'ObjectPath' => '/' } );

    my $URI = '/epages/TestRequest.';
    if( !ExistsRequestByHostID( $HostID, $URI ) ) {
        InsertRequest({ 'SiteID' => $SiteID, 'HostID' => $HostID, 'URI' => $URI });
    }

    return;
}

#========================================================================================
# function     uninstall
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->uninstall;
#----------------------------------------------------------------------------------------
# description  uninstall the cartridge
#========================================================================================
sub uninstall {
    my $self = shift;
    $self->startUninstall;
    $self->SUPER::uninstall;
    $self->finishUninstall;
}

# skip call for super function in 6.17.5
sub patch {
    my $self = shift;

    return if ($self->version() eq '6.17.5');

    $self->SUPER::patch;

    return;
}

1;
